/*
 * winstore.c: Windows-specific implementation of the interface
 * defined in storage.h.
 */

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#if (UNDER_CE < 0x400) && defined(_MSC_VER)
#define assert(x)
#else
#include <assert.h>
#endif

#include "putty.h"
#include "storage.h"
#include "wince_res.h"

static const wchar_t *const reg_jumplist_key = PUTTY_REG_POS L"\\Jumplist";
static const wchar_t *const reg_jumplist_value = L"Recent sessions";
static const wchar_t *const puttystr = PUTTY_REG_POS L"\\Sessions";

static wchar_t seedpath[FILENAME_MAX + 10];

static const char hex[17] = "0123456789ABCDEF";
//static int tried_shgetfolderpath = FALSE;
//typedef HRESULT (WINAPI *p_SHGetFolderPath_t)
//    (HWND, int, HANDLE, DWORD, LPTSTR);
//static p_SHGetFolderPath_t p_SHGetFolderPath = NULL;

static void mungestr(const char *in, char *out)
{
    int candot = 0;

    while (*in) {
	if (*in == ' ' || *in == '\\' || *in == '*' || *in == '?' ||
	    *in == '%' || *in < ' ' || *in > '~' || (*in == '.'
						     && !candot)) {
	    *out++ = '%';
	    *out++ = hex[((unsigned char) *in) >> 4];
	    *out++ = hex[((unsigned char) *in) & 15];
	} else
	    *out++ = *in;
	in++;
	candot = 1;
    }
    *out = '\0';
    return;
}

static void unmungestr(const char *in, char *out, int outlen)
{
    while (*in) {
	if (*in == '%' && in[1] && in[2]) {
	    int i, j;

	    i = in[1] - '0';
	    i -= (i > 9 ? 7 : 0);
	    j = in[2] - '0';
	    j -= (j > 9 ? 7 : 0);

	    *out++ = (i << 4) + j;
	    if (!--outlen)
		return;
	    in += 3;
	} else {
	    *out++ = *in++;
	    if (!--outlen)
		return;
	}
    }
    *out = '\0';
    return;
}

void *open_settings_w(const char *sessionname, char **errmsg)
{
    HKEY subkey1, sesskey;
    int ret;
    char *p;
    wchar_t *wp;
    char *ap;

    *errmsg = NULL;

    if (!sessionname || !*sessionname)
	sessionname = "Default Settings";

    p = snewn(3 * strlen(sessionname) + 1, char);
    mungestr(sessionname, p);

    ap = unicode2ansi(puttystr);

    ret = RegCreateKey(HKEY_CURRENT_USER, puttystr, &subkey1);
    if (ret != ERROR_SUCCESS) {
	sfree(p);
        *errmsg = dupprintf("%s\n"
                            "HKEY_CURRENT_USER\\%s",
                            load_ansi_string(IDS_WARN_UNABLE_CREATE_REGKEY), ap);
        sfree(ap);
	return NULL;
    }
    wp = ansi2unicode(p);
    ret = RegCreateKey(subkey1, wp, &sesskey);
    RegCloseKey(subkey1);
    if (ret != ERROR_SUCCESS) {
        *errmsg = dupprintf("%s\n"
                            "HKEY_CURRENT_USER\\%s\\%s",
                            load_ansi_string(IDS_WARN_UNABLE_CREATE_REGKEY), ap, p);
        sfree(ap);
	sfree(p);
	sfree(wp);
	return NULL;
    }
    sfree(ap);
    sfree(p);
    sfree(wp);
    return (void *) sesskey;
}

void write_setting_s(void *handle, const char *key, const char *value)
{
	wchar_t *wkey, *wvalue;

	if (handle) {
		wkey = ansi2unicode(key);
		wvalue = ansi2unicode(value);
		RegSetValueEx((HKEY) handle, wkey, 0, REG_SZ, (LPBYTE)wvalue,
			(wcslen(wvalue) +1) *sizeof(wchar_t));
		sfree(wkey);
		sfree(wvalue);
	}
}

void write_setting_i(void *handle, const char *key, int value)
{
	wchar_t *wkey;

	if (handle) {
		wkey = ansi2unicode(key);
		RegSetValueEx((HKEY) handle, wkey, 0, REG_DWORD,
			(LPBYTE)&value, sizeof(value));
		sfree(wkey);
	}
}

void close_settings_w(void *handle)
{
    RegCloseKey((HKEY) handle);
}

void *open_settings_r(const char *sessionname)
{
    HKEY subkey1, sesskey;
    char *p;
    wchar_t *wp;

    if (!sessionname || !*sessionname)
	sessionname = "Default Settings";

    p = snewn(3 * strlen(sessionname) + 1, char);
    mungestr(sessionname, p);

    if (RegOpenKey(HKEY_CURRENT_USER, puttystr, &subkey1) != ERROR_SUCCESS) {
	    sesskey = NULL;
    } else {
	wp = ansi2unicode(p);
	if (RegOpenKey(subkey1, wp, &sesskey) != ERROR_SUCCESS) {
		sesskey = NULL;
	}
	RegCloseKey(subkey1);
	sfree(wp);
    }

    sfree(p);
    return (void *) sesskey;
}

char *read_setting_s(void *handle, const char *key)
{
	DWORD type, allocsize, size;
	wchar_t	*wkey, *wbuffer;
	char *ret;

	if (!handle)
		return NULL;

	wkey = ansi2unicode(key);

	/* Find out the type and size of the data. */
	type = size = 0;
	if (RegQueryValueEx((HKEY) handle, wkey, 0,
			    &type, NULL, &size) != ERROR_SUCCESS ||
		type != REG_SZ) {
		sfree(wkey);
		return NULL;
	}

	allocsize = size + 1;		/* allow for an extra NUL if needed */
	wbuffer = snewn(allocsize, wchar_t);
	if (RegQueryValueEx((HKEY) handle, wkey, 0,
			    &type, (BYTE *) wbuffer, &size) != ERROR_SUCCESS ||
		type != REG_SZ) {
		sfree(wkey);
		sfree(wbuffer);
		return NULL;
	}
	assert(size < allocsize);
	wbuffer[size] = L'\0'; /* add an extra NUL in case RegQueryValueEx
				* didn't supply one */

	ret = unicode2ansi(wbuffer);
	sfree(wkey);
	sfree(wbuffer);

	return ret;
}

int read_setting_i(void *handle, const char *key, int defvalue)
{
	DWORD type, val, size;
	wchar_t *wkey;
	size = sizeof(val);

	wkey = ansi2unicode(key);
	if (!handle ||
		RegQueryValueEx((HKEY) handle, wkey, 0, &type,
		(LPBYTE)&val, &size) != ERROR_SUCCESS ||
		size != sizeof(val) || type != REG_DWORD) {
		sfree(wkey);
		return defvalue;
	} else {
		sfree(wkey);
		return val;
	}
}

FontSpec *read_setting_fontspec(void *handle, const char *name)
{
    char *settingname;
    char *fontname;
    FontSpec *ret;
    int isbold, height, charset;

    fontname = read_setting_s(handle, name);
    if (!fontname)
	return NULL;

    settingname = dupcat(name, "IsBold", NULL);
    isbold = read_setting_i(handle, settingname, -1);
    sfree(settingname);
    if (isbold == -1) {
        sfree(fontname);
        return NULL;
    }

    settingname = dupcat(name, "CharSet", NULL);
    charset = read_setting_i(handle, settingname, -1);
    sfree(settingname);
    if (charset == -1) {
        sfree(fontname);
        return NULL;
    }

    settingname = dupcat(name, "Height", NULL);
    height = read_setting_i(handle, settingname, INT_MIN);
    sfree(settingname);
    if (height == INT_MIN) {
        sfree(fontname);
        return NULL;
    }

    ret = fontspec_new(fontname, isbold, height, charset);
    sfree(fontname);
    return ret;
}

void write_setting_fontspec(void *handle, const char *name, FontSpec *font)
{
    char *settingname;

    write_setting_s(handle, name, font->name);
    settingname = dupcat(name, "IsBold", NULL);
    write_setting_i(handle, settingname, font->isbold);
    sfree(settingname);
    settingname = dupcat(name, "CharSet", NULL);
    write_setting_i(handle, settingname, font->charset);
    sfree(settingname);
    settingname = dupcat(name, "Height", NULL);
    write_setting_i(handle, settingname, font->height);
    sfree(settingname);
}

Filename *read_setting_filename(void *handle, const char *name)
{
    char *tmp = read_setting_s(handle, name);
    if (tmp) {
        Filename *ret = filename_from_str(tmp);
	sfree(tmp);
	return ret;
    } else
	return NULL;
}

void write_setting_filename(void *handle, const char *name, Filename *result)
{
    write_setting_s(handle, name, result->path);
}

void close_settings_r(void *handle)
{
    RegCloseKey((HKEY) handle);
}

void del_settings(const char *sessionname)
{
    HKEY subkey1;
    char *p;
    wchar_t *wp;

    if (RegOpenKey(HKEY_CURRENT_USER, puttystr, &subkey1) != ERROR_SUCCESS)
	return;

    p = snewn(3 * strlen(sessionname) + 1, char);
    mungestr(sessionname, p);
    wp = ansi2unicode(p);
    RegDeleteKey(subkey1, wp);
    sfree(p);
    sfree(wp);

    RegCloseKey(subkey1);
}

struct enumsettings {
    HKEY key;
    int i;
};

void *enum_settings_start(void)
{
    struct enumsettings *ret;
    HKEY key;

    if (RegOpenKey(HKEY_CURRENT_USER, puttystr, &key) != ERROR_SUCCESS)
	return NULL;

    ret = snew(struct enumsettings);
    if (ret) {
	ret->key = key;
	ret->i = 0;
    }

    return ret;
}

char *enum_settings_next(void *handle, char *buffer, int buflen)
{
    struct enumsettings *e = (struct enumsettings *) handle;
    char *otherbuf;
    wchar_t *wotherbuf;

    otherbuf = snewn(3 * buflen, char);
    wotherbuf = snewn(buflen, wchar_t);

    if (RegEnumKey(e->key, e->i++, wotherbuf, 3 * buflen) == ERROR_SUCCESS) {
	wcstombs(otherbuf, wotherbuf, buflen *3);
	unmungestr(otherbuf, buffer, buflen);
	sfree(otherbuf);
	sfree(wotherbuf);
	return buffer;
    } else {
	sfree(otherbuf);
	sfree(wotherbuf);
	return NULL;
    }
}

void enum_settings_finish(void *handle)
{
    struct enumsettings *e = (struct enumsettings *) handle;
    RegCloseKey(e->key);
    sfree(e);
}

static void hostkey_regname(char *buffer, const char *hostname,
			    int port, const char *keytype)
{
    int len;
    strcpy(buffer, keytype);
    strcat(buffer, "@");
    len = strlen(buffer);
    len += sprintf(buffer + len, "%d:", port);
    mungestr(hostname, buffer + strlen(buffer));
}

int verify_host_key(const char *hostname, int port,
		    const char *keytype, const char *key)
{
    char *otherstr, *regname;
    int len;
    HKEY rkey;
    DWORD readlen;
    DWORD type;
    int ret, compare;
    wchar_t *wregname;
    wchar_t *wotherstr;

    len = 1 + strlen(key);

    /*
     * Now read a saved key in from the registry and see what it
     * says.
     */
    otherstr = snewn(len, char);
    wotherstr = snewn(len * sizeof(wchar_t), wchar_t);
    regname = snewn(3 * (strlen(hostname) + strlen(keytype)) + 15, char);

    hostkey_regname(regname, hostname, port, keytype);

    if (RegOpenKey(HKEY_CURRENT_USER, PUTTY_REG_POS TEXT("\\SshHostKeys"),
		   &rkey) != ERROR_SUCCESS)
	return 1;		       /* key does not exist in registry */

    readlen = len * sizeof(wchar_t);
    wregname = ansi2unicode(regname);
    ret = RegQueryValueEx(rkey, wregname, NULL, &type, (LPBYTE)wotherstr, &readlen);
    otherstr = unicode2ansi(wotherstr);

    if (ret != ERROR_SUCCESS && ret != ERROR_MORE_DATA &&
	!strcmp(keytype, "rsa")) {
	/*
	 * Key didn't exist. If the key type is RSA, we'll try
	 * another trick, which is to look up the _old_ key format
	 * under just the hostname and translate that.
	 */
	char *justhost = regname + 1 + strcspn(regname, ":");
	char *oldstyle;
	wchar_t *woldstyle = snewn((len + 10) * 3, wchar_t);
	wchar_t *wjusthost;
	readlen = len;
	wjusthost = ansi2unicode(justhost);
	ret = RegQueryValueEx(rkey, wjusthost, NULL, &type, (LPBYTE)woldstyle, &readlen);
	oldstyle = unicode2ansi(woldstyle);
	sfree(woldstyle);

	if (ret == ERROR_SUCCESS && type == REG_SZ) {
	    /*
	     * The old format is two old-style bignums separated by
	     * a slash. An old-style bignum is made of groups of
	     * four hex digits: digits are ordered in sensible
	     * (most to least significant) order within each group,
	     * but groups are ordered in silly (least to most)
	     * order within the bignum. The new format is two
	     * ordinary C-format hex numbers (0xABCDEFG...XYZ, with
	     * A nonzero except in the special case 0x0, which
	     * doesn't appear anyway in RSA keys) separated by a
	     * comma. All hex digits are lowercase in both formats.
	     */
	    char *p = otherstr;
	    char *q = oldstyle;
	    int i, j;

	    for (i = 0; i < 2; i++) {
		int ndigits, nwords;
		*p++ = '0';
		*p++ = 'x';
		ndigits = strcspn(q, "/");	/* find / or end of string */
		nwords = ndigits / 4;
		/* now trim ndigits to remove leading zeros */
		while (q[(ndigits - 1) ^ 3] == '0' && ndigits > 1)
		    ndigits--;
		/* now move digits over to new string */
		for (j = 0; j < ndigits; j++)
		    p[ndigits - 1 - j] = q[j ^ 3];
		p += ndigits;
		q += nwords * 4;
		if (*q) {
		    q++;	       /* eat the slash */
		    *p++ = ',';	       /* add a comma */
		}
		*p = '\0';	       /* terminate the string */
	    }

	    /*
	     * Now _if_ this key matches, we'll enter it in the new
	     * format. If not, we'll assume something odd went
	     * wrong, and hyper-cautiously do nothing.
	     */
	    if (!strcmp(otherstr, key)) {
		RegSetValueEx(rkey, wregname, 0, REG_SZ, (LPBYTE)wotherstr,
			      (wcslen(wotherstr) + 1) * sizeof(wchar_t));
	    }
	}
    }

    RegCloseKey(rkey);

    compare = strcmp(otherstr, key);

    sfree(otherstr);
    sfree(regname);
    sfree(wregname);
    sfree(wotherstr);

    if (ret == ERROR_MORE_DATA ||
	(ret == ERROR_SUCCESS && type == REG_SZ && compare))
	return 2;		       /* key is different in registry */
    else if (ret != ERROR_SUCCESS || type != REG_SZ)
	return 1;		       /* key does not exist in registry */
    else
	return 0;		       /* key matched OK in registry */
}

void store_host_key(const char *hostname, int port,
		    const char *keytype, const char *key)
{
    char *regname;
    wchar_t *wregname;
    wchar_t *wkey;
    HKEY rkey;

    regname = snewn(3 * (strlen(hostname) + strlen(keytype)) + 15, char);
    hostkey_regname(regname, hostname, port, keytype);
    wregname = ansi2unicode(regname);
    wkey = ansi2unicode(key);

    if (RegCreateKey(HKEY_CURRENT_USER, PUTTY_REG_POS TEXT("\\SshHostKeys"),
		     &rkey) == ERROR_SUCCESS) {
	RegSetValueEx(rkey, wregname, 0, REG_SZ, (const BYTE *)wkey, (wcslen(wkey) + 1) * sizeof(wchar_t));
	RegCloseKey(rkey);
    } /* else key does not exist in registry */

    sfree(regname);
    sfree(wregname);
    sfree(wkey);
}

/*
 * Find the random seed file path and store it in `seedpath'.
 */
static void get_seedpath(void)
{
    HKEY rkey;
    DWORD type, size;

    size = sizeof(seedpath);

    if (RegOpenKey(HKEY_CURRENT_USER, PUTTY_REG_POS, &rkey) ==
	ERROR_SUCCESS) {
	int ret = RegQueryValueEx(rkey, TEXT("RandSeedFile"),
				  0, &type, (LPBYTE) seedpath, &size);
	if (ret != ERROR_SUCCESS || type != REG_SZ)
	    seedpath[0] = '\0';
	RegCloseKey(rkey);
    } else
	seedpath[0] = '\0';

    if (!seedpath[0]) {
	int len;

	wcscpy(seedpath, L"\\My Documents");
	len = wcslen(seedpath) - 1;
	wcscpy(seedpath + len, L"\\PUTTY.RND");
    }
}

void read_random_seed(noise_consumer_t consumer)
{
    HANDLE seedf;

    if (!seedpath[0])
	get_seedpath();

    seedf = CreateFile(seedpath, GENERIC_READ,
		       FILE_SHARE_READ | FILE_SHARE_WRITE,
		       NULL, OPEN_EXISTING, 0, NULL);

    if (seedf != INVALID_HANDLE_VALUE) {
	while (1) {
	    char buf[1024];
	    DWORD len;

	    if (ReadFile(seedf, buf, sizeof(buf), &len, NULL) && len)
		consumer(buf, len);
	    else
		break;
	}
	CloseHandle(seedf);
    }
}

void write_random_seed(void *data, int len)
{
    HANDLE seedf;

    if (!seedpath[0])
	get_seedpath();

    seedf = CreateFile(seedpath, GENERIC_WRITE, 0,
		       NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);

    if (seedf != INVALID_HANDLE_VALUE) {
	DWORD lenwritten;

	WriteFile(seedf, data, len, &lenwritten, NULL);
	CloseHandle(seedf);
    }
}

/*
 * Internal function supporting the jump list registry code. All the
 * functions to add, remove and read the list have substantially
 * similar content, so this is a generalisation of all of them which
 * transforms the list in the registry by prepending 'add' (if
 * non-null), removing 'rem' from what's left (if non-null), and
 * returning the resulting concatenated list of strings in 'out' (if
 * non-null).
 */
static int transform_jumplist_registry
    (const char *add, const char *rem, char **out)
{
    int ret;
    HKEY pjumplist_key, psettings_tmp;
    DWORD type;
    DWORD value_length;
    wchar_t *old_value, *new_value;
    wchar_t *piterator_old, *piterator_new, *piterator_tmp;
    char *aout;

    ret = RegCreateKeyEx(HKEY_CURRENT_USER, reg_jumplist_key, 0, NULL,
                         REG_OPTION_NON_VOLATILE, (KEY_READ | KEY_WRITE), NULL,
                         &pjumplist_key, NULL);
    if (ret != ERROR_SUCCESS) {
	return JUMPLISTREG_ERROR_KEYOPENCREATE_FAILURE;
    }

    /* Get current list of saved sessions in the registry. */
    value_length = 200;
    old_value = snewn(value_length, wchar_t);
    ret = RegQueryValueEx(pjumplist_key, reg_jumplist_value, NULL, &type,
                          (PBYTE) old_value, &value_length);
    /* When the passed buffer is too small, ERROR_MORE_DATA is
     * returned and the required size is returned in the length
     * argument. */
    if (ret == ERROR_MORE_DATA) {
        sfree(old_value);
        old_value = snewn(value_length, wchar_t);
        ret = RegQueryValueEx(pjumplist_key, reg_jumplist_value, NULL, &type,
                              (PBYTE) old_value, &value_length);
    }

    if (ret == ERROR_FILE_NOT_FOUND) {
        /* Value doesn't exist yet. Start from an empty value. */
        *old_value = L'\0';
        *(old_value + 1) = L'\0';
    } else if (ret != ERROR_SUCCESS) {
        /* Some non-recoverable error occurred. */
        sfree(old_value);
        RegCloseKey(pjumplist_key);
        return JUMPLISTREG_ERROR_VALUEREAD_FAILURE;
    } else if (type != REG_MULTI_SZ) {
        /* The value present in the registry has the wrong type: we
         * try to delete it and start from an empty value. */
        ret = RegDeleteValue(pjumplist_key, reg_jumplist_value);
        if (ret != ERROR_SUCCESS) {
            sfree(old_value);
            RegCloseKey(pjumplist_key);
            return JUMPLISTREG_ERROR_VALUEREAD_FAILURE;
        }

        *old_value = L'\0';
        *(old_value + 1) = L'\0';
    }

    /* Check validity of registry data: REG_MULTI_SZ value must end
     * with \0\0. */
    piterator_tmp = old_value;
    while (((piterator_tmp - old_value) < (value_length - 1)) &&
           !(*piterator_tmp == L'\0' && *(piterator_tmp+1) == L'\0')) {
        ++piterator_tmp;
    }

    if ((piterator_tmp - old_value) >= (value_length-1)) {
        /* Invalid value. Start from an empty value. */
        *old_value = L'\0';
        *(old_value + 1) = L'\0';
    }

    /*
     * Modify the list, if we're modifying.
     */
    if (add || rem) {
        wchar_t *wadd = 0, *wrem = 0;
        if (add) wadd = ansi2unicode(add);
        if (rem) wrem = ansi2unicode(rem);

        /* Walk through the existing list and construct the new list of
         * saved sessions. */
        new_value = snewn(value_length + (wadd ? wcslen(wadd) + 1 : 0), wchar_t);
        piterator_new = new_value;
        piterator_old = old_value;

        /* First add the new item to the beginning of the list. */
        if (wadd) {
            wcscpy(piterator_new, wadd);
            piterator_new += wcslen(piterator_new) + 1;
        }
        /* Now add the existing list, taking care to leave out the removed
         * item, if it was already in the existing list. */
        while (*piterator_old != L'\0') {
            if (!wrem || wcscmp(piterator_old, wrem) != 0) {
                char *apitr;
                /* Check if this is a valid session, otherwise don't add. */
                apitr = unicode2ansi(piterator_old);
                psettings_tmp = open_settings_r(apitr);
                sfree(apitr);
                if (psettings_tmp != NULL) {
                    close_settings_r(psettings_tmp);
                    wcscpy(piterator_new, piterator_old);
                    piterator_new += wcslen(piterator_new) + 1;
                }
            }
            piterator_old += wcslen(piterator_old) + 1;
        }
        *piterator_new = L'\0';
        ++piterator_new;

        /* Save the new list to the registry. */
        ret = RegSetValueEx(pjumplist_key, reg_jumplist_value, 0, REG_MULTI_SZ,
                            (PBYTE) new_value, (DWORD) (piterator_new - new_value));

        sfree(wrem);
        sfree(wadd);
        sfree(old_value);
        old_value = new_value;
    } else
        ret = ERROR_SUCCESS;

    /*
     * Either return or free the result.
     */
    if (out && ret == ERROR_SUCCESS) {
        aout = unicode2ansi(old_value);
        *out = aout;
    }

    sfree(old_value);

    /* Clean up and return. */
    RegCloseKey(pjumplist_key);

    if (ret != ERROR_SUCCESS) {
        return JUMPLISTREG_ERROR_VALUEWRITE_FAILURE;
    } else {
        return JUMPLISTREG_OK;
    }
}

/* Adds a new entry to the jumplist entries in the registry. */
int add_to_jumplist_registry(const char *item)
{
    return transform_jumplist_registry(item, item, NULL);
}

/* Removes an item from the jumplist entries in the registry. */
int remove_from_jumplist_registry(const char *item)
{
    return transform_jumplist_registry(NULL, item, NULL);
}

/* Returns the jumplist entries from the registry. Caller must free
 * the returned pointer. */
char *get_jumplist_registry_entries (void)
{
    char *list_value;

    if (transform_jumplist_registry(NULL,NULL,&list_value) != JUMPLISTREG_OK) {
	list_value = snewn(2, char);
        *list_value = '\0';
        *(list_value + 1) = '\0';
    }
    return list_value;
}

/*
 * Recursively delete a registry key and everything under it.
 */
static void registry_recursive_remove(HKEY key)
{
    DWORD i;
    wchar_t name[MAX_PATH + 1];
    HKEY subkey;

    i = 0;
    while (RegEnumKey(key, i, name, sizeof(name)) == ERROR_SUCCESS) {
	if (RegOpenKey(key, name, &subkey) == ERROR_SUCCESS) {
	    registry_recursive_remove(subkey);
	    RegCloseKey(subkey);
	}
	RegDeleteKey(key, name);
    }
}

BOOL remove(TCHAR *filename)
{
	return DeleteFile(filename);
}

void cleanup_all(void)
{
    HKEY key;
    int ret;
    wchar_t name[MAX_PATH + 1];

    /* ------------------------------------------------------------
     * Wipe out the random seed file.
     */
    if (!seedpath[0])
	get_seedpath();
    remove(seedpath);

    /* ------------------------------------------------------------
     * Ask Windows to delete any jump list information associated
     * with this installation of PuTTY.
     */
    clear_jumplist();

    /* ------------------------------------------------------------
     * Destroy all registry information associated with PuTTY.
     */

    /*
     * Open the main PuTTY registry key and remove everything in it.
     */
    if (RegOpenKey(HKEY_CURRENT_USER, PUTTY_REG_POS, &key) ==
	ERROR_SUCCESS) {
	registry_recursive_remove(key);
	RegCloseKey(key);
    }
    /*
     * Now open the parent key and remove the PuTTY main key. Once
     * we've done that, see if the parent key has any other
     * children.
     */
    if (RegOpenKey(HKEY_CURRENT_USER, PUTTY_REG_PARENT,
		   &key) == ERROR_SUCCESS) {
	RegDeleteKey(key, PUTTY_REG_PARENT_CHILD);
	ret = RegEnumKey(key, 0, name, sizeof(name));
	RegCloseKey(key);
	/*
	 * If the parent key had no other children, we must delete
	 * it in its turn. That means opening the _grandparent_
	 * key.
	 */
	if (ret != ERROR_SUCCESS) {
	    if (RegOpenKey(HKEY_CURRENT_USER, PUTTY_REG_GPARENT,
			   &key) == ERROR_SUCCESS) {
		RegDeleteKey(key, PUTTY_REG_GPARENT_CHILD);
		RegCloseKey(key);
	    }
	}
    }
    /*
     * Now we're done.
     */
}
