/*
 * wince_window.c - the PuTTY(tel) main program, which runs a PuTTY terminal
 * emulator and backend in a window.
 */

#include <stdio.h>
#include <windows.h>
#include <stdlib.h>
#include <ctype.h>
#include <limits.h>
#include <commctrl.h>

#ifdef __GNUC__
#include <tchar.h>
extern int _wtoi(const wchar_t*);
#endif

#define PUTTY_DO_GLOBALS /* actually _define_ globals */
#include "putty.h"
#if (UNDER_CE < 0x400) & defined(_MSC_VER)
#include "wince_time.h"
#define assert(x)
#else
#include <time.h>
#include <assert.h>
#endif
#include "terminal.h"
#include "storage.h"
#include "wince_res.h"

#ifdef SHELL_AYGSHELL
#include "wince_aygshell.h"
#endif
#include "wince_device.h"
#include "wince_config.h"

#include <imm.h>
#include <sipapi.h>
#include <commctrl.h>
#include <richedit.h>
#include <mmsystem.h>

static HWND hwndCB;
static int g_bTerminate; /* Thread termination flag */
static int pending_event = FALSE;
static int fatalbox_visible = FALSE;

#define IDM_SPECIALSEP 0x41F0
#define IDM_SPECIAL_MIN 0x4200
#define IDM_SPECIAL_MAX 0x4800

#define IDM_SAVED_MIN 0x4C00
#define IDM_SAVED_MAX 0x5000
#define MENU_SAVED_STEP 16
/* Maximum number of sessions on saved-session submenu */
#define MENU_SAVED_MAX ((IDM_SAVED_MAX-IDM_SAVED_MIN) / MENU_SAVED_STEP)

#define WM_IGNORE_CLIP    (WM_XUSER + 2)
#define WM_AGENT_CALLBACK (WM_XUSER + 4)
#define WM_GOT_CLIPDATA   (WM_XUSER + 6)

/* Needed for Chinese support and apparently not always defined. */
#ifndef VK_PROCESSKEY
#define VK_PROCESSKEY 0xE5
#endif

/*
 * This could easily be made resizable at runtime if we so desired.
 */

static struct wince_socket_registration
{
	SOCKET sock;
	HWND   hwnd;
	UINT   message;
} wince_sockets[WINCE_MAX_SOCKETS];

static Mouse_Button translate_button(Mouse_Button button);
static LRESULT CALLBACK WndProc(HWND, UINT, WPARAM, LPARAM);
static int TranslateKey(UINT message, WPARAM wParam, LPARAM lParam,
			unsigned char *output);
static void conftopalette(void);
static void systopalette(void);
static void init_palette(void);
static void init_fonts(int, int);
static void another_font(int);
static void deinit_fonts(void);
static void set_input_locale(HKL);
static void update_savedsess_menu(void);
static void reset_window(int reinit);

static BOOL current_full_screen;
static int is_full_screen(void);
static void make_full_screen(void);
static void clear_full_screen(void);
static void flip_full_screen(void);
static int process_clipdata(HLOCAL clipdata, int unicode);
static void enact_pending_netevent(void);
static BOOL set_window_style(HWND hwnd, DWORD_PTR add, DWORD_PTR remove);

int debug_level = DBG_ALL;
int wince_device = 0;
static int font_width, font_height, font_dualwidth, font_varpitch;
static int offset_width, offset_height;
static UINT override_menu_height;
static int was_zoomed = 0;
static int prev_rows, prev_cols;

static void sys_cursor_update(void);
static int get_fullscreen_rect(RECT * ss);

static int caret_x = -1, caret_y = -1;

static int kbd_codepage;

static void *ldisc;
static Backend *back;
static void *backhandle;

static int got_focus;
static int pending_netevent = 0;
static WPARAM pend_netevent_wParam = 0;
static LPARAM pend_netevent_lParam = 0;
static HANDLE wince_net_event;
#ifdef SHELL_AYGSHELL
static SHACTIVATEINFO sai;
#endif
static HWND CreateRpCommandBar(HWND hwnd);
static enum { CEHPC, CEPPC, CESP };
static int cetype;

static struct unicode_data ucsdata;
static int session_closed;
static int reconfiguring;

static const struct telnet_special *specials = NULL;
static HMENU specials_menu = NULL;
static int n_specials = 0;

static wchar_t *clipboard_contents;
static size_t clipboard_length;

#define TIMING_TIMER_ID 1234
static long timing_next_time;

static struct {
	HMENU menu;
} popup_menus[2];
enum { SYSMENU, CTXMENU };
static HMENU savedsess_menu;
#define MENU_SAVEDSESS_LENGTH 64

struct wm_netevent_params {
	/* Used to pass data to wm_netevent_callback */
	WPARAM wParam;
	LPARAM lParam;
};

Conf *conf;					/* exported to wince_dlg.c */

static void conf_cache_data(void);
int cursor_type;
int vtmode;

static struct sesslist sesslist;       /* for saved-session menu */

struct agent_callback {
	void (*callback)(void *, void *, int);
	void *callback_ctx;
	void *data;
	int len;
};

#define FONT_NORMAL 0
#define FONT_BOLD 1
#define FONT_UNDERLINE 2
#define FONT_BOLDUND 3
#define FONT_WIDE	0x04
#define FONT_HIGH	0x08
#define FONT_NARROW	0x10

#define FONT_OEM 	0x20
#define FONT_OEMBOLD 	0x21
#define FONT_OEMUND 	0x22
#define FONT_OEMBOLDUND 0x23

#define FONT_MAXNO 	0x2F
#define FONT_SHIFT	5
static HFONT fonts[FONT_MAXNO];
static LOGFONT lfont;
static int fontflag[FONT_MAXNO];
static enum {
	BOLD_NONE, BOLD_SHADOW, BOLD_FONT
} bold_font_mode;
static int bold_colours;
static enum {
	UND_LINE, UND_FONT
} und_mode;
static int descent;

#define NCFGCOLOURS 22
#define NEXTCOLOURS 240
#define NALLCOLOURS (NCFGCOLOURS + NEXTCOLOURS)
static COLORREF colours[NALLCOLOURS];
static HPALETTE pal;
static LPLOGPALETTE logpal;
static RGBTRIPLE defpal[NALLCOLOURS];

static HBITMAP caretbm;

static int dbltime, lasttime, lastact;
static Mouse_Button lastbtn;

/* this allows xterm-style mouse handling. */
static int send_raw_mouse = 0;

static int busy_status = BUSY_NOT;

static char *window_name, *icon_name;

static int compose_state = 0;

const int share_can_be_downstream = TRUE;
const int share_can_be_upstream = TRUE;

void logevent(void *frontend, const char *string) { debug_level((DBG_FUNCTION, "log_event(%s);\n", string)); }
void frontend_keypress(void *handle) { debug_level((DBG_FUNCTION, "frontend_keypress();\n")); }
void set_iconic(void *frontend, int iconic) { debug_level((DBG_FUNCTION, "set_iconic();\n")); }
void ldisc_update(void *frontend, int echo, int edit) { debug_level((DBG_FUNCTION, "ldisc_update();\n")); }
static void flash_window(int mode) { debug_level((DBG_FUNCTION, "flash_window();\n")); }
BOOL IsZoomed(HWND hwnd);
int agent_exists(void) {  debug_level((DBG_FUNCTION, "agent_exists();\n")); return FALSE; }
int is_iconic(void *frontend) { debug_level((DBG_FUNCTION, "is_iconic();\n")); return FALSE; }

void wince_dispatch_net_events(WSAEVENT event)
{
	WSANETWORKEVENTS events;
	int              i;
	int              j;

	debug_level((DBG_FUNCTION, "wince_dispatch_net_events();\n"));

	/*
	 * Loop through our set of registered sockets and discover which, if any,
	 * actually have a pending network event.  We'll send the window message
	 * for each that does.  In reality, it would be better to process the data
	 * inline instead of making the call through the windowing system, but we
	 * are aiming to minimize sweeping changes.  That part of PuTTY really
	 * needs to be rewritten even for non-WinCE anyways, though.
	 */

	for (i = 0;
	     i < WINCE_MAX_SOCKETS;
	     i += 1)
	{
		if (wince_sockets[i].sock == INVALID_SOCKET ||
		    wince_sockets[i].hwnd == NULL)
			continue;

		if (p_WSAEnumNetworkEvents(wince_sockets[i].sock, event, &events) == SOCKET_ERROR)
		{
			debug_level((DBG_FUNCTION, "wince_dispatch_net_events():"
				    " - WSAEnumNetworkEvents failed for socket %lu.  Canceling the registration.\n",
				    wince_sockets[i].sock));

			/*
			 * Cancel the registration with Winsock but leave the socket there,
			 * so that we can remove it.
			 */

			p_WSAEventSelect(wince_sockets[i].sock, event, 0);

			/*
			 * Clear the ``hwnd'' field as an indication that we have removed
			 * the registration but are still waiting for someone to close the
			 * actual socket's select entry.
			 */

			wince_sockets[i].hwnd = 0;
		}

		/*
		 * Send the message on it's way for each bit that was set.
		 */

		for (j = 0;
		     j < FD_MAX_EVENTS;
		     j += 1)
		{
			if (events.lNetworkEvents & (1 << j))
			{
				debug_level((DBG_VERBOSE, "wince_dispatch_net_events():"
					    " - socket %lu has event %08X (error %lu)\n",
					    wince_sockets[i].sock,
					    1 << j,
					    events.iErrorCode[j]));

				SendMessage(wince_sockets[i].hwnd,
					    wince_sockets[i].message,
					    (WPARAM)wince_sockets[i].sock,
					    (LPARAM)((1 << j) | (events.iErrorCode[j] << 16))
					    );
			}
		}
	}

}

/* Dummy routine, only required in plink. */
void frontend_echoedit_update(void *frontend, int echo, int edit)
{
}

char *get_ttymode(void *frontend, const char *mode)
{
	debug_level((DBG_FUNCTION, "get_ttymode(%s);\n", mode));
	return term_get_ttymode(term, mode);
}

static void start_backend(void)
{
	const char *error;
	char msg[1024], *title;
	char *realhost;
	int i;

	/*
	 * Select protocol. This is farmed out into a table in a
	 * separate file to enable an ssh-free variant.
	 */
	back = backend_from_proto(conf_get_int(conf, CONF_protocol));
	if (back == NULL) {
		char *cap = dupprintf(load_ansi_string(IDS_WARN_CAPTION_INTERNAL_ERROR), appname);
		MessageBoxA(NULL, load_ansi_string(IDS_WARN_UNSUPPORTED_PROTOCOL_NUM),
			    cap, MB_OK | MB_ICONEXCLAMATION | MB_TOPMOST);
		sfree(cap);
		cleanup_exit(1);
	}

	error = back->init(NULL, &backhandle, conf,
			   conf_get_str(conf, CONF_host),
			   conf_get_int(conf, CONF_port),
			   &realhost,
			   conf_get_int(conf, CONF_tcp_nodelay),
			   conf_get_int(conf, CONF_tcp_keepalives));
	back->provide_logctx(backhandle, logctx);
	if (error) {
		char *cap = dupprintf(load_ansi_string(IDS_WARN_CAPTION_ERROR), appname);
		sprintf(msg, load_ansi_string(IDS_WARN_UNABLE_TO_OPEN_CONNECTION),
			conf_dest(conf), error);
		MessageBoxA(NULL, msg, cap, MB_ICONERROR | MB_OK | MB_TOPMOST);
		sfree(cap);
		exit(0);
	}
	window_name = icon_name = NULL;
	title = conf_get_str(conf, CONF_wintitle);
	if (!title || !*title) {
		sprintf(msg, "%.930s - %.70s", realhost, appname);
		title =	msg;
	}
	sfree(realhost);
	set_title(NULL, title);
	set_icon(NULL, title);

	/*
	* Connect the terminal to the backend for resize purposes.
	*/
	term_provide_resize_fn(term, back->size, backhandle);

	/*
	* Set up a line discipline.
	*/
	ldisc = ldisc_create(conf, term, back, backhandle, NULL);

	/*
	 * Destroy the Restart Session menu item. (This will return
	 * failure if it's already absent, as it will be the very first
	 * time we call this function. We ignore that, because as long
	 * as the menu item ends up not being there, we don't care
	 * whether it was us who removed it or not!)
	 */
	i = 1;
	EnableMenuItem(popup_menus[i].menu, IDM_RESTART, MF_BYCOMMAND | MF_GRAYED);
	EnableMenuItem(popup_menus[i].menu, IDM_CLOSESESSION, MF_BYCOMMAND | MF_ENABLED);

	session_closed = FALSE;
}

static void close_session(void *ignored_context)
{
	char morestuff[100];
	int  i;

	debug_level((DBG_FUNCTION, "close_session();\n"));

	session_closed = TRUE;
	sprintf(morestuff, "%.70s (inactive)", appname);
	set_icon(NULL, morestuff);
	set_title(NULL, morestuff);

	if (ldisc) {
		ldisc_free(ldisc);
		ldisc = NULL;
	}
	if (back) {
		back->free(backhandle);
		backhandle = NULL;
		back = NULL;
		term_provide_resize_fn(term, NULL, NULL);
		update_specials_menu(NULL);
	}

	/*
	 * Show the Restart Session menu item. Do a precautionary
	 * delete first to ensure we never end up with more than one.
	 */
	i = 1;
	EnableMenuItem(popup_menus[i].menu, IDM_RESTART, MF_BYCOMMAND | MF_ENABLED);
	EnableMenuItem(popup_menus[i].menu, IDM_CLOSESESSION, MF_BYCOMMAND | MF_GRAYED);
}

int WINAPI WinMain(HINSTANCE inst, HINSTANCE prev, LPWSTR wcmdline, int show)
{
	WNDCLASS wndclass;
	MSG msg;
	HRESULT hr;
	HANDLE event;
	char *cmdline, *plat, *oem;
	wchar_t wplat[MAX_PATH], woem[MAX_PATH];
	char *ad;
	struct wince_dev_t *wd;
	int i;

	hinst = inst;
	hwnd = NULL;
	flags = FLAG_VERBOSE | FLAG_INTERACTIVE;

	sk_init();
#ifdef SHELL_AYGSHELL
	aygshell_init();
#endif

	/* useless? --ynezz
	*/
	InitCommonControls();
#ifdef SHELL_AYGSHELL
	if (as_SHInitExtraControls)
		as_SHInitExtraControls();
#endif

	/* Ensure a Maximize setting in Explorer doesn't maximise the
	 * config box. */
	defuse_showwindow();
	event = NULL;

	if (!init_winver())
	{
		char *cap = dupprintf(load_ansi_string(IDS_WARN_CAPTION_FALAL_ERROR), appname);
		MessageBoxA(NULL, "Windows refuses to report a version",
			    cap, MB_OK | MB_ICONEXCLAMATION | MB_TOPMOST);
		sfree(cap);
		return 1;
	}

#ifdef DEBUG_UDP
	udplog_init();
#endif

	wince_device = wince_dev_id(&osVersion);

	init_help();

	event = p_WSACreateEvent();

	wince_net_event = event;

	for (i = 0;
	     i < WINCE_MAX_SOCKETS;
	     i += 1)
	{
		wince_sockets[i].sock = INVALID_SOCKET;
	}

	wince_platform(wplat, MAX_PATH-1);
	wince_oeminfo(woem, MAX_PATH-1);
	wd = wince_dev(wince_device);

	if (wcsstr(wplat, L"PocketPC")) {
		cetype = CEPPC;
	}
	else if (wcsstr(wplat, L"Smartphone")) {
		cetype = CESP;
	}
	else {
		cetype = CEHPC;
	}

	ad = unicode2ansi(wd->wince_dev_str);
	plat = unicode2ansi(wplat);
	oem = unicode2ansi(woem);

	debug_level((DBG_USER, "\n\n"
				"PocketPutty: %s\n"
				"Build: %s\n"
				"Platform: \"%s\" OEM: \"%s\"\n"
				"OS Version MA: %i MI: %i BN: %i PI: %i\n"
				"PocketPutty device ID: %s\n\n",
				ver,
				__TIMESTAMP__,
				plat, oem,
				osVersion.dwMajorVersion, osVersion.dwMinorVersion,
				osVersion.dwBuildNumber, osVersion.dwPlatformId,
				ad));
	sfree(plat);
	sfree(oem);
	sfree(ad);

	conf = conf_new();

	/*
	 * Initialize COM.
	 */
	hr = CoInitializeEx(NULL, 0);
	if (hr != S_OK && hr != S_FALSE) {
		char *cap = dupprintf(load_ansi_string(IDS_WARN_CAPTION_FALAL_ERROR), appname);
		MessageBoxA(NULL, "Failed to initialize COM subsystem",
			    cap, MB_OK | MB_ICONEXCLAMATION | MB_TOPMOST);
		sfree(cap);
		return 1;
	}

	/*
	 * Process the command line.
	 */
	{
		char *p;
		int got_host = 0;
		/* By default, we bring up the config dialog, rather than launching
		 * a session. This gets set to TRUE if something happens to change
		 * that (e.g., a hostname is specified on the command-line). */
		int allow_launch = FALSE;

		default_protocol = be_default_protocol;
		/* Find the appropriate default port. */
		{
			Backend *b = backend_from_proto(default_protocol);
			default_port = 0; /* illegal */
			if (b)
				default_port = b->default_port;
		}
		conf_set_int(conf, CONF_logtype, LGTYP_NONE);

		do_defaults(NULL, conf);

		cmdline = unicode2ansi(wcmdline);
		p = cmdline;

		/*
		 * Process a couple of command-line options which are more
		 * easily dealt with before the line is broken up into words.
		 * These are the old-fashioned but convenient @sessionname and
		 * the internal-use-only &sharedmemoryhandle, neither of which
		 * are combined with anything else.
		 */
		while (*p && isspace(*p))
			p++;
		if (*p == '@') {
			/*
			 * An initial @ means that the whole of the rest of the
			 * command line should be treated as the name of a saved
			 * session, with _no quoting or escaping_. This makes it a
			 * very convenient means of automated saved-session
			 * launching, via IDM_SAVEDSESS or Windows 7 jump lists.
			 */
			int i = strlen(p);
			while (i > 1 && isspace(p[i - 1]))
				i--;
			p[i] = '\0';
			do_defaults(p + 1, conf);
			if (!conf_launchable(conf) && !do_config()) {
				cleanup_exit(0);
			}
			allow_launch = TRUE;    /* allow it to be launched directly */
#if 0
		} else if (*p == '&') {
			/*
			 * An initial & means we've been given a command line
			 * containing the hex value of a HANDLE for a file
			 * mapping object, which we must then interpret as a
			 * serialised Conf.
			 */
			HANDLE filemap = NULL;
			Conf *cp = NULL;
			unsigned cpsize = conf_serialised_size(conf);
			if (sscanf(p + 1, "%p:%u", &filemap, &cpsize) == 2 &&
				(cp = (Conf *)MapViewOfFile(filemap, FILE_MAP_READ,
							    0, 0, cpsize)) != NULL) {
				conf_deserialise(conf, cp, cpsize);
				UnmapViewOfFile(cp);
				CloseHandle(filemap);
			} else if (!do_config()) {
				cleanup_exit(0);
			}
			allow_launch = TRUE;
#endif
		} else {
			/*
			* Otherwise, break up the command line and deal with

			* it sensibly.
			*/

			int argc, i;
			char **argv;

			split_into_argv(cmdline, &argc, &argv, NULL);

			for (i = 0; i < argc; i++) {
				char *p = argv[i];
				int ret;

				ret = cmdline_process_param(p, i+1<argc?argv[i+1]:NULL,
							    1, conf);
				if (ret == -2) {
					cmdline_error("option \"%s\" requires an argument", p);
				} else if (ret == 2) {
					i++;	       /* skip next argument */
				} else if (ret == 1) {
					continue;	       /* nothing further needs doing */
				} else if (!strcmp(p, "-cleanup") ||
					!strcmp(p, "-cleanup-during-uninstall")) {
					/*
					* `putty -cleanup'. Remove all registry
					* entries associated with PuTTY, and also find
					* and delete the random seed file.
					*/
					char *s1, *s2;
					/* Are we being invoked from an uninstaller? */
					if (!strcmp(p, "-cleanup-during-uninstall")) {
						s1 = dupprintf("Remove saved sessions and random seed file?\n"
							"\n"
							"If you hit Yes, ALL Registry entries associated\n"
							"with %s will be removed, as well as the\n"
							"random seed file. THIS PROCESS WILL\n"
							"DESTROY YOUR SAVED SESSIONS.\n"
							"(This only affects the currently logged-in user.)\n"
							"\n"
							"If you hit No, uninstallation will proceed, but\n"
							"saved sessions etc will be left on the machine.",
							appname);
						s2 = dupprintf("%s Uninstallation", appname);
					} else {
						s1 = dupprintf("This procedure will remove ALL Registry entries\n"
							"associated with %s, and will also remove\n"
							"the random seed file. (This only affects the\n"
							"currently logged-in user.)\n"
							"\n"
							"THIS PROCESS WILL DESTROY YOUR SAVED SESSIONS.\n"
							"Are you really sure you want to continue?",
							appname);
						s2 = dupprintf("%s Warning", appname);
					}
					if (MessageBoxA(NULL, s1, s2,
							MB_YESNO | MB_ICONWARNING | MB_DEFBUTTON2 |
							MB_TOPMOST) == IDYES) {
						cleanup_all();
					}
					sfree(s1);
					sfree(s2);
					exit(0);
				} else if (!strcmp(p, "-pgpfp")) {
					pgp_fingerprints();
					exit(1);
				} else if (*p != '-') {
					char *q = p;
					if (got_host) {
						/*
						 * If we already have a host name, treat
						 * this argument as a port number. NB we
						 * have to treat this as a saved -P
						 * argument, so that it will be deferred
						 * until it's a good moment to run it.
						 */
						int ret = cmdline_process_param("-P", p, 1, conf);
						assert(ret == 2);
					} else if (!strncmp(q, "telnet:", 7)) {
						/*
						 * If the hostname starts with "telnet:",
						 * set the protocol to Telnet and process
						 * the string as a Telnet URL.
						 */
						char c;

						q += 7;
						if (q[0] == '/' && q[1] == '/')
							q += 2;
						conf_set_int(conf, CONF_protocol, PROT_TELNET);
						p = q;
						p += host_strcspn(p, ":/");
						c = *p;
						if (*p)
							*p++ = '\0';
						if (c == ':')
							conf_set_int(conf, CONF_port, atoi(p));
						else
							conf_set_int(conf, CONF_port, -1);
						conf_set_str(conf, CONF_host, q);
						got_host = 1;
					} else {
						/*
						 * Otherwise, treat this argument as a host
						 * name.
						 */
						while (*p && !isspace(*p))
							p++;
						if (*p)
							*p++ = '\0';
						conf_set_str(conf, CONF_host, q);
						got_host = 1;
					}
				} else {
					cmdline_error("unknown option \"%s\"", p);
				}
			}
		}

		cmdline_run_saved(conf);

		if (loaded_session || got_host)
			allow_launch = TRUE;

		if (!allow_launch || !conf_launchable(conf)) {
#if defined(CONFIG_SINGLE_INSTANCE)
			LPWSTR class = ansi2unicode(appname);
			const LPWSTR caption = load_unicode_string(IDS_CFG_CAPTION);
			/** Do not single-instance PuTTY; we will launch multiple sessions to different hosts. */
			HWND hwnd = FindWindowW(class, NULL);
			if (!hwnd)
				hwnd = FindWindowW(NULL, caption);
			if (hwnd) {
				Sleep(1000);
				hwnd = FindWindowW(class, NULL);
				if (!hwnd)
					hwnd = FindWindowW(NULL, caption);
				if (hwnd) {
					SetForegroundWindow((HWND)((ULONG)hwnd | 0x00000001));
					cleanup_exit(0);
				}
			}
			sfree(class);
#endif
			if (!do_config())
				cleanup_exit(0);
		}

		/*
		 * Muck about with the hostname in various ways.
		 */
		{
		    char *hostbuf = dupstr(conf_get_str(conf, CONF_host));
		    char *host = hostbuf;
		    char *p, *q;

		    /*
		     * Trim leading whitespace.
		     */
		    host += strspn(host, " \t");

		    /*
		     * See if host is of the form user@host, and separate
		     * out the username if so.
		     */
		    if (host[0] != '\0') {
			char *atsign = strrchr(host, '@');
			if (atsign) {
			    *atsign = '\0';
			    conf_set_str(conf, CONF_username, host);
			    host = atsign + 1;
			}
		    }

		    /*
		     * Trim a colon suffix off the hostname if it's there. In
		     * order to protect unbracketed IPv6 address literals
		     * against this treatment, we do not do this if there's
		     * _more_ than one colon.
		     */
		    {
			char *c = host_strchr(host, ':');

			if (c) {
			    char *d = host_strchr(c+1, ':');
			    if (!d)
				*c = '\0';
			}
		    }

		    /*
		     * Remove any remaining whitespace.
		     */
		    p = hostbuf;
		    q = host;
		    while (*q) {
			if (*q != ' ' && *q != '\t')
			    *p++ = *q;
			q++;
		    }
		    *p = '\0';

		    conf_set_str(conf, CONF_host, hostbuf);
		    sfree(hostbuf);
		}
	} /* process cmdline */

	/* Check for invalid Port number (i.e. zero) */
	if (conf_get_int(conf, CONF_port) == 0) {
		char *cap = dupprintf(load_ansi_string(IDS_WARN_CAPTION_INTERNAL_ERROR), appname);
		MessageBoxA(NULL, "Invalid Port Number",
			    cap,
			    MB_OK | MB_ICONEXCLAMATION | MB_TOPMOST);
		sfree(cap);
		cleanup_exit(1);
	}

	if (!prev) {
		LPWSTR wappname = ansi2unicode(appname);
		wndclass.style = CS_HREDRAW | CS_VREDRAW;
		wndclass.lpfnWndProc = WndProc;
		wndclass.cbClsExtra = 0;
		wndclass.cbWndExtra = 0;
		wndclass.hInstance = inst;
		wndclass.hIcon = LoadIcon(inst, MAKEINTRESOURCE(IDI_MAINICON));
		wndclass.hCursor = LoadCursor(NULL, IDC_IBEAM);
		wndclass.hbrBackground	= (HBRUSH) GetStockObject(BLACK_BRUSH);
		wndclass.lpszMenuName = NULL;
		wndclass.lpszClassName = wappname;

		RegisterClassW(&wndclass);
		sfree(wappname);
	}

	memset(&ucsdata, 0, sizeof(ucsdata));

	conf_cache_data();

	conftopalette();

	/*
	 * Guess some defaults for the window size. This all gets
	 * updated later, so we don't really care too much. However, we
	 * do want the font width/height guesses to correspond to a
	 * large font rather than a small one...
	 */

	font_width = 10;
	font_height = 20;
	{
		LPWSTR wappname;
		DWORD_PTR winmode = WS_OVERLAPPED | WS_CLIPCHILDREN | WS_VSCROLL;
		DWORD_PTR exwinmode = 0;
		if (conf_get_int(conf, CONF_alwaysontop))
			exwinmode |= WS_EX_TOPMOST;
		if (conf_get_int(conf, CONF_sunken_edge))
			exwinmode |= WS_EX_CLIENTEDGE;
		wappname = ansi2unicode(appname);
		hwnd = CreateWindowExW(exwinmode, wappname, wappname,
				       winmode, CW_USEDEFAULT, CW_USEDEFAULT,
				       CW_USEDEFAULT, CW_USEDEFAULT,
				       NULL, NULL, inst, NULL);
		sfree(wappname);

		// On WinCE caret flickers awfully
		HideCaret(hwnd);

		if (!conf_get_int(conf, CONF_scrollbar))
			set_window_style(hwnd, 0, WS_VSCROLL);
	}

	/*
	 * Initialise the fonts, simultaneously correcting the guesses
	 * for font_{width,height}.
	 */
	init_fonts(0, 0);

	/*
	 * Initialise the terminal. (We have to do this _after_
	 * creating the window, since the terminal is the first thing
	 * which will call schedule_timer(), which will in turn call
	 * timer_change_notify() which will expect hwnd to exist.)
	 */
	term = term_init(conf, &ucsdata, NULL);
	logctx = log_init(NULL, conf);
	term_provide_logctx(term, logctx);
	term_size(term, conf_get_int(conf, CONF_height),
		  conf_get_int(conf, CONF_width),
		  conf_get_int(conf, CONF_savelines));

	/*
	 * Resize the window, now we know what size we _really_ want it
	 * to be.
	 */
	reset_window(0);

	/*
	 * Set up a caret bitmap, with no content.
	 */
	{
		char *bits;
		int size = (font_width + 15) / 16 * 2 * font_height;
		bits = snewn(size, char);
		memset(bits, 0, size);
		caretbm = CreateBitmap(font_width, font_height, 1, 1, bits);
		sfree(bits);
	}
	CreateCaret(hwnd, caretbm, font_width, font_height);

	/*
	 * Initialise the scroll bar.
	 */
	{
		SCROLLINFO si;

		si.cbSize = sizeof(si);
		si.fMask = SIF_ALL | SIF_DISABLENOSCROLL;
		si.nMin = 0;
		si.nMax = term->rows - 1;
		si.nPage = term->rows;
		si.nPos = 0;
		SetScrollInfo(hwnd, SB_VERT, &si, FALSE);
	}

	/*
	 * Prepare the mouse handler.
	 */
	lastact = MA_NOTHING;
	lastbtn = MBT_NOTHING;
	dbltime = GetDoubleClickTime();

	/*
	 * Set up the session-control options on the system menu.
	 */
	savedsess_menu = CreatePopupMenu();
	if (savedsess_menu) {
		get_sesslist(&sesslist, TRUE);
		update_savedsess_menu();
		get_sesslist(&sesslist, FALSE);
		InsertMenuW(popup_menus[1].menu, IDM_SAVEDSESS, MF_BYCOMMAND | MF_POPUP,
			    (UINT_PTR) savedsess_menu,
			    load_unicode_string(IDS_SAVEDSESS));
		DeleteMenu(popup_menus[1].menu, IDM_SAVEDSESS, MF_BYCOMMAND);
	}

	start_backend();

	/*
	 * Set up the initial input locale.
	 */
	set_input_locale(GetKeyboardLayout(0));

	/*
	 * Finally show the window!
	 */
	ShowWindow(hwnd, show);
	SetForegroundWindow(hwnd);

	/*
	 * Set the palette up.
	 */
	pal = NULL;
	logpal = NULL;
	init_palette();

	term_set_focus(term, GetForegroundWindow() == hwnd);

	UpdateWindow(hwnd);

#if 0
	{
		RECT rc;

		GetWindowRect(hwnd, &rc);
		SetWindowPos(hwnd, NULL, 0, 0, rc.right - rc.left, rc.bottom - rc.top,
			     SWP_NOMOVE | SWP_NOREDRAW | SWP_NOZORDER);
	}
#endif

	debug_level((DBG_WMSG, "\nMessage queue processing:\n"));

	while (1) {
		HANDLE *handles;
		HANDLE rhandles[ MAXIMUM_WAIT_OBJECTS ];
		int nhandles, n;
		DWORD timeout;

		if (toplevel_callback_pending() ||
			PeekMessage(&msg, NULL, 0, 0, PM_NOREMOVE)) {
			/*
			 * If we have anything we'd like to do immediately, set
			 * the timeout for MsgWaitForMultipleObjects to zero so
			 * that we'll only do a quick check of our handles and
			 * then get on with whatever that was.
			 *
			 * One such option is a pending toplevel callback. The
			 * other is a non-empty Windows message queue, which you'd
			 * think we could leave to MsgWaitForMultipleObjects to
			 * check for us along with all the handles, but in fact we
			 * can't because once PeekMessage in one iteration of this
			 * loop has removed a message from the queue, the whole
			 * queue is considered uninteresting by the next
			 * invocation of MWFMO. So we check ourselves whether the
			 * message queue is non-empty, and if so, set this timeout
			 * to zero to ensure MWFMO doesn't block.
			 */
			timeout = 0;
		} else {
			timeout = 1000;
			/* The messages seem unreliable; especially if we're being tricky */
			term_set_focus(term, got_focus);
		}

		handles = handle_get_events(&nhandles);

		memcpy(handles, rhandles, nhandles * sizeof( HANDLE ));

		rhandles[ nhandles ] = event;

		n = MsgWaitForMultipleObjects(nhandles + 1, rhandles, FALSE,
					      timeout, QS_ALLINPUT);

		if ((unsigned)(n - WAIT_OBJECT_0) < (unsigned)nhandles) {
			handle_got_event(handles[n - WAIT_OBJECT_0]);
			sfree(handles);
		} else
			sfree(handles);

		/*
		 * If the network event was set then let's process networ kdata.
		 */

		if (n == nhandles)
		{
			wince_dispatch_net_events(event);
		}

		while (PeekMessage(&msg, NULL, 0, 0, PM_NOREMOVE)) {
			const int f = GetMessage(&msg, NULL, 0, 0);
			if (msg.message == WM_QUIT || f == -1)
				goto finished;	/* two-level break */

			DispatchMessage(&msg);
		}

		/* The messages seem unreliable; especially if we're being tricky */
		term_set_focus(term, got_focus);

		if (pending_netevent)
			enact_pending_netevent();

		run_toplevel_callbacks();

		{
			const int minwait = 20, offwait = 100, idlewait = 300;
			Sleep(got_focus ? (session_closed ? offwait : minwait) : idlewait);
		}
	}

finished:
	if (event)
		p_WSACloseEvent( event );
	cleanup_exit(msg.wParam);	/* this doesn't return... */
	return msg.wParam;	/* ... but optimiser doesn't know */
}

/*
 * Clean up and exit.
 */
void cleanup_exit(int code)
{
	debug_level((DBG_FUNCTION, "cleanup_exit(%i);\n", code));

	/*
	 * Clean up.
	 */
	deinit_fonts();
	sfree(logpal);

	if (pal)
		DeleteObject(pal);

#ifdef DEBUG_UDP
	udplog_cleanup();
#endif
	sk_cleanup();
#ifdef SHELL_AYGSHELL
	aygshell_cleanup();
#endif

	if (conf_get_int(conf, CONF_protocol) == PROT_SSH) {
		random_save_seed();
#ifdef MSCRYPTOAPI
		crypto_wrapup();
#endif
	}
	shutdown_help();

	/* Clean up COM. */
	CoUninitialize();

	debug_level((DBG_FUNCTION, "done - cleanup_exit(%i);\n", code));

	conf_free(conf);
	conf = NULL;

	exit(code);
}

/*
 * Set up, or shut down, an AsyncSelect. Called from winnet.c.
 */
char *do_select(SOCKET skt, int startup)
{
	int i;
	int success;

	success = FALSE;

	debug_level((DBG_FUNCTION, "do_select();"));
	if (startup) {
		debug_level((DBG_FUNCTION, "do_select(startup);"));

		for (i = 0;
		     i < WINCE_MAX_SOCKETS;
		     i += 1)
		{
			if (wince_sockets[i].sock == INVALID_SOCKET)
			{
				wince_sockets[i].sock    = skt;
				wince_sockets[i].hwnd    = hwnd;
				wince_sockets[i].message = WM_NETEVENT;

				if (p_WSAEventSelect(wince_sockets[i].sock,
						     wince_net_event,
						     FD_CONNECT |
						     FD_READ    |
						     FD_WRITE   |
						     FD_OOB     |
						     FD_CLOSE   | FD_ACCEPT
						     ) == SOCKET_ERROR)
				{
					fatalbox("WSAEventSelect fails - %lu", p_WSAGetLastError());
					break;
				}

				success                    = TRUE;

				break;
			}
		}

		/*
		 * TODO: Add code to resize the array here if we wanted to.
		 */
		if (!success)
			fatalbox("Too many active sockets.");

	} else {
		debug_level((DBG_FUNCTION, "do_select(shutdown);"));

		if (skt == INVALID_SOCKET)
			fatalbox("Attempted to unregister invalid socket.");

		for (i = 0;
		     i < WINCE_MAX_SOCKETS;
		     i += 1)
		{
			if (wince_sockets[i].sock == skt)
			{
				wince_sockets[i].sock    = INVALID_SOCKET;
				wince_sockets[i].hwnd    = 0;
				wince_sockets[i].message = WM_NULL;

				success                  = TRUE;

				break;
			}
		}

		if (!success)
			fatalbox("Attempted to unregister unknown socket.");
	}

	debug_level((DBG_FUNCTION, "do_select - end;\n"));
	return NULL;
}

/*
 * Refresh the saved-session submenu from `sesslist'.
 */
static void update_savedsess_menu(void)
{
	int i;

	debug_level((DBG_FUNCTION, "update_savedsess_menu();\n"));

	InsertMenu(savedsess_menu, -1, MF_BYPOSITION, 0, TEXT(""));
	while (DeleteMenu(savedsess_menu, 0, MF_BYPOSITION)) ;

	/* skip sesslist.sessions[0] == Default Settings */
	for (i = 1;
		i < ((sesslist.nsessions <= MENU_SAVED_MAX+1) ? sesslist.nsessions
							      : MENU_SAVED_MAX+1);
		i++) {

		WCHAR *str = ansi2unicode(sesslist.sessions[i]);
		if (wcslen(str) >= MENU_SAVEDSESS_LENGTH)
			str[MENU_SAVEDSESS_LENGTH - 1] = L'\0';
		InsertMenuW(savedsess_menu, -1, MF_BYPOSITION,
			    IDM_SAVED_MIN + (i-1)*MENU_SAVED_STEP,
			    str);
		sfree(str);
	}

	if (sesslist.nsessions <= 1)
		InsertMenu(savedsess_menu, -1, MF_BYPOSITION | MF_GRAYED, IDM_SAVED_MIN, TEXT("(No sessions)"));
}

/*
 * Update the Special Commands submenu.
 */
void update_specials_menu(void *frontend)
{
	HMENU new_menu;
	int i, j;
	wchar_t *s;
	debug_level((DBG_FUNCTION, "update_specials_menu();\n"));

	if (back)
		specials = back->get_specials(backhandle);
	else
		specials = NULL;

	if (specials) {
		/* We can't use Windows to provide a stack for submenus, so
		 * here's a lame "stack" that will do for now. */
		HMENU saved_menu = NULL;
		int nesting = 1;
		new_menu = CreatePopupMenu();
		for (i = 0; nesting > 0; i++) {
			assert(IDM_SPECIAL_MIN + 0x10 * i < IDM_SPECIAL_MAX);
			switch (specials[i].code) {
			case TS_SEP:
				AppendMenu(new_menu, MF_SEPARATOR, 0, 0);
				break;
			case TS_SUBMENU:
				assert(nesting < 2);
				nesting++;
				saved_menu = new_menu; /* XXX lame stacking */
				new_menu = CreatePopupMenu();
				s = ansi2unicode(specials[i].name);
				AppendMenuW(saved_menu, MF_POPUP | MF_ENABLED,
					    (UINT_PTR) new_menu, s);
				sfree(s);
				break;
			case TS_EXITMENU:
				nesting--;
				if (nesting) {
					new_menu = saved_menu; /* XXX lame stacking */
					saved_menu = NULL;
				}
				break;
			default:
				s = ansi2unicode(specials[i].name);
				AppendMenuW(new_menu, MF_ENABLED,
					    IDM_SPECIAL_MIN + 0x10 * i, s);
				sfree(s);
				break;
			}
		}
		/* Squirrel the highest special. */
		n_specials = i - 1;
	} else {
		new_menu = NULL;
		n_specials = 0;
	}

	j = 0;
	{
		if (specials_menu) {
			DeleteMenu(popup_menus[j].menu, 0, MF_BYPOSITION);
			DeleteMenu(popup_menus[j].menu, IDM_SPECIALSEP, MF_BYCOMMAND);
		}
		if (new_menu) {
			InsertMenu(popup_menus[j].menu, 0,
				   MF_BYPOSITION | MF_SEPARATOR, IDM_SPECIALSEP, 0);
			InsertMenuW(popup_menus[j].menu, 0,
				    MF_BYPOSITION | MF_POPUP | MF_ENABLED,
				    (UINT_PTR) new_menu, load_unicode_string(IDS_SPECIALCOMMAND));
		}
	}
	specials_menu = new_menu;
}

static void update_mouse_pointer(void)
{
	LPTSTR curstype;
	int force_visible = FALSE;
	static int forced_visible = FALSE;

	debug_level((DBG_FUNCTION, "update_mouse_pointer();\n"));

	switch (busy_status) {
	case BUSY_NOT:
		if (send_raw_mouse)
			curstype = IDC_ARROW;
		else
			curstype = IDC_IBEAM;
		break;
	case BUSY_WAITING:
		curstype = IDC_APPSTARTING; /* this may be an abuse */
		force_visible = TRUE;
		break;
	case BUSY_CPU:
		curstype = IDC_WAIT;
		force_visible = TRUE;
		break;
	default:
		assert(0);
		curstype = IDC_ARROW;
	}
	{
		HCURSOR cursor = LoadCursor(NULL, curstype);
		SetClassLongPtr(hwnd, GCLP_HCURSOR, (LONG_PTR)cursor);
		SetCursor(cursor); /* force redraw of cursor at current posn */
	}
	if (force_visible != forced_visible) {
		/* We want some cursor shapes to be visible always.
		 * Along with show_mouseptr(), this manages the ShowCursor()
		 * counter such that if we switch back to a non-force_visible
		 * cursor, the previous visibility state is restored. */
		ShowCursor(force_visible);
		forced_visible = force_visible;
	}
}

void set_busy_status(void *frontend, int status)
{
    busy_status = status;
    update_mouse_pointer();
}

/*
 * set or clear the "raw mouse message" mode
 */
void set_raw_mouse_mode(void *frontend, int activate)
{
	debug_level((DBG_FUNCTION, "set_raw_mouse_mode(%i);\n", activate));

	activate = activate && !conf_get_int(conf, CONF_no_mouse_rep);
	send_raw_mouse = activate;
	update_mouse_pointer();
}

/*
 * Print a message box and close the connection.
 */
void connection_fatal(void *frontend, char *fmt, ...)
{
	va_list ap;
	char *stuff;
	char *cap = dupprintf(load_ansi_string(IDS_WARN_CAPTION_FALAL_ERROR), appname);

	va_start(ap, fmt);
	stuff = dupvprintf(fmt, ap);
	va_end(ap);

	MessageBoxA(hwnd, stuff, cap, MB_ICONERROR | MB_OK);
	sfree(stuff);
	sfree(cap);

	if (conf_get_int(conf, CONF_close_on_exit) == FORCE_ON)
		PostQuitMessage(1);
	else {
		queue_toplevel_callback(close_session, NULL);
	}
}

/*
 * Report an error at the command-line parsing stage.
 */
void cmdline_error(char *fmt, ...)
{
	va_list ap;
	char *stuff;
	char *cap = dupprintf(load_ansi_string(IDS_WARN_CAPTION_COMMANDLINE_ERROR), appname);

	va_start(ap, fmt);
	stuff = dupvprintf(fmt, ap);
	va_end(ap);

	MessageBoxA(hwnd, stuff, cap, MB_ICONERROR | MB_OK);

	sfree(stuff);
	sfree(cap);
	exit(1);
}

/*
 * Actually do the job requested by a WM_NETEVENT
 */
static void wm_netevent_callback(void *vctx)
{
    struct wm_netevent_params *params = (struct wm_netevent_params *)vctx;
    select_result(params->wParam, params->lParam);
    sfree(vctx);
}

/*
 * Copy the colour palette from the configuration data into defpal.
 * This is non-trivial because the colour indices are different.
 */
static void conftopalette(void)
{
	int i;
	static const int ww[] = {
		256, 257, 258, 259, 260, 261,
		0, 8, 1, 9, 2, 10, 3, 11,
		4, 12, 5, 13, 6, 14, 7, 15
	};

	for (i = 0; i < 22; i++) {
		int w = ww[i];
		defpal[w].rgbtRed = conf_get_int_int(conf, CONF_colours, i*3+0);
		defpal[w].rgbtGreen = conf_get_int_int(conf, CONF_colours, i*3+1);
		defpal[w].rgbtBlue = conf_get_int_int(conf, CONF_colours, i*3+2);
	}
	for (i = 0; i < NEXTCOLOURS; i++) {
		if (i < 216) {
			int r = i / 36, g = (i / 6) % 6, b = i % 6;
			defpal[i+16].rgbtRed = r ? r * 40 + 55 : 0;
			defpal[i+16].rgbtGreen = g ? g * 40 + 55 : 0;
			defpal[i+16].rgbtBlue = b ? b * 40 + 55 : 0;
		} else {
			int shade = i - 216;
			shade = shade * 10 + 8;
			defpal[i+16].rgbtRed = defpal[i+16].rgbtGreen =
			defpal[i+16].rgbtBlue = shade;
		}
	}

	/* Override with system colours if appropriate */
	if (conf_get_int(conf, CONF_system_colour))
		systopalette();
}

/*
* Override bit of defpal with colours from the system.
* (NB that this takes a copy the system colours at the time this is called,
* so subsequent colour scheme changes don't take effect. To fix that we'd
* probably want to be using GetSysColorBrush() and the like.)
*/
static void systopalette(void)
{
	int i;
	static const struct { int nIndex; int norm; int bold; } or[] =
	{
		{ COLOR_WINDOWTEXT,	256, 257 }, /* Default Foreground */
		{ COLOR_WINDOW,		258, 259 }, /* Default Background */
		{ COLOR_HIGHLIGHTTEXT,	260, 260 }, /* Cursor Text */
		{ COLOR_HIGHLIGHT,	261, 261 }, /* Cursor Colour */
	};

	for (i = 0; i < (sizeof(or)/sizeof(or[0])); i++) {
		COLORREF colour = GetSysColor(or[i].nIndex);
		defpal[or[i].norm].rgbtRed =
			defpal[or[i].bold].rgbtRed = GetRValue(colour);
		defpal[or[i].norm].rgbtGreen =
			defpal[or[i].bold].rgbtGreen = GetGValue(colour);
		defpal[or[i].norm].rgbtBlue =
			defpal[or[i].bold].rgbtBlue = GetBValue(colour);
	}
}

/*
 * Set up the colour palette.
 */
static void init_palette(void)
{
	int i;
	HDC hdc = GetDC(hwnd);

	debug_level((DBG_FUNCTION, "init_palette();\n"));

	if (hdc) {
		if (conf_get_int(conf, CONF_try_palette) &&
			GetDeviceCaps(hdc, RASTERCAPS) & RC_PALETTE) {
			/*
			 * This is a genuine case where we must use smalloc
			 * because the snew macros can't cope.
			 */
			logpal = (LPLOGPALETTE) smalloc(sizeof(*logpal)
							- sizeof(logpal->palPalEntry)
							+ NALLCOLOURS * sizeof(PALETTEENTRY));
			logpal->palVersion = 0x300;
			logpal->palNumEntries = NALLCOLOURS;
			for (i = 0; i < NALLCOLOURS; i++) {
				logpal->palPalEntry[i].peRed = defpal[i].rgbtRed;
				logpal->palPalEntry[i].peGreen = defpal[i].rgbtGreen;
				logpal->palPalEntry[i].peBlue = defpal[i].rgbtBlue;
				logpal->palPalEntry[i].peFlags = PC_NOCOLLAPSE;
			}
			pal = CreatePalette(logpal);
			if (pal) {
				SelectPalette(hdc, (HPALETTE) pal, FALSE);
				RealizePalette(hdc);
				SelectPalette(hdc, (HPALETTE) GetStockObject(DEFAULT_PALETTE), FALSE);
			}
		}
		ReleaseDC(hwnd, hdc);
	}
	if (pal) {
		for (i = 0; i < NALLCOLOURS; i++) {
			colours[i] = PALETTERGB(defpal[i].rgbtRed,
			defpal[i].rgbtGreen,
			defpal[i].rgbtBlue);
		}
	} else {
		for (i = 0; i < NALLCOLOURS; i++) {
			colours[i] = RGB(defpal[i].rgbtRed,
			defpal[i].rgbtGreen, defpal[i].rgbtBlue);
		}
	}
}

/*
 * This is a wrapper to ExtTextOut() to force Windows to display
 * the precise glyphs we give it. Otherwise it would do its own
 * bidi and Arabic shaping, and we would end up uncertain which
 * characters it had put where.
 */
static void exact_textout(HDC hdc, int x, int y, CONST RECT *lprc,
			  unsigned short *lpString, UINT cbCount,
			  CONST INT *lpDx, int opaque)
{
	wchar_t *buffer = snewn(cbCount*2+2, wchar_t);
	memset(buffer, 0, (cbCount*2+2)*sizeof(wchar_t));
	ExtTextOut(hdc, x, y,
		   ETO_CLIPPED | (opaque ? ETO_OPAQUE : 0),
		   lprc, buffer, cbCount, lpDx);
}

/*
 * The exact_textout() wrapper, unfortunately, destroys the useful
 * Windows `font linking' behaviour: automatic handling of Unicode
 * code points not supported in this font by falling back to a font
 * which does contain them. Therefore, we adopt a multi-layered
 * approach: for any potentially-bidi text, we use exact_textout(),
 * and for everything else we use a simple ExtTextOut as we did
 * before exact_textout() was introduced.
 */
static void general_textout(HDC hdc, int x, int y, CONST RECT *lprc,
			    unsigned short *lpString, UINT cbCount,
			    CONST INT *lpDx, int opaque)
{
	int i, j, xp, xn;
	RECT newrc;

	xp = xn = x;

	for (i = 0; i < (int)cbCount ;) {
		int rtl = is_rtl(lpString[i]);

		xn += lpDx[i];

		for (j = i+1; j < (int)cbCount; j++) {
			if (rtl != is_rtl(lpString[j]))
				break;
			xn += lpDx[j];
		}

		/*
		 * Now [i,j) indicates a maximal substring of lpString
		 * which should be displayed using the same textout
		 * function.
		 */
		if (rtl) {
			newrc.left = lprc->left + xp - x;
			newrc.right = lprc->left + xn - x;
			newrc.top = lprc->top;
			newrc.bottom = lprc->bottom;
			exact_textout(hdc, xp, y, &newrc, lpString+i, j-i, lpDx+i, opaque);
		} else {
			newrc.left = lprc->left + xp - x;
			newrc.right = lprc->left + xn - x;
			newrc.top = lprc->top;
			newrc.bottom = lprc->bottom;
			ExtTextOutW(hdc, xp, y, ETO_CLIPPED | (opaque ? ETO_OPAQUE : 0),
				    &newrc, lpString+i, j-i, lpDx+i);
		}

		i = j;
		xp = xn;
	}

	assert(xn - x == lprc->right - lprc->left);
}

static int get_font_width(HDC hdc, const TEXTMETRIC *tm)
{
	return tm->tmAveCharWidth;
}

/*
 * Initialise all the fonts we will need initially. There may be as many as
 * three or as few as one.  The other (potentially) twenty-one fonts are done
 * if/when they are needed.
 *
 * We also:
 *
 * - check the font width and height, correcting our guesses if
 *   necessary.
 *
 * - verify that the bold font is the same width as the ordinary
 *   one, and engage shadow bolding if not.
 *
 * - verify that the underlined font is the same width as the
 *   ordinary one (manual underlining by means of line drawing can
 *   be done in a pinch).
 */
static void init_fonts(int pick_width, int pick_height)
{
	TEXTMETRIC tm;
	CPINFO cpinfo;
	FontSpec *font;
	int fontsize[3];
	int i;
	int quality;
	HDC hdc;
	int fw_dontcare, fw_bold;

	debug_level((DBG_FUNCTION, "init_fonts(w: %i h: %i);\n", pick_width, pick_height));

	for (i = 0; i < FONT_MAXNO; i++)
		fonts[i] = NULL;

	bold_font_mode = conf_get_int(conf, CONF_bold_style) & 1 ?
			 BOLD_FONT : BOLD_NONE;
	bold_colours = conf_get_int(conf, CONF_bold_style) & 2 ? TRUE : FALSE;
	und_mode = UND_FONT;

	font = conf_get_fontspec(conf, CONF_font);
	if (font->isbold) {
		fw_dontcare = FW_BOLD;
		fw_bold = FW_HEAVY;
	} else {
		fw_dontcare = FW_DONTCARE;
		fw_bold = FW_BOLD;
	}

	hdc = GetDC(hwnd);

	if (pick_height)
		font_height = pick_height;
	else {
		font_height = font->height;
		if (font_height > 0) {
			font_height =
				-MulDiv(font_height, GetDeviceCaps(hdc, LOGPIXELSY), 72);
		}
	}
	font_width = pick_width;

	quality = conf_get_int(conf, CONF_font_quality);
#define f(i,c,w,u) \
	fonts[i] = CreateFontA(font_height, font_width, 0, 0, w, FALSE, u, FALSE, \
			       c, OUT_DEFAULT_PRECIS, \
			       CLIP_DEFAULT_PRECIS, FONT_QUALITY(quality), \
			       FIXED_PITCH | FF_DONTCARE, font->name)

	f(FONT_NORMAL, font->charset, fw_dontcare, FALSE);

	SelectObject(hdc, fonts[FONT_NORMAL]);
	GetTextMetrics(hdc, &tm);

	GetObject(fonts[FONT_NORMAL], sizeof(LOGFONT), &lfont);

	/* Note that the TMPF_FIXED_PITCH bit is defined upside down :-( */
	if (!(tm.tmPitchAndFamily & TMPF_FIXED_PITCH)) {
		font_varpitch = FALSE;
		font_dualwidth = (tm.tmAveCharWidth != tm.tmMaxCharWidth);
	} else {
		font_varpitch = TRUE;
		font_dualwidth = TRUE;
	}
	if (pick_width == 0 || pick_height == 0) {
		font_height = tm.tmHeight;
		font_width = get_font_width(hdc, &tm);
	}

#ifdef RDB_DEBUG_PATCH
	debug(23, "Primary font H=%d, AW=%d, MW=%d",
	tm.tmHeight, tm.tmAveCharWidth, tm.tmMaxCharWidth);
#endif
	debug_level((DBG_USER, "Primary font %s, charset: %i, H=%d, AW=%d, MW=%d",
	font->name, font->charset, tm.tmHeight, tm.tmAveCharWidth, tm.tmMaxCharWidth));
	{
		CHARSETINFO info;
		DWORD cset = tm.tmCharSet;
		memset(&info, 0xFF, sizeof(info));

		/* !!! Yes the next line is right */
		if (cset == OEM_CHARSET)
			ucsdata.font_codepage = GetOEMCP();
		else
		if (TranslateCharsetInfo ((DWORD *) cset, &info, TCI_SRCCHARSET))
			ucsdata.font_codepage = info.ciACP;
		else
			ucsdata.font_codepage = -1;

		GetCPInfo(ucsdata.font_codepage, &cpinfo);
		ucsdata.dbcs_screenfont = (cpinfo.MaxCharSize > 1);
	}

	f(FONT_UNDERLINE, font->charset, fw_dontcare, TRUE);

	/*
	 * Some fonts, e.g. 9-pt Courier, draw their underlines
	 * outside their character cell. We successfully prevent
	 * screen corruption by clipping the text output, but then
	 * we lose the underline completely. Here we try to work
	 * out whether this is such a font, and if it is, we set a
	 * flag that causes underlines to be drawn by hand.
	 *
	 * Having tried other more sophisticated approaches (such
	 * as examining the TEXTMETRIC structure or requesting the
	 * height of a string), I think we'll do this the brute
	 * force way: we create a small bitmap, draw an underlined
	 * space on it, and test to see whether any pixels are
	 * foreground-coloured. (Since we expect the underline to
	 * go all the way across the character cell, we only search
	 * down a single column of the bitmap, half way across.)
	 */
	{
		HDC und_dc;
		HBITMAP und_bm, und_oldbm;
		int i, gotit;
		COLORREF c;

		und_dc = CreateCompatibleDC(hdc);
		und_bm = CreateCompatibleBitmap(hdc, font_width, font_height);
		und_oldbm = (HBITMAP) SelectObject(und_dc, und_bm);
		SelectObject(und_dc, fonts[FONT_UNDERLINE]);
#if !(UNDER_CE < 0x400)
		SetTextAlign(und_dc, TA_TOP | TA_LEFT | TA_NOUPDATECP);
#endif
		SetTextColor(und_dc, RGB(255, 255, 255));
		SetBkColor(und_dc, RGB(0, 0, 0));
		SetBkMode(und_dc, OPAQUE);
		ExtTextOut(und_dc, 0, 0, ETO_OPAQUE, NULL, TEXT(" "), 1, NULL);
		gotit = FALSE;
		for (i = 0; i < font_height; i++) {
			c = GetPixel(und_dc, font_width / 2, i);
			if (c != RGB(0, 0, 0))
				gotit = TRUE;
		}
		SelectObject(und_dc, und_oldbm);
		DeleteObject(und_bm);
		DeleteDC(und_dc);
		if (!gotit) {
			und_mode = UND_LINE;
			DeleteObject(fonts[FONT_UNDERLINE]);
			fonts[FONT_UNDERLINE] = 0;
		}
	}

	if (bold_font_mode == BOLD_FONT) {
		f(FONT_BOLD, font->charset, fw_bold, FALSE);
	}
#undef f

	descent = tm.tmAscent + 1;
	if (descent >= font_height)
		descent = font_height - 1;

	for (i = 0; i < 3; i++) {
		if (fonts[i]) {
			if (SelectObject(hdc, fonts[i]) && GetTextMetrics(hdc, &tm))
				fontsize[i] = get_font_width(hdc, &tm) + 256 * tm.tmHeight;
			else
				fontsize[i] = -i;
		} else
			fontsize[i] = -i;
	}

	ReleaseDC(hwnd, hdc);

	if (fontsize[FONT_UNDERLINE] != fontsize[FONT_NORMAL]) {
		und_mode = UND_LINE;
		DeleteObject(fonts[FONT_UNDERLINE]);
		fonts[FONT_UNDERLINE] = 0;
	}

	if (bold_font_mode == BOLD_FONT &&
	    fontsize[FONT_BOLD] != fontsize[FONT_NORMAL]) {
		bold_font_mode = BOLD_SHADOW;
		DeleteObject(fonts[FONT_BOLD]);
		fonts[FONT_BOLD] = 0;
	}
	fontflag[0] = fontflag[1] = fontflag[2] = 1;

	init_ucs(conf, &ucsdata);
}

static void another_font(int fontno)
{
	int basefont;
	int fw_dontcare, fw_bold, quality;
	int c, u, w, x;
	char *s;
	FontSpec *font;

	if (fontno < 0 || fontno >= FONT_MAXNO || fontflag[fontno])
		return;

	basefont = (fontno & ~(FONT_BOLDUND));
	if (basefont != fontno && !fontflag[basefont])
		another_font(basefont);

	font = conf_get_fontspec(conf, CONF_font);

	if (font->isbold) {
		fw_dontcare = FW_BOLD;
		fw_bold = FW_HEAVY;

	} else {
		fw_dontcare = FW_DONTCARE;
		fw_bold = FW_BOLD;
	}

	c = font->charset;
	w = fw_dontcare;
	u = FALSE;
	s = font->name;
	x = font_width;

	if (fontno & FONT_WIDE)
		x *= 2;
	if (fontno & FONT_NARROW)
		x = (x+1)/2;
	if (fontno & FONT_OEM)
		c = OEM_CHARSET;
	if (fontno & FONT_BOLD)
		w = fw_bold;
	if (fontno & FONT_UNDERLINE)
		u = TRUE;

	quality = conf_get_int(conf, CONF_font_quality);

	fonts[fontno] =
		CreateFontA(font_height * (1 + !!(fontno & FONT_HIGH)), x, 0, 0, w,
			    FALSE, u, FALSE, c, OUT_DEFAULT_PRECIS,
			    CLIP_DEFAULT_PRECIS, FONT_QUALITY(quality),
			    FIXED_PITCH | FF_DONTCARE, s);

	fontflag[fontno] = 1;
}

static void deinit_fonts(void)
{
	int i;

	debug_level((DBG_FUNCTION, "deinit_fonts();\n"));

	for (i = 0; i < FONT_MAXNO; i++) {
		if (fonts[i])
			DeleteObject(fonts[i]);
		fonts[i] = 0;
		fontflag[i] = 0;
	}
}

void request_resize(void *frontend, int w, int h)
{
	int width, height;

	debug_level((DBG_FUNCTION, "request_resize(w: %i h: %i);\n", w, h));

	/* If the window is maximized supress resizing attempts */
	if (IsZoomed(hwnd)) {
		if (conf_get_int(conf, CONF_resize_action) == RESIZE_TERM)
			return;
	}

	if (conf_get_int(conf, CONF_resize_action) == RESIZE_DISABLED) return;
	if (h == term->rows && w == term->cols) return;

	/* Sanity checks ... */
	{
		static int first_time = 1;
		static RECT ss;

		switch (first_time) {
		case 1:
			/* Get the size of the screen */
			if (get_fullscreen_rect(&ss))
				/* first_time = 0 */ ;
			else {
				first_time = 2;
				break;
			}
		case 0:
			/* Make sure the values are sane */
			width = (ss.right - ss.left) / 4;
			height = (ss.bottom - ss.top) / 6;

			if (w > width || h > height)
				return;
			if (w < 15)
				w = 15;
			if (h < 1)
				h = 1;
		}
	}

	term_size(term, h, w, conf_get_int(conf, CONF_savelines));

	if (conf_get_int(conf, CONF_resize_action) != RESIZE_FONT) {
	} else
		reset_window(0);

	InvalidateRect(hwnd, NULL, TRUE);
}

static void reset_window(int reinit)
{
	/*
	 * This function decides how to resize or redraw when the
	 * user changes something.
	 *
	 * This function doesn't like to change the terminal size but if the
	 * font size is locked that may be it's only soluion.
	 */
	int win_width, win_height, resize_action, window_border;
	RECT cr, wr;

	if(!term)
		return;

	debug_level((DBG_FUNCTION, "reset_window(%i);\n", reinit));

#ifdef RDB_DEBUG_PATCH
	debug_level((27, "reset_window()"));
#endif

	/* Current window sizes ... */
	GetWindowRect(hwnd, &wr);
	GetClientRect(hwnd, &cr);
	wr.top += override_menu_height;
	cr.bottom -= override_menu_height;

	win_width  = cr.right - cr.left;
	win_height = cr.bottom - cr.top;

	resize_action = conf_get_int(conf, CONF_resize_action);
	window_border = conf_get_int(conf, CONF_window_border);

	if (resize_action == RESIZE_DISABLED)
		reinit = 2;

	/* Are we being forced to reload the fonts ? */
	if (reinit>1) {
#ifdef RDB_DEBUG_PATCH
		debug_level((27, "reset_window() -- Forced deinit"));
#endif
		deinit_fonts();
		init_fonts(0, 0);
	}

	/* Oh, looks like we're minimised */
	if (win_width == 0 || win_height == 0)
		return;

	/* Is the window out of position ? */
	if ( !reinit &&
	    (offset_width != (win_width-font_width*term->cols)/2 ||
	     offset_height != (win_height-font_height*term->rows)/2) ){
		offset_width = (win_width-font_width*term->cols)/2;
		offset_height = (win_height-font_height*term->rows)/2;
		if (offset_height < override_menu_height)
			offset_height = override_menu_height;
		InvalidateRect(hwnd, NULL, TRUE);
#ifdef RDB_DEBUG_PATCH
		debug_level((27, "reset_window() -> Reposition terminal"));
#endif
	}

	/* We're fullscreen, this means we must not change the size of
	 * the window so it's the font size or the terminal itself.
	 */

	if (resize_action == RESIZE_FONT ||
	    (resize_action == RESIZE_EITHER && is_full_screen()))
	{
		if (resize_action == RESIZE_EITHER) {
			term_size(term, conf_get_int(conf, CONF_height),
				  conf_get_int(conf, CONF_width),
				  conf_get_int(conf, CONF_savelines));
		}
		if (font_width != win_width/term->cols ||
		    font_height != win_height/term->rows) {
			deinit_fonts();
			init_fonts(win_width/term->cols, win_height/term->rows);
			offset_width = (win_width-font_width*term->cols)/2;
			offset_height = (win_height-font_height*term->rows)/2;
			if (offset_height < override_menu_height)
				offset_height = override_menu_height;
			InvalidateRect(hwnd, NULL, TRUE);
#ifdef RDB_DEBUG_PATCH
			debug_level((25, "reset_window() -> Z font resize to (%d, %d)",
				    font_width, font_height));
#endif
		}
	}
	else if (resize_action == RESIZE_TERM ||
		 resize_action == RESIZE_EITHER)
	{
		if (resize_action == RESIZE_EITHER) {
			deinit_fonts();
			init_fonts(0, 0);
		}
		if (font_width * term->cols != win_width ||
		    font_height * term->rows != win_height) {
			/* Our only choice at this point is to change the
			 * size of the terminal; Oh well.
			 */
			term_size(term, win_height/font_height, win_width/font_width,
				  conf_get_int(conf, CONF_savelines));
			offset_width = (win_width-font_width*term->cols)/2;
			offset_height = (win_height-font_height*term->rows)/2;
			if (offset_height < override_menu_height)
				offset_height = override_menu_height;
			InvalidateRect(hwnd, NULL, TRUE);
#ifdef RDB_DEBUG_PATCH
			debug_level((27, "reset_window() -> Zoomed term_size"));
#endif
		}
	}

	return;
}

static void set_input_locale(HKL kl)
{
	wchar_t lbuf[20];

	debug_level((DBG_FUNCTION, "set_input_locale();\n"));
	GetLocaleInfo(LOWORD(kl), LOCALE_IDEFAULTANSICODEPAGE, lbuf, sizeof(lbuf));
	kbd_codepage = _wtoi(lbuf);
}

int agent_query(void *in, int inlen, void **out, int *outlen,
		void (*callback)(void *, void *, int), void *callback_ctx)
{
	debug_level((DBG_FUNCTION, "agent_query();\n"));
	return 1; /* failure */
}

static void click(Mouse_Button b, int x, int y, int shift, int ctrl, int alt)
{
	int thistime = GetTickCount();

	if (send_raw_mouse && !(conf_get_int(conf, CONF_mouse_override) && shift)) {
		lastbtn = MBT_NOTHING;
		term_mouse(term, b, translate_button(b), MA_CLICK,
			   x, y, shift, ctrl, alt);
		return;
	}

	if (lastbtn == b && thistime - lasttime < dbltime) {
		lastact = (lastact == MA_CLICK ? MA_2CLK :
			   lastact == MA_2CLK ? MA_3CLK :
			   lastact == MA_3CLK ? MA_CLICK : MA_NOTHING);
	} else {
		lastbtn = b;
		lastact = MA_CLICK;
	}
	if (lastact != MA_NOTHING)
		term_mouse(term, b, translate_button(b), lastact,
			   x, y, shift, ctrl, alt);
	lasttime = thistime;
}

/*
 * Translate a raw mouse button designation (LEFT, MIDDLE, RIGHT)
 * into a cooked one (SELECT, EXTEND, PASTE).
 */
static Mouse_Button translate_button(Mouse_Button button)
{
	if (button == MBT_LEFT)
		return MBT_SELECT;
	if (button == MBT_MIDDLE)
		return conf_get_int(conf, CONF_mouse_is_xterm) == 1 ?
			MBT_PASTE : MBT_EXTEND;
	if (button == MBT_RIGHT)
		return conf_get_int(conf, CONF_mouse_is_xterm) == 1 ?
			MBT_EXTEND : MBT_PASTE;
	return 0;			       /* shouldn't happen */
}

static void show_mouseptr(int show)
{
	/* NB that the counter in ShowCursor() is also frobbed by
	* update_mouse_pointer() */
	static int cursor_visible = 1;

	debug_level((DBG_FUNCTION, "show_mouseptr(%i);\n", show));

	if (!conf_get_int(conf, CONF_hide_mouseptr))
		show = 1;			/* override if this feature disabled */
	if (cursor_visible && !show)
		ShowCursor(FALSE);
	else if (!cursor_visible && show)
		ShowCursor(TRUE);
	cursor_visible = show;
}

static int is_alt_pressed(void)
{
	BYTE keystate[256];
	int r = GetKeyboardState(keystate);
	if (!r)
		return FALSE;
	if (keystate[VK_MENU] & 0x80)
		return TRUE;
	if (keystate[VK_RMENU] & 0x80)
		return TRUE;
	return FALSE;
}

static int resizing;

void notify_remote_exit(void *fe)
{
	int exitcode, close_on_exit;

	debug_level((DBG_FUNCTION, "notify_remote_exit();\n"));

	if (!session_closed &&
		(exitcode = back->exitcode(backhandle)) >= 0) {
		close_on_exit = conf_get_int(conf, CONF_close_on_exit);
		/* Abnormal exits will already have set session_closed and taken
		 * appropriate action. */
		if (close_on_exit == FORCE_ON ||
		    (close_on_exit == AUTO && exitcode != INT_MAX)) {
			PostQuitMessage(0);
		} else {
			queue_toplevel_callback(close_session, NULL);
			session_closed = TRUE;
			/* exitcode == INT_MAX indicates that the connection was closed
			 * by a fatal error, so an error box will be coming our way and
			 * we should not generate this informational one. */
			if (exitcode != INT_MAX)
				MessageBoxA(hwnd, load_ansi_string(IDS_WARN_CLOSED_BY_REMOTEHOST), 
					    appname, MB_OK | MB_ICONINFORMATION);
		}
	}
}

void timer_change_notify(unsigned long next)
{
	unsigned long now = GETTICKCOUNT();
	long ticks;

	debug_level((DBG_FUNCTION, "timer_change_notify(%lu);\n", next));

	if (now - next < INT_MAX)
		ticks = 0;
	else
		ticks = next - now;
	KillTimer(hwnd, TIMING_TIMER_ID);
	SetTimer(hwnd, TIMING_TIMER_ID, ticks, NULL);
	timing_next_time = next;
}

static void conf_cache_data(void)
{
	/* Cache some items from conf to speed lookups in very hot code */
	cursor_type = conf_get_int(conf, CONF_cursor_type);
	vtmode = conf_get_int(conf, CONF_vtmode);
}

static LRESULT CALLBACK WndProc(HWND hwnd, UINT message,
				WPARAM wParam, LPARAM lParam)
{
    HDC hdc;
    static int ignore_clip = FALSE;
    static int need_backend_resize = FALSE;
    static int fullscr_on_max = FALSE;
    static UINT last_mousemove = 0;

    debug_level((DBG_WMSG, "\nWndProc Message %d(%x, %x)\n", message, wParam, lParam));

    switch (message) {
      case WM_TIMER:
	debug_level((DBG_WMSG, "WM_TIMER\n"));
	if ((UINT_PTR)wParam == TIMING_TIMER_ID) {
	    unsigned long next;

	    KillTimer(hwnd, TIMING_TIMER_ID);
	    if (run_timers(timing_next_time, &next)) {
		timer_change_notify(next);
	    }
	}
	return 0;
      case WM_CREATE:
	debug_level((DBG_WMSG, "WM_CREATE\n"));
	{
		HIMC himc = ImmGetContext(hwnd);
		if (himc) {
			ImmSetOpenStatus(himc, FALSE);
			ImmReleaseContext(hwnd, himc);
		}

#ifdef SHELL_AYGSHELL
		memset(&sai, 0, sizeof(sai));
		sai.cbSize = sizeof(sai);
#endif

		hwndCB = CreateRpCommandBar(hwnd);
		/* Remove (x) close button in top-right corner
		 * Button moves window to background instead of closing app. Doh!
		 * next option is leave button displayed and subclass "hhtaskbar" window
		 * and close/minimize putty according to user option -- ynezz
		 */
		set_window_style(hwnd, WS_MINIMIZEBOX, 0);

#ifdef SHELL_AYGSHELL
		if (as_SHDoneButton)
			as_SHDoneButton(hwnd, SHDB_SHOWCANCEL);
		if (as_SHFullScreen)
			as_SHFullScreen(hwnd, SHFS_SHOWTASKBAR | SHFS_SHOWSTARTICON);
#endif
	}
	break;
      case WM_CLOSE:
	debug_level((DBG_WMSG, "WM_CLOSE\n"));
	if (is_full_screen())
	    clear_full_screen();
	{
	    char *cap;
	    show_mouseptr(1);
	    cap = dupprintf(load_ansi_string(IDS_WARN_CAPTION_EXIT_CONFIRM), appname);
	    if (session_closed || !conf_get_int(conf, CONF_warn_on_close) ||
		MessageBoxA(hwnd, load_ansi_string(IDS_WARN_WANT_TO_CLOSE),
			    cap, MB_ICONWARNING | MB_OKCANCEL | MB_DEFBUTTON1)
		== IDOK)
		DestroyWindow(hwnd);
	    sfree(cap);
	}
	return 0;
      case WM_QUIT:
      case WM_DESTROY:
	debug_level((DBG_WMSG, "WM_QUIT/WM_DESTROY\n"));
	show_mouseptr(1);
	PostQuitMessage(0);
	return 0;
	break;
      case WM_SYSCOMMAND:
      case WM_COMMAND:
	debug_level((DBG_WMSG, "WM_COMMAND\n"));
	{
	    switch (wParam & ~0xF) {       /* low 4 bits reserved to Windows */
	    case IDM_NEWSESS:
	    case IDM_SAVEDSESS:
	      {
		WCHAR b[FILENAME_MAX];
		char *cl = NULL;
		WCHAR *wcl = NULL;
		DWORD fcreate = CREATE_NEW_CONSOLE;
		PROCESS_INFORMATION pi;

		get_sesslist(&sesslist, TRUE);
		if (wParam == IDM_SAVEDSESS) {
		    const unsigned sessno = ((lParam - IDM_SAVED_MIN)
					     / MENU_SAVED_STEP) + 1;
		    if (sessno < (unsigned)sesslist.nsessions) {
			const char *session = sesslist.sessions[sessno];
			cl = dupprintf("@%s", session);
		    } else {
			get_sesslist(&sesslist, FALSE);
			break;
		    }
		} else /* IDM_NEWSESS */ {
		}
		get_sesslist(&sesslist, FALSE);

		if (is_full_screen())
		    clear_full_screen();

#if defined(CONFIG_SINGLE_INSTANCE)
		queue_toplevel_callback(close_session, NULL);
		run_toplevel_callbacks();
		PostMessage(hwnd, WM_CLOSE, 0, 0);
#endif

		GetModuleFileNameW(NULL, b, sizeof(b) - 1);
		ZeroMemory(&pi, sizeof(PROCESS_INFORMATION));
		if (cl)
		    wcl = ansi2unicode(cl);
		CreateProcessW(b, wcl, NULL, NULL, FALSE,
			       fcreate, NULL, NULL, NULL, &pi);
		CloseHandle(pi.hProcess);
		CloseHandle(pi.hThread);

		if (wcl)
		    sfree(wcl);
		if (cl)
		    sfree(cl);
	      }
	      break;
	    case IDM_RESTART:
		if (!back) {
		    logevent(NULL, "----- Session restarted -----");
		    term_pwron(term, FALSE);
		    start_backend();
		}
		break;
	    case IDM_COPYALL:
		term_copyall(term);
		break;
	    case IDM_PASTE:
		request_paste(NULL);
		break;
	    case IDM_CLRSB:
		term_clrsb(term);
		break;
	    case IDM_RESET:
		term_pwron(term, TRUE);
		if (ldisc)
		    ldisc_echoedit_update(ldisc);
		break;
	    case IDM_CLOSESESSION:
		{
		    char *cap;
		    show_mouseptr(1);
		    cap = dupprintf(load_ansi_string(IDS_WARN_CAPTION_EXIT_CONFIRM), appname);
		    if (session_closed || !conf_get_int(conf, CONF_warn_on_close) ||
			MessageBoxA(hwnd, load_ansi_string(IDS_WARN_WANT_TO_CLOSE),
				    cap, MB_ICONWARNING | MB_OKCANCEL | MB_DEFBUTTON1)
			== IDOK)
			queue_toplevel_callback(close_session, NULL);
		    sfree(cap);
		}
		break;

	    case IDM_ABOUT:
		showabout(hwnd);
		break;
	    case IDM_FULLSCREEN:
		flip_full_screen();
		break;
	    case IDM_CLOSE:
	    case IDOK:
		PostMessage(hwnd, WM_CLOSE, 0, 0);
		break;
	    default:
		if (wParam >= IDM_SAVED_MIN && wParam < IDM_SAVED_MAX) {
			PostMessage(hwnd, WM_COMMAND, IDM_SAVEDSESS, wParam);
		}
		if (wParam >= IDM_SPECIAL_MIN && wParam <= IDM_SPECIAL_MAX) {
			int i = (wParam -IDM_SPECIAL_MIN) /0x10;
			/*
			 * Ensure we haven't been sent a bogus SYSCOMMAND
			 * which would cause us to reference invalid memory
			 * and crash. Perhaps I'm just too paranoid here.
			 */
			if (i >= n_specials)
				break;
			if (back)
				back->special(backhandle, specials[i].code);
			//net_pending_errors();
		}
		return (DefWindowProc(hwnd, message, wParam, lParam));
	    }
	}
	break;

#define X_POS(l) ((int)(short)LOWORD(l))
#define Y_POS(l) ((int)(short)HIWORD(l))

#define TO_CHR_X(x) ((((x)<0 ? (x)-font_width+1 : (x))-offset_width) / font_width)
#define TO_CHR_Y(y) ((((y)<0 ? (y)-font_height+1: (y))-offset_height) / font_height)
      case WM_LBUTTONDOWN:
      case WM_MBUTTONDOWN:
      case WM_RBUTTONDOWN:
      case WM_LBUTTONUP:
      case WM_MBUTTONUP:
      case WM_RBUTTONUP:
	debug_level((DBG_WMSG, "WM_BUTTON*\n"));
	{
	    int button, press;

	    switch (message) {
	      case WM_LBUTTONDOWN:
		{
#ifdef SHELL_AYGSHELL
		  SHRGINFO shrginfo;
		  shrginfo.cbSize = sizeof(SHRGINFO);
		  shrginfo.hwndClient = hwnd;
		  shrginfo.ptDown.x = LOWORD(lParam);
		  shrginfo.ptDown.y = HIWORD(lParam);
		  shrginfo.dwFlags = SHRG_RETURNCMD;
		  if (as_SHRecognizeGesture && as_SHRecognizeGesture(&shrginfo) == GN_CONTEXTMENU) {
		    button = MBT_RIGHT;
		    wParam |= MK_RBUTTON;
		  }
		  else
#endif
		  if (is_alt_pressed()) {
		    button = MBT_RIGHT;
		    wParam |= MK_RBUTTON;
		  } else {
		    button = MBT_LEFT;
		    wParam |= MK_LBUTTON;
		  }
		  press = 1;
		}
		break;
	      case WM_MBUTTONDOWN:
		button = MBT_MIDDLE;
		wParam |= MK_MBUTTON;
		press = 1;
		break;
	      case WM_RBUTTONDOWN:
		button = MBT_RIGHT;
		wParam |= MK_RBUTTON;
		press = 1;
		break;
	      case WM_LBUTTONUP:
		button = MBT_LEFT;
		wParam &= ~MK_LBUTTON;
		press = 0;
		break;
	      case WM_MBUTTONUP:
		button = MBT_MIDDLE;
		wParam &= ~MK_MBUTTON;
		press = 0;
		break;
	      case WM_RBUTTONUP:
		button = MBT_RIGHT;
		wParam &= ~MK_RBUTTON;
		press = 0;
		break;
	      default:
		button = press = 0;    /* shouldn't happen */
	    }
	    show_mouseptr(1);

	    if (press) {
		click(button,
		      TO_CHR_X(X_POS(lParam)), TO_CHR_Y(Y_POS(lParam)),
		      wParam & MK_SHIFT, wParam & MK_CONTROL,
		      is_alt_pressed());
		SetCapture(hwnd);
	    } else {
		term_mouse(term, button, translate_button(button), MA_RELEASE,
			   TO_CHR_X(X_POS(lParam)),
			   TO_CHR_Y(Y_POS(lParam)), wParam & MK_SHIFT,
			   wParam & MK_CONTROL, is_alt_pressed());
		if (!(wParam & (MK_LBUTTON | MK_MBUTTON | MK_RBUTTON)))
		    ReleaseCapture();
	    }
	}
	return 0;
      case WM_MOUSEMOVE:
	debug_level((DBG_WMSG, "WM_MOUSEMOVE\n"));
	{
	    /*
	     * Windows seems to like to occasionally send MOUSEMOVE
	     * events even if the mouse hasn't moved. Don't unhide
	     * the mouse pointer in this case.
	     */
	    static WPARAM wp = 0;
	    static LPARAM lp = 0;
	    if (wParam != wp || lParam != lp ||
		last_mousemove != WM_MOUSEMOVE) {
		show_mouseptr(1);
		wp = wParam; lp = lParam;
		last_mousemove = WM_MOUSEMOVE;
	    }
	}
	/*
	 * Add the mouse position and message time to the random
	 * number noise.
	 */
	noise_ultralight(lParam);

	if (wParam & (MK_LBUTTON | MK_MBUTTON | MK_RBUTTON) &&
	    GetCapture() == hwnd) {
	    Mouse_Button b;
	    if (wParam & MK_LBUTTON)
		b = MBT_LEFT;
	    else if (wParam & MK_MBUTTON)
		b = MBT_MIDDLE;
	    else
		b = MBT_RIGHT;
	    term_mouse(term, b, translate_button(b), MA_DRAG,
		       TO_CHR_X(X_POS(lParam)),
		       TO_CHR_Y(Y_POS(lParam)), wParam & MK_SHIFT,
		       wParam & MK_CONTROL, is_alt_pressed());
	}
	return 0;
      case WM_IGNORE_CLIP:
	debug_level((DBG_WMSG, "WM_IGNORE_CLIP\n"));
	ignore_clip = wParam;	       /* don't panic on DESTROYCLIPBOARD */
	break;
      case WM_DESTROYCLIPBOARD:
	debug_level((DBG_WMSG, "WM_DESTROYCLIPBOARD\n"));
	if (!ignore_clip)
	    term_deselect(term);
	ignore_clip = FALSE;
	return 0;
      case WM_PAINT:
	debug_level((DBG_WMSG, "WM_PAINT\n"));
	{
	    PAINTSTRUCT p;

	    HideCaret(hwnd);
	    hdc = BeginPaint(hwnd, &p);
	    if (pal) {
		SelectPalette(hdc, pal, TRUE);
		RealizePalette(hdc);
	    }
	    if (p.rcPaint.top < override_menu_height) {
		p.rcPaint.top = override_menu_height;
	    }

	    /*
	     * We have to be careful about term_paint(). It will
	     * set a bunch of character cells to INVALID and then
	     * call do_paint(), which will redraw those cells and
	     * _then mark them as done_. This may not be accurate:
	     * when painting in WM_PAINT context we are restricted
	     * to the rectangle which has just been exposed - so if
	     * that only covers _part_ of a character cell and the
	     * rest of it was already visible, that remainder will
	     * not be redrawn at all. Accordingly, we must not
	     * paint any character cell in a WM_PAINT context which
	     * already has a pending update due to terminal output.
	     * The simplest solution to this - and many, many
	     * thanks to Hung-Te Lin for working all this out - is
	     * not to do any actual painting at _all_ if there's a
	     * pending terminal update: just mark the relevant
	     * character cells as INVALID and wait for the
	     * scheduled full update to sort it out.
	     *
	     * I have a suspicion this isn't the _right_ solution.
	     * An alternative approach would be to have terminal.c
	     * separately track what _should_ be on the terminal
	     * screen and what _is_ on the terminal screen, and
	     * have two completely different types of redraw (one
	     * for full updates, which syncs the former with the
	     * terminal itself, and one for WM_PAINT which syncs
	     * the latter with the former); yet another possibility
	     * would be to have the Windows front end do what the
	     * GTK one already does, and maintain a bitmap of the
	     * current terminal appearance so that WM_PAINT becomes
	     * completely trivial. However, this should do for now.
	     */
	    term_paint(term, hdc,
		       (p.rcPaint.left-offset_width)/font_width,
		       (p.rcPaint.top-offset_height)/font_height,
		       (p.rcPaint.right-offset_width-1)/font_width,
		       (p.rcPaint.bottom-offset_height-1)/font_height,
		       !term->window_update_pending);

	    if (p.fErase ||
	        p.rcPaint.left  < offset_width  ||
		p.rcPaint.top   < offset_height ||
		p.rcPaint.right >= offset_width + font_width*term->cols ||
		p.rcPaint.bottom>= offset_height + font_height*term->rows)
	    {
		HBRUSH fillcolour, oldbrush;
		HPEN   edge, oldpen;
		fillcolour = CreateSolidBrush (
				    colours[ATTR_DEFBG>>ATTR_BGSHIFT]);
		oldbrush = (HBRUSH) SelectObject(hdc, fillcolour);
		edge = CreatePen(PS_SOLID, 0,
				    colours[ATTR_DEFBG>>ATTR_BGSHIFT]);
		oldpen = (HPEN) SelectObject(hdc, edge);

		/*
		 * Jordan Russell reports that this apparently
		 * ineffectual IntersectClipRect() call masks a
		 * Windows NT/2K bug causing strange display
		 * problems when the PuTTY window is taller than
		 * the primary monitor. It seems harmless enough...
		 */
		IntersectClipRect(hdc,
			p.rcPaint.left, p.rcPaint.top,
			p.rcPaint.right, p.rcPaint.bottom);

		ExcludeClipRect(hdc,
			offset_width, offset_height,
			offset_width+font_width*term->cols,
			offset_height+font_height*term->rows);

		Rectangle(hdc, p.rcPaint.left, p.rcPaint.top,
			  p.rcPaint.right, p.rcPaint.bottom);

		/* SelectClipRgn(hdc, NULL); */

		SelectObject(hdc, oldbrush);
		DeleteObject(fillcolour);
		SelectObject(hdc, oldpen);
		DeleteObject(edge);
	    }
	    SelectObject(hdc, GetStockObject(SYSTEM_FONT));
	    SelectObject(hdc, GetStockObject(WHITE_PEN));
	    EndPaint(hwnd, &p);
	    ShowCaret(hwnd);
	}
	return 0;
      case WM_NETEVENT:
        {
	    debug_level((DBG_WMSG, "WM_NETEVENT\n"));
	    /* Notice we can get multiple netevents, FD_READ, FD_WRITE etc
	     * but the only one that's likely to try to overload us is FD_READ.
	     * This means buffering just one is fine.
	     */
	    if (pending_netevent)
		enact_pending_netevent();

	    pending_netevent = TRUE;
	    pend_netevent_wParam = wParam;
	    pend_netevent_lParam = lParam;
	    if (WSAGETSELECTEVENT(lParam) != FD_READ)
		enact_pending_netevent();

	    //net_pending_errors();
	    pending_event = FALSE;

            /*
             * To protect against re-entrancy when Windows's recv()
             * immediately triggers a new WSAAsyncSelect window
             * message, we don't call select_result directly from this
             * handler but instead wait until we're back out at the
             * top level of the message loop.
             */
	    struct wm_netevent_params *params =
		snew(struct wm_netevent_params);
	    params->wParam = wParam;
	    params->lParam = lParam;
	    queue_toplevel_callback(wm_netevent_callback, params);
	}
	return 0;
      case WM_SETFOCUS:
	debug_level((DBG_WMSG, "WM_SETFOCUS\n"));
	if (term && fatalbox_visible == FALSE) {
		term_set_focus(term, TRUE);
		CreateCaret(hwnd, caretbm, font_width, font_height);
		ShowCaret(hwnd);
		flash_window(0);	       /* stop */
		compose_state = 0;
		term_update(term);
	}
	break;
      case WM_KILLFOCUS:
	debug_level((DBG_WMSG, "WM_KILLFOCUS\n"));
	if (fatalbox_visible == TRUE) {
		SetForegroundWindow(hwnd);
	} else {
		show_mouseptr(1);
		term_set_focus(term, FALSE);
		DestroyCaret();
		caret_x = caret_y = -1; /* ensure caret is replaced next time */
		term_update(term);
	}
	break;
      case WM_MOVE:
	debug_level((DBG_WMSG, "WM_MOVE\n"));
	sys_cursor_update();
	break;
      case WM_ACTIVATE:
	debug_level((DBG_WMSG, "WM_ACTIVATE\n"));
#ifdef SHELL_AYGSHELL
	if (as_SHHandleWMActivate)
		as_SHHandleWMActivate(hwnd, wParam, lParam, &sai, FALSE);
#endif
	if (LOWORD(wParam) == WA_INACTIVE)
	{
	    got_focus = FALSE;
	    {
		char *name;
		wchar_t *wname;
		name = conf_get_int(conf, CONF_win_name_always) ? window_name : icon_name;
		if (!name)
		    name = window_name;
		if (name) {
		    wname = ansi2unicode(name);
		    SetWindowTextW(hwnd, wname);
		    sfree(wname);
		}
	    }
	    if (is_full_screen())
		clear_full_screen();
	}
	else {
	    got_focus = TRUE;
	    if (window_name) {
		wchar_t *wname;
		wname = ansi2unicode(window_name);
		SetWindowTextW(hwnd, wname);
		sfree(wname);
	    }
	}
	return 0;
      case WM_SETTINGCHANGE:
	debug_level((DBG_WMSG, "WM_SETTINGCHANGE\n"));
#ifdef SHELL_AYGSHELL
	if (wParam == SPI_SETSIPINFO) {
	    if (!is_full_screen()) {
		if (as_SHHandleWMSettingChange)
		    as_SHHandleWMSettingChange(hwnd, -1, 0, &sai); // send wParam=-1, lParam=0
	    }
	}
	else {
	    if (as_SHHandleWMSettingChange)
		as_SHHandleWMSettingChange(hwnd, wParam, lParam, &sai);
	}
#endif
	if (wParam == SPI_SETWORKAREA || wParam == SPI_SETSIPINFO) {
	    if (is_full_screen()) {
		make_full_screen();
	    }
	    else {
		clear_full_screen();
	    }
	}
	return 0;
      case WM_SIZE:
	debug_level((DBG_WMSG, "WM_SIZE\n"));
#ifdef RDB_DEBUG_PATCH
	debug_level((27, "WM_SIZE %s (%d,%d)",
		(wParam == SIZE_MINIMIZED) ? "SIZE_MINIMIZED":
		(wParam == SIZE_MAXIMIZED) ? "SIZE_MAXIMIZED":
		(wParam == SIZE_RESTORED && resizing) ? "to":
		(wParam == SIZE_RESTORED) ? "SIZE_RESTORED":
		"...",
	    LOWORD(lParam), HIWORD(lParam)));
#endif
	if (wParam == SIZE_MINIMIZED)
	    if (is_full_screen())
		clear_full_screen();
#if 0
	if (wParam == SIZE_MINIMIZED) {
		wchar_t *wname;
		wname = ansi2unicode(conf_get_int(conf, CONF_win_name_always) ? window_name : icon_name);
		SetWindowTextW(hwnd, wname);
		sfree(wname);
	}
	if (wParam == SIZE_RESTORED || wParam == SIZE_MAXIMIZED)
		if(window_name) {
			wchar_t *wname;
			wname = ansi2unicode(window_name);
			SetWindowTextW(hwnd, wname);
			sfree(wname);
		}
	if (conf_get_int(conf, CONF_resize_action) == RESIZE_DISABLED) {
	    /* A resize, well it better be a minimize. */
	    reset_window(-1);
	} else {
	    int width, height, w, h;

	    width = LOWORD(lParam);
	    height = HIWORD(lParam) - override_menu_height;

	    if (!resizing) {
		if (wParam == SIZE_MAXIMIZED && !was_zoomed) {
		    was_zoomed = 1;
		    prev_rows = term->rows;
		    prev_cols = term->cols;
		    if (conf_get_int(conf, CONF_resize_action) == RESIZE_TERM) {
			w = width / font_width;
			if (w < 1) w = 1;
			h = height / font_height;
			if (h < 1) h = 1;

			term_size(term, h, w, conf_get_int(conf, CONF_savelines));
		    }
		    reset_window(0);
		} else if (wParam == SIZE_RESTORED && was_zoomed) {
		    was_zoomed = 0;
		    if (conf_get_int(conf, CONF_resize_action) == RESIZE_TERM)
			term_size(term, prev_rows, prev_cols, conf_get_int(conf, CONF_savelines));
		    if (conf_get_int(conf, CONF_resize_action) != RESIZE_FONT)
			reset_window(2);
		    else
			reset_window(0);
		}
		/* This is an unexpected resize, these will normally happen
		 * if the window is too large. Probably either the user
		 * selected a huge font or the screen size has changed.
		 *
		 * This is also called with minimize.
		 */
		else reset_window(-1);
	    }
	}
#endif
	sys_cursor_update();
	return 0;
      case WM_VSCROLL:
	debug_level((DBG_WMSG, "WM_VSCROLL\n"));
	switch (LOWORD(wParam)) {
	  case SB_BOTTOM:
	    term_scroll(term, -1, 0);
	    break;
	  case SB_TOP:
	    term_scroll(term, +1, 0);
	    break;
	  case SB_LINEDOWN:
	    term_scroll(term, 0, +1);
	    break;
	  case SB_LINEUP:
	    term_scroll(term, 0, -1);
	    break;
	  case SB_PAGEDOWN:
	    term_scroll(term, 0, +term->rows / 2);
	    break;
	  case SB_PAGEUP:
	    term_scroll(term, 0, -term->rows / 2);
	    break;
	  case SB_THUMBPOSITION:
	  case SB_THUMBTRACK:
	    /*
	     * Use GetScrollInfo instead of HIWORD(wParam) to get
	     * 32-bit scroll position.
	     */
	    {
		SCROLLINFO si;

		si.cbSize = sizeof(si);
		si.fMask = SIF_TRACKPOS;
		if (GetScrollInfo(hwnd, SB_VERT, &si) == 0)
		    si.nTrackPos = HIWORD(wParam);
		term_scroll(term, 1, si.nTrackPos);
	    }
	    break;
	}
	break;
	case WM_KEYUP:
	{
	    unsigned char buf[20];
	    int len;

		debug_level((DBG_WMSG, "WM_KEYUP\n"));
		if (wParam == VK_CLEAR && lParam == 0xffffffff80000001) {
			MSG m;
			m.hwnd = hwnd;
			m.message = WM_KEYDOWN;
			m.wParam = VK_SPACE;
			m.lParam = lParam & 0xdfff;
			TranslateMessage(&m);
		}

		/*
		 * Push the message to TranslateKey as it needs to do special things on
		 * WM_KEYUP to support page up / page down and the scroll wheel.  Note
		 * that most keys continue to be handled exclusively by the key down
		 * handler.
		 */

		len = TranslateKey(message, wParam, lParam, buf);
		if (len == -1) {
			MSG m;
			m.hwnd = hwnd;
			m.message = WM_KEYDOWN;
			m.wParam = wParam;
			m.lParam = lParam;
			TranslateMessage(&m);
			return DefWindowProc(hwnd, message, wParam, lParam);
		}
		if (len != 0) {
		    /*
		     * Interrupt an ongoing paste. I'm not sure
		     * this is sensible, but for the moment it's
		     * preferable to having to faff about buffering
		     * things.
		     */
		    term_nopaste(term);

		    /*
		     * We need not bother about stdin backlogs
		     * here, because in GUI PuTTY we can't do
		     * anything about it anyway; there's no means
		     * of asking Windows to hold off on KEYDOWN
		     * messages. We _have_ to buffer everything
		     * we're sent.
		     */
		    term_seen_key_event(term);
		    if (ldisc)
			ldisc_send(ldisc, (char *)buf, len, 1);
		    show_mouseptr(0);
		}
	}
	break;

	case WM_KEYDOWN:
	case WM_SYSKEYDOWN:
	debug_level((DBG_WMSG, "WM_SYSKEYDOWN/WM_KEYDOWN\n"));
	/*
	 * Add the scan code and keypress timing to the random
	 * number noise.
	 */
	noise_ultralight(lParam);

	/*
	 * We don't do TranslateMessage since it disassociates the
	 * resulting CHAR message from the KEYDOWN that sparked it,
	 * which we occasionally don't want. Instead, we process
	 * KEYDOWN, and call the Win32 translator functions so that
	 * we get the translations under _our_ control.
	 */
	{
	    unsigned char buf[20];
	    int len;
		if (wParam == VK_PROCESSKEY) { /* IME PROCESS key */
		if (message == WM_KEYDOWN) {
			MSG m;
		    m.hwnd = hwnd;
		    m.message = WM_KEYDOWN;
		    m.wParam = wParam;
		    m.lParam = lParam & 0xdfff;
		    TranslateMessage(&m);
		} else break; /* pass to Windows for default processing */
	    } else {
#if UNDER_CE < 0x400
		if (wParam == VK_CLEAR) len = -1;
		else
#endif
		len = TranslateKey(message, wParam, lParam, buf);
		if (len == -1) {
			MSG m;
			m.hwnd = hwnd;
			m.message = WM_KEYDOWN;
			m.wParam = wParam;
			m.lParam = lParam;
			TranslateMessage(&m);
			return DefWindowProc(hwnd, message, wParam, lParam);
		}
		if (len != 0) {
		    /*
		     * Interrupt an ongoing paste. I'm not sure
		     * this is sensible, but for the moment it's
		     * preferable to having to faff about buffering
		     * things.
		     */
		    term_nopaste(term);

		    /*
		     * We need not bother about stdin backlogs
		     * here, because in GUI PuTTY we can't do
		     * anything about it anyway; there's no means
		     * of asking Windows to hold off on KEYDOWN
		     * messages. We _have_ to buffer everything
		     * we're sent.
		     */
		    term_seen_key_event(term);
		    if (ldisc)
			ldisc_send(ldisc, (char *)buf, len, 1);
		    show_mouseptr(0);
		}
	    }
	}
	//net_pending_errors();
	return 0;
	break;
      case WM_INPUTLANGCHANGE:
	/* wParam == Font number */
	/* lParam == Locale */
	set_input_locale((HKL)lParam);
	sys_cursor_update();
	break;
      case WM_IME_STARTCOMPOSITION:
	debug_level((DBG_WMSG, "WM_IME_STARTCOMPOSITION\n"));
	{
		HIMC hImc = ImmGetContext(hwnd);
		ImmSetCompositionFont(hImc, &lfont);
		ImmReleaseContext(hwnd, hImc);
		return 0;
	}
	break;
      case WM_IME_NOTIFY:
	debug_level((DBG_WMSG, "WM_IME_NOTIFY\n"));
	{
		if(wParam == IMN_SETOPENSTATUS) {
			HIMC hImc = ImmGetContext(hwnd);
			ImmSetCompositionFont(hImc, &lfont);
			ImmReleaseContext(hwnd, hImc);
			return 0;
		}
	}
	break;
      case WM_IME_COMPOSITION:
        debug_level((DBG_WMSG, "WM_IME_COMPOSITION\n"));
	{
	    HIMC hIMC;
	    int n;
	    char *buff;

	    if(osVersion.dwPlatformId == VER_PLATFORM_WIN32_WINDOWS ||
	        osVersion.dwPlatformId == VER_PLATFORM_WIN32s) break; /* no Unicode */

	    if ((lParam & GCS_RESULTSTR) == 0) /* Composition unfinished. */
		break; /* fall back to DefWindowProc */

	    hIMC = ImmGetContext(hwnd);
	    n = ImmGetCompositionStringW(hIMC, GCS_RESULTSTR, NULL, 0);

	    if (n > 0) {
		int i;
		buff = snewn(n, char);
		ImmGetCompositionStringW(hIMC, GCS_RESULTSTR, buff, n);
		/*
		 * Jaeyoun Chung reports that Korean character
		 * input doesn't work correctly if we do a single
		 * luni_send() covering the whole of buff. So
		 * instead we luni_send the characters one by one.
		 */
		term_seen_key_event(term);
		/* don't divide SURROGATE PAIR */
		if (ldisc) {
		    for (i = 0; i < n; i += 2) {
			WCHAR hs = *(unsigned short *)(buff+i);
			if (IS_HIGH_SURROGATE(hs) && i+2 < n) {
			    WCHAR ls = *(unsigned short *)(buff+i+2);
			    if (IS_LOW_SURROGATE(ls)) {
				luni_send(ldisc, (unsigned short *)(buff+i), 2, 1);
				i += 2;
				continue;
			    }
			}
			luni_send(ldisc, (unsigned short *)(buff+i), 1, 1);
                    }
		}
		free(buff);
	    }
	    ImmReleaseContext(hwnd, hIMC);
	    return 1;
	}

      case WM_IME_CHAR:
	debug_level((DBG_WMSG, "WM_IME_CHAR\n"));
	if (wParam & 0xFF00) {
	    char buf[2];

	    buf[1] = wParam;
	    buf[0] = wParam >> 8;
	    term_seen_key_event(term);
	    if (ldisc)
		lpage_send(ldisc, kbd_codepage, buf, 2, 1);
	} else {
	    char c = (char) wParam;
	    term_seen_key_event(term);
	    if (ldisc)
		lpage_send(ldisc, kbd_codepage, &c, 1, 1);
	}
	return (0);
      case WM_CHAR:
      case WM_SYSCHAR:
	debug_level((DBG_WMSG, "WM_CHAR/WM_SYSCHAR\n"));
	/*
	 * Nevertheless, we are prepared to deal with WM_CHAR
	 * messages, should they crop up. So if someone wants to
	 * post the things to us as part of a macro manoeuvre,
	 * we're ready to cope.
	 */
	{
	    char c = (char) wParam;
	    term_seen_key_event(term);
	    if (ldisc)
		lpage_send(ldisc, CP_ACP, &c, 1, 1);
	}
	return 0;
	break;
      case WM_AGENT_CALLBACK:
	{
	    struct agent_callback *c = (struct agent_callback *)lParam;
	    c->callback(c->callback_ctx, c->data, c->len);
	    sfree(c);
	}
	return 0;
      case WM_GOT_CLIPDATA:
	if (process_clipdata((HLOCAL)lParam, wParam))
	    term_do_paste(term);
	return 0;
      default:
	      break;
    }

    /*
     * Any messages we don't process completely above are passed through to
     * DefWindowProc() for default processing.
     */
     debug_level((DBG_WMSG, "DefWndProc\n"));
     return DefWindowProc(hwnd, message, wParam, lParam);
}

/*
 * Move the system caret. (We maintain one, even though it's
 * invisible, for the benefit of blind people: apparently some
 * helper software tracks the system caret, so we should arrange to
 * have one.)
 */
void sys_cursor(void *frontend, int x, int y)
{
	int cx, cy;

	debug_level((DBG_FUNCTION, "sys_cursor(x: %i y: %i);\n", x, y));

	if (!term->has_focus) return;

	/*
	 * Avoid gratuitously re-updating the cursor position and IMM
	 * window if there's no actual change required.
	 */
	cx = x * font_width + offset_width;
	cy = y * font_height + offset_height;
	if (cx == caret_x && cy == caret_y)
		return;
	caret_x = cx;
	caret_y = cy;

	sys_cursor_update();
}

static void sys_cursor_update(void)
{
	COMPOSITIONFORM cf;
	HIMC hIMC;

	debug_level((DBG_FUNCTION, "sys_cursor_update();\n"));

	if(!term) return;

	if (!term->has_focus)
		return;

	if (caret_x < 0 || caret_y < 0)
		return;

	SetCaretPos(caret_x, caret_y);

	/* IMM calls on Win98 and beyond only */
	if(osVersion.dwPlatformId == VER_PLATFORM_WIN32s) return; /* 3.11 */

	if(osVersion.dwPlatformId == VER_PLATFORM_WIN32_WINDOWS &&
		osVersion.dwMinorVersion == 0) return; /* 95 */

	/* we should have the IMM functions */
	hIMC = ImmGetContext(hwnd);
	cf.dwStyle = CFS_POINT;
	cf.ptCurrentPos.x = caret_x;
	cf.ptCurrentPos.y = caret_y;
	ImmSetCompositionWindow(hIMC, &cf);

	ImmReleaseContext(hwnd, hIMC);
}

/*
 * Draw a line of text in the window, at given character
 * coordinates, in given attributes.
 *
 * We are allowed to fiddle with the contents of `text'.
 */
void do_text_internal(Context ctx, int x, int y, wchar_t *text, int len,
		      unsigned long attr, int lattr)
{
    COLORREF fg, bg, t;
    int nfg, nbg, nfont;
    HDC hdc = ctx;
    RECT line_box;
    int force_manual_underline = 0;
    int fnt_width, char_width;
    int text_adjust = 0;
    static int *IpDx = 0, IpDxLEN = 0;

    lattr &= LATTR_MODE;

    char_width = fnt_width = font_width * (1 + (lattr != LATTR_NORM));

    if (attr & ATTR_WIDE)
	char_width *= 2;

    if (len > IpDxLEN || IpDx[0] != char_width) {
	int i;
	if (len > IpDxLEN) {
	    sfree(IpDx);
	    IpDx = snewn(len + 16, int);
	    IpDxLEN = (len + 16);
	}
	for (i = 0; i < IpDxLEN; i++)
	    IpDx[i] = char_width;
    }

    /* Only want the left half of double width lines */
    if (lattr != LATTR_NORM && x*2 >= term->cols)
	return;

    x *= fnt_width;
    y *= font_height;
    x += offset_width;
    y += offset_height;

    if ((attr & TATTR_ACTCURS) && (conf_get_int(conf, CONF_cursor_type) == 0 || term->big_cursor)) {
	attr &= ~(ATTR_REVERSE|ATTR_BLINK|ATTR_COLOURS);
	if (bold_colours == BOLD_NONE)
	    attr &= ~ATTR_BOLD;

	/* cursor fg and bg */
	attr |= (260 << ATTR_FGSHIFT) | (261 << ATTR_BGSHIFT);
    }

    nfont = 0;
    if (conf_get_int(conf, CONF_vtmode) == VT_POORMAN && lattr != LATTR_NORM) {
	/* Assume a poorman font is borken in other ways too. */
	lattr = LATTR_WIDE;
    } else
	switch (lattr) {
	  case LATTR_NORM:
	    break;
	  case LATTR_WIDE:
	    nfont |= FONT_WIDE;
	    break;
	  default:
	    nfont |= FONT_WIDE + FONT_HIGH;
	    break;
	}
    if (attr & ATTR_NARROW)
	nfont |= FONT_NARROW;

    /* Special hack for the VT100 linedraw glyphs. */
    if (text[0] >= 0x23BA && text[0] <= 0x23BD) {
	switch ((unsigned char) (text[0])) {
	  case 0xBA:
	    text_adjust = -2 * font_height / 5;
	    break;
	  case 0xBB:
	    text_adjust = -1 * font_height / 5;
	    break;
	  case 0xBC:
	    text_adjust = font_height / 5;
	    break;
	  case 0xBD:
	    text_adjust = 2 * font_height / 5;
	    break;
	}
	if (lattr == LATTR_TOP || lattr == LATTR_BOT)
	    text_adjust *= 2;
	text[0] = ucsdata.unitab_xterm['q'];
	if (attr & ATTR_UNDER) {
	    attr &= ~ATTR_UNDER;
	    force_manual_underline = 1;
	}
    }

    /* Anything left as an original character set is unprintable. */
    if (DIRECT_CHAR(text[0])) {
	int i;
	for (i = 0; i < len; i++)
	    text[i] = 0xFFFD;
    }

    /* OEM CP */
    if ((text[0] & CSET_MASK) == CSET_OEMCP)
	nfont |= FONT_OEM;

    nfg = ((attr & ATTR_FGMASK) >> ATTR_FGSHIFT);
    nbg = ((attr & ATTR_BGMASK) >> ATTR_BGSHIFT);
    if (bold_colours == BOLD_FONT && (attr & ATTR_BOLD))
	nfont |= FONT_BOLD;
    if (und_mode == UND_FONT && (attr & ATTR_UNDER))
	nfont |= FONT_UNDERLINE;
    another_font(nfont);
    if (!fonts[nfont]) {
	if (nfont & FONT_UNDERLINE)
	    force_manual_underline = 1;
	/* Don't do the same for manual bold, it could be bad news. */

	nfont &= ~(FONT_BOLD | FONT_UNDERLINE);
    }
    another_font(nfont);
    if (!fonts[nfont])
	nfont = FONT_NORMAL;
    if (attr & ATTR_REVERSE) {
	t = nfg;
	nfg = nbg;
	nbg = t;
    }
    if (bold_colours == BOLD_NONE && (attr & ATTR_BOLD)) {
	if (nfg < 16) nfg |= 8;
	else if (nfg >= 256) nfg |= 1;
    }
    if (bold_colours == BOLD_NONE && (attr & ATTR_BLINK)) {
	if (nbg < 16) nbg |= 8;
	else if (nbg >= 256) nbg |= 1;
    }
    fg = colours[nfg];
    bg = colours[nbg];
    SelectObject(hdc, fonts[nfont]);
    SetTextColor(hdc, fg);
    SetBkColor(hdc, bg);
    if (attr & TATTR_COMBINING)
	SetBkMode(hdc, TRANSPARENT);
    else
	SetBkMode(hdc, OPAQUE);
    line_box.left = x;
    line_box.top = y;
    line_box.right = x + char_width * len;
    line_box.bottom = y + font_height;

    /* Only want the left half of double width lines */
    if (line_box.right > font_width*term->cols+offset_width)
	line_box.right = font_width*term->cols+offset_width;

    /* We're using a private area for direct to font. (512 chars.) */
    if (ucsdata.dbcs_screenfont && (text[0] & CSET_MASK) == CSET_ACP) {
	/* Ho Hum, dbcs fonts are a PITA! */
	/* To display on W9x I have to convert to UCS */
	static wchar_t *uni_buf = 0;
	static int uni_len = 0;
	int nlen, mptr;
	if (len > uni_len) {
	    sfree(uni_buf);
	    uni_len = len;
	    uni_buf = snewn(uni_len, wchar_t);
	}

	for(nlen = mptr = 0; mptr<len; mptr++) {
	    uni_buf[nlen] = 0xFFFD;
	    if (IsDBCSLeadByteEx(ucsdata.font_codepage, (BYTE) text[mptr])) {
		char dbcstext[2];
		dbcstext[0] = text[mptr] & 0xFF;
		dbcstext[1] = text[mptr+1] & 0xFF;
		IpDx[nlen] += char_width;
	        MultiByteToWideChar(ucsdata.font_codepage, MB_USEGLYPHCHARS,
				    dbcstext, 2, uni_buf+nlen, 1);
		mptr++;
	    }
	    else
	    {
		char dbcstext[1];
		dbcstext[0] = text[mptr] & 0xFF;
	        MultiByteToWideChar(ucsdata.font_codepage, MB_USEGLYPHCHARS,
				    dbcstext, 1, uni_buf+nlen, 1);
	    }
	    nlen++;
	}
	if (nlen <= 0)
	    return;		       /* Eeek! */

	ExtTextOutW(hdc, x,
		    y - font_height * (lattr == LATTR_BOT) + text_adjust,
		    ETO_CLIPPED | ETO_OPAQUE, &line_box, uni_buf, nlen, IpDx);
	if (bold_colours == BOLD_SHADOW && (attr & ATTR_BOLD)) {
	    SetBkMode(hdc, TRANSPARENT);
	    ExtTextOutW(hdc, x - 1,
			y - font_height * (lattr ==
					   LATTR_BOT) + text_adjust,
			ETO_CLIPPED, &line_box, uni_buf, nlen, IpDx);
	}

	IpDx[0] = -1;
    } else if (DIRECT_FONT(text[0])) {
	static char *directbuf = NULL;
	static int directlen = 0;
	wchar_t *wdirectbuf;
	int i;

	if (len > directlen) {
	    directlen = len;
	    directbuf = sresize(directbuf, directlen, char);
	}

	for (i = 0; i < len; i++)
	    directbuf[i] = text[i] & 0xFF;

	wdirectbuf = ansi2unicode(directbuf);
	ExtTextOutW(hdc, x,
		    y - font_height * (lattr == LATTR_BOT) + text_adjust,
		    ETO_CLIPPED | ETO_OPAQUE, &line_box, wdirectbuf, len, IpDx);
	if (bold_colours == BOLD_SHADOW && (attr & ATTR_BOLD)) {
	    SetBkMode(hdc, TRANSPARENT);

	    /* GRR: This draws the character outside it's box and can leave
	     * 'droppings' even with the clip box! I suppose I could loop it
	     * one character at a time ... yuk.
	     *
	     * Or ... I could do a test print with "W", and use +1 or -1 for this
	     * shift depending on if the leftmost column is blank...
	     */
	    ExtTextOutW(hdc, x - 1,
			y - font_height * (lattr ==
					   LATTR_BOT) + text_adjust,
			ETO_CLIPPED, &line_box, wdirectbuf, len, IpDx);
	}
	sfree(wdirectbuf);
    } else {
	/* And 'normal' unicode characters */
	static WCHAR *wbuf = NULL;
	static int wlen = 0;
	int i;

	if (wlen < len) {
	    sfree(wbuf);
	    wlen = len;
	    wbuf = snewn(wlen, WCHAR);
	}

	for (i = 0; i < len; i++)
	    wbuf[i] = text[i];

	/* print Glyphs as they are, without Windows' Shaping*/
	general_textout(hdc, x, y - font_height * (lattr == LATTR_BOT) + text_adjust,
			&line_box, wbuf, len, IpDx, !(attr & TATTR_COMBINING));

	/* And the shadow bold hack. */
	if (bold_colours == BOLD_SHADOW && (attr & ATTR_BOLD)) {
	    SetBkMode(hdc, TRANSPARENT);
	    ExtTextOutW(hdc, x - 1,
			y - font_height * (lattr ==
					   LATTR_BOT) + text_adjust,
			ETO_CLIPPED, &line_box, wbuf, len, IpDx);
	}
    }
    if (lattr != LATTR_TOP && (force_manual_underline ||
			       (und_mode == UND_LINE
				&& (attr & ATTR_UNDER)))) {
	HPEN oldpen;
	int dec = descent;
	if (lattr == LATTR_BOT)
	    dec = dec * 2 - font_height;

	oldpen = (HPEN)SelectObject(hdc, CreatePen(PS_SOLID, 0, fg));
#if UNDER_CE < 0x400
	{
		POINT poly[2];
		poly[0].x = x;
		poly[0].y = y + dec;
		poly[1].x = x + len * char_width;
		poly[1].y = y + dec;
	}
#else
	MoveToEx(hdc, x, y + dec, NULL);
	LineTo(hdc, x + len * char_width, y + dec);
#endif
	oldpen = (HPEN)SelectObject(hdc, oldpen);
	DeleteObject(oldpen);
    }
}

/*
 * Wrapper that handles combining characters.
 */
void do_text(Context ctx, int x, int y, wchar_t *text, int len,
		unsigned long attr, int lattr)
{
	debug_level((DBG_FUNCTION, "do_text(%i);\n", len));

	if (attr & TATTR_COMBINING) {
		unsigned long a = 0;
		attr &= ~TATTR_COMBINING;

		while (len--) {
			do_text_internal(ctx, x, y, text, 1, attr | a, lattr);
			text++;
			a = TATTR_COMBINING;
		}
	} else {
		do_text_internal(ctx, x, y, text, len, attr, lattr);
	}
}

void do_cursor(Context ctx, int x, int y, wchar_t *text, int len,
	       unsigned long attr, int lattr)
{

    int fnt_width;
    int char_width;
    HDC hdc = ctx;
    int ctype = conf_get_int(conf, CONF_cursor_type);

    debug_level((DBG_FUNCTION, "do_cursor(%i);\n", len));

    lattr &= LATTR_MODE;

    if ((attr & TATTR_ACTCURS) && (ctype == 0 || term->big_cursor)) {
	if (*text != UCSWIDE) {
	    do_text(ctx, x, y, text, len, attr, lattr);
	    return;
	}
	ctype = 2;
	attr |= TATTR_RIGHTCURS;
    }

    fnt_width = char_width = font_width * (1 + (lattr != LATTR_NORM));
    if (attr & ATTR_WIDE)
	char_width *= 2;
    x *= fnt_width;
    y *= font_height;
    x += offset_width;
    y += offset_height;

    if ((attr & TATTR_PASCURS) && (ctype == 0 || term->big_cursor)) {
	POINT pts[5];
	HPEN oldpen;
	pts[0].x = pts[1].x = pts[4].x = x;
	pts[2].x = pts[3].x = x + char_width - 1;
	pts[0].y = pts[3].y = pts[4].y = y;
	pts[1].y = pts[2].y = y + font_height - 1;
	oldpen = (HPEN) SelectObject(hdc, CreatePen(PS_SOLID, 0, colours[261]));
	Polyline(hdc, pts, 5);
	oldpen = (HPEN) SelectObject(hdc, oldpen);
	DeleteObject(oldpen);
    } else if ((attr & (TATTR_ACTCURS | TATTR_PASCURS)) && ctype != 0) {
	int startx, starty, dx, dy, length, i;
	if (ctype == 1) {
	    startx = x;
	    starty = y + descent;
	    dx = 1;
	    dy = 0;
	    length = char_width;
	} else {
	    int xadjust = 0;
	    if (attr & TATTR_RIGHTCURS)
		xadjust = char_width - 1;
	    startx = x + xadjust;
	    starty = y;
	    dx = 0;
	    dy = 1;
	    length = font_height;
	}
	if (attr & TATTR_ACTCURS) {
	    HPEN oldpen;
	    oldpen =
		(HPEN) SelectObject(hdc, CreatePen(PS_SOLID, 0, colours[261]));
#if UNDER_CE < 0x400
		{
			POINT poly[2];
			poly[0].x = startx;
			poly[0].y = starty;
			poly[1].x = startx +dx *length;
			poly[1].y = starty +dx *length;
		}
#else
	    MoveToEx(hdc, startx, starty, NULL);
	    LineTo(hdc, startx + dx * length, starty + dy * length);
#endif
	    oldpen = (HPEN) SelectObject(hdc, oldpen);
	    DeleteObject(oldpen);
	} else {
	    for (i = 0; i < length; i++) {
		if (i % 2 == 0) {
		    SetPixel(hdc, startx, starty, colours[261]);
		}
		startx += dx;
		starty += dy;
	    }
	}
    }
}

/* This function gets the actual width of a character in the normal font.
 */
int char_width(Context ctx, int uc) {
    HDC hdc = ctx;
    int ibuf = 0;

    /* If the font max is the same as the font ave width then this
     * function is a no-op.
     */
    if (!font_dualwidth) return 1;

    switch (uc & CSET_MASK) {
      case CSET_ASCII:
	uc = ucsdata.unitab_line[uc & 0xFF];
	break;
      case CSET_LINEDRW:
	uc = ucsdata.unitab_xterm[uc & 0xFF];
	break;
      case CSET_SCOACS:
	uc = ucsdata.unitab_scoacs[uc & 0xFF];
	break;
    }
    if (DIRECT_FONT(uc)) {
	if (ucsdata.dbcs_screenfont) return 1;

	/* Speedup, I know of no font where ascii is the wrong width */
	if ((uc&~CSET_MASK) >= ' ' && (uc&~CSET_MASK)<= '~')
	    return 1;

	if ( (uc & CSET_MASK) == CSET_ACP ) {
	    SelectObject(hdc, fonts[FONT_NORMAL]);
	} else if ( (uc & CSET_MASK) == CSET_OEMCP ) {
	    another_font(FONT_OEM);
	    if (!fonts[FONT_OEM]) return 0;

	    SelectObject(hdc, fonts[FONT_OEM]);
	} else
	    return 0;

#if UNDER_CE < 0x400
	{
	    TCHAR ucw[2];
	    RECT rect;
	    char c = (char) (uc&~CSET_MASK);
	    rect.top = 0;
	    rect.left = 0;
	    rect.right = 0;
	    rect.right = 0;
	    mbstowcs(ucw, &c, 1);
	    ucw[1] = TEXT('\0');
	    DrawText(hdc, ucw, 1, &rect, DT_CALCRECT);
	    ibuf = rect.right - rect.left;
	}
#else
	if ( GetCharWidth32(hdc, uc&~CSET_MASK, uc&~CSET_MASK, &ibuf) != 1)
	    return 0;
#endif
    } else {
	/* Speedup, I know of no font where ascii is the wrong width */
	if (uc >= ' ' && uc <= '~') return 1;

	SelectObject(hdc, fonts[FONT_NORMAL]);
#if UNDER_CE < 0x400
	{
	    TCHAR ucw[2];
	    RECT rect;
	    char c = (char) (uc&~CSET_MASK);
	    rect.top = 0;
	    rect.left = 0;
	    rect.right = 0;
	    rect.right = 0;
	    mbstowcs(ucw, &c, 1);
	    ucw[1] = TEXT('\0');
	    DrawText(hdc, ucw, 1, &rect, DT_CALCRECT);
	    ibuf = rect.right - rect.left;
	}
#else
	if ( GetCharWidth32W(hdc, uc, uc, &ibuf) != 1 )
	    return 0;
#endif
    }

    ibuf += font_width / 2 -1;
    ibuf /= font_width;

    return ibuf;
}

/*
 * Translate a WM_(SYS)?KEY(UP|DOWN) message into a string of ASCII
 * codes. Returns number of bytes used, zero to drop the message,
 * -1 to forward the message to Windows, or another negative number
 * to indicate a NUL-terminated "special" string.
 */
static int TranslateKey(UINT message, WPARAM wParam, LPARAM lParam,
			unsigned char *output)
{
	BYTE keystate[256];
	int scan, left_alt = 0, key_down, shift_state;
	int r, i, code;
	unsigned char *p = output;
	static int alt_sum = 0;
	int funky_type = conf_get_int(conf, CONF_funky_type);
	int no_applic_k = conf_get_int(conf, CONF_no_applic_k);
	int ctrlaltkeys = conf_get_int(conf, CONF_ctrlaltkeys);
	int nethack_keypad = conf_get_int(conf, CONF_nethack_keypad);

#if defined(CONFIG_HACK_PAGE_KEYS)
	/*
	 * Giant hack for page up / page down support.  These are reported as a
	 * key down for up arrow / down arrow, *followed by* a key down for a
	 * magical key (VK_F21 - 84 for the keyboard, VK_F20 - 83 for the scroll
	 * wheel.  As far as I can tell, the key state is always identical in these
	 * cases, and the message parameters are the same.  So I don't really see
	 * a good way to handle this other than to check if were previously
	 * handling an up arrow / down arrow event and now got one of the magical
	 * function keys.  Feel free to replace this code with something more sane
	 * if you determine a better way of supporting this.  I determined the
	 * magical VK_F20/VK_F21 keys on a Verizon XV6800 (WM6 Pro).  - Skywing
	 */

	static int up_arrow_state = 0;
	static int down_arrow_state = 0;
	static int last_arrow_key = 0;
#endif /* defined(CONFIG_HACK_PAGE_KEYS) */

	HKL kbd_layout = GetKeyboardLayout(0);

	/* keys is for ToAsciiEx. There's some ick here, see below. */
	static WORD keys[3];
	static int compose_char = 0;
	static WPARAM compose_keycode = 0;

	debug_level((DBG_FUNCTION, "TranslateKey("));
	debug_level((DBG_VERBOSE, "HKL=0x%08X,", (DWORD)kbd_layout));

	r = GetKeyboardState(keystate);
	if (!r) {
		memset(keystate, 0, sizeof(keystate));
	} else {
		if(debug_level & DBG_VERBOSE) {
			/* Tell us all about key events */
			static BYTE oldstate[256];
			static int first = 1;
			static int scan;
			int ch;
			if (first)
				memcpy(oldstate, keystate, sizeof(oldstate));
			first = 0;

			if ((HIWORD(lParam) & (KF_UP | KF_REPEAT)) == KF_REPEAT) {
				debug_level((DBG_VERBOSE, "+"));
			} else if ((HIWORD(lParam) & KF_UP) && scan == (HIWORD(lParam) & 0xFF)) {
				debug_level((DBG_VERBOSE, ". U"));
		} else {
			if (wParam >= VK_F1 && wParam <= VK_F20) {
				debug_level((DBG_VERBOSE, "K_F%d", wParam + 1 - VK_F1));
			} else {

				switch (wParam) {
				case VK_SHIFT:
					debug_level((DBG_VERBOSE, "SHIFT"));
					break;
				case VK_CONTROL:
					debug_level((DBG_VERBOSE, "CTRL"));
					break;
				case VK_MENU:
					debug_level((DBG_VERBOSE, "ALT"));
					break;
				default:
					debug_level((DBG_VERBOSE, "VK_%02x", wParam));
				}

				if (message == WM_SYSKEYDOWN || message == WM_SYSKEYUP) {
					debug_level((DBG_VERBOSE, "*"));
				}

				debug_level((DBG_VERBOSE, ", S%02x", scan = (HIWORD(lParam) & 0xFF)));

				ch = MapVirtualKey(wParam, 2);
				if (ch >= ' ' && ch <= '~') {
					debug_level((DBG_VERBOSE, ", '%c'", ch));
				} else if (ch) {
					debug_level((DBG_VERBOSE, ", $%02x", ch));
				}

				if (keys[0]) {
					debug_level((DBG_VERBOSE, ", KB0=%02x", keys[0]));
				}

				if (keys[1]) {
					debug_level((DBG_VERBOSE, ", KB1=%02x", keys[1]));
				}

				if (keys[2]) {
					debug_level((DBG_VERBOSE, ", KB2=%02x", keys[2]));
				}

				if ((keystate[VK_SHIFT] & 0x80) != 0) {
					debug_level((DBG_VERBOSE, ", S"));
				}

				if ((keystate[VK_CONTROL] & 0x80) != 0) {
					debug_level((DBG_VERBOSE, ", C"));
				}

				if ((HIWORD(lParam) & KF_EXTENDED)) {
					debug_level((DBG_VERBOSE, ", E"));
				}

				if ((HIWORD(lParam) & KF_UP)) {
					debug_level((DBG_VERBOSE, ", U"));
				}
			}

			if ((HIWORD(lParam) & (KF_UP | KF_REPEAT)) == KF_REPEAT) {
			} else if ((HIWORD(lParam) & KF_UP)) {
				oldstate[wParam & 0xFF] ^= 0x80;
			} else {
				oldstate[wParam & 0xFF] ^= 0x81;
			}

			for (ch = 0; ch < 256; ch++) {
				if (oldstate[ch] != keystate[ch]) {
					debug_level((DBG_VERBOSE, ", M%02x=%02x", ch, keystate[ch]));
				}
			}

			memcpy(oldstate, keystate, sizeof(oldstate));
		} /* if(debug_level & DBG_VERBOSE) */
	} /* if(!r) */

#if defined(CONFIG_HACK_PAGE_KEYS)
	/* 
	 * First, let's check if we're handling an up arrow or down arrow.  If we
	 * are the first to enter this state, then we shall defer the processing
	 * until the next message about that key.  If we get one of the magical
	 * function keys while in this state, we'll generate a page up or page down
	 * event and exit the up arrow / down arrow state as appropriate.  What a
	 * giant hack, but it works, and the old behavior was to treat all of those
	 * keys as up arrow / down arrow (which is highly suboptimal).
	 *
	 * Note that this does have the slightly unpleasant side effect of making a
	 * down arrow or up arrow keystroke only register on either *key repeat* or
	 * *key release*.  In either case, we don't actually drop any keystrokes,
	 * so the correct number of keypresses are actually registered, just that
	 * they may appear to show up a fraction of a second later than they would
	 * for other keys.  This is because we (necessarily) are waiting for the
	 * next message to see if it's going to be a 
	 */

	switch (wParam)
	{

	case VK_UP:
	case VK_DOWN:
		/*
		 * Record the last arrow key so we know who to pair the function key
		 * press with if the user has lots of keys mashed down at the same
		 * time.
		 */

		last_arrow_key = (int)wParam;

		if ((lParam & 0x80000000) == 0)
		{
			/*
			 * Key is being pressed.  Let's check if we were already in the
			 * state.  If so then we should dispatch the keystroke, otherwise
			 * we shall hold it for the next message.
			 */

			if (wParam == VK_UP)
			{
				if (!up_arrow_state)
				{
					debug_level((DBG_VERBOSE, "entering UP_ARROW_STATE\n"));
					up_arrow_state = 1;
					return 0;
				}
				else
				{
					debug_level((DBG_VERBOSE, "dispatching UP ARROW KEYPRESS\n"));

					/* 
					 * Fallthrough.
					 */
				}
			}
			else if (wParam == VK_DOWN)
			{
				if (!down_arrow_state)
				{
					debug_level((DBG_VERBOSE, "entering DOWN_ARROW_STATE\n"));
					down_arrow_state = 1;
					return 0;
				}
				else
				{
					debug_level((DBG_VERBOSE, "dispatching DOWN ARROW KEYPRESS\n"));

					/* 
					 * Fallthrough.
					 */
				}
			}
		}
		else /* Key Released */
		{
			/*
			 * Key is being released.  If we were still in the up or down
			 * arrow states then clear those flags and send the event.
			 */

			if (wParam == VK_UP && up_arrow_state)
			{
				debug_level((DBG_VERBOSE, "dispatching UP ARROW KEYPRESS for KEY RELEASE\n"));
				up_arrow_state = 0;

				goto check_for_arrow_key; /* Aaargh! Restructure me, please! */
			}
			else if (wParam == VK_DOWN && down_arrow_state)
			{
				debug_level((DBG_VERBOSE, "dispatching DOWN ARROW KEYPRESS for KEY RELEASE\n"));
				down_arrow_state = 0;

				goto check_for_arrow_key; /* Aaargh! Restructure me, please! */
			}
			else
			{
				/*
				 * It is actually expected that we will have a "mismatched"
				 * state here, because we clear the state flag on the function
				 * key handlers if we get those events.  We want to silently
				 * eat the message in this case, as that is actually the right
				 * thing to do here.
				 */

				debug_level((DBG_VERBOSE, "ARROW KEY released after we already handled the event due to a function key, ignoring the message\n"));

				return 0;
			}
			return 0;
		}
		break;

	case VK_F20: /* Scroll wheel */
	case VK_F21: /* Keyboard Blue/Fn key */

		/*
		 * If the last key was up and we're processing an up arrow, or we have
		 * already consumed the down arrow state and have an up arrow state
		 * still, then create a page up event.
		 */

		if ((last_arrow_key == VK_UP && up_arrow_state) ||
		    (!down_arrow_state && up_arrow_state))
		{
			debug_level((DBG_VERBOSE, "generate PAGE UP event\n"));

			up_arrow_state = 0;

			*p++ = 0x1B;
			*p++ = '[';
			*p++ = '5';
			*p++ = '~';

			return (p - output);
		}

		/*
		 * If the last key was down and we're processing an down arrow, or we
		 * have already consumed the up arrow state and have a down arrow state
		 * still, then create a page down event.
		 */

		if ((last_arrow_key == VK_DOWN && down_arrow_state) ||
		    (!up_arrow_state && down_arrow_state))
		{
			debug_level((DBG_VERBOSE, "generate PAGE DOWN event\n"));

			down_arrow_state = 0;

			*p++ = 0x1B;
			*p++ = '[';
			*p++ = '6';
			*p++ = '~';

			return (p - output);
		}

		debug_level((DBG_VERBOSE, "ignoring other function key %02X event\n", wParam));

		return 0;

	default:
		/*
		 * If we got a key release event then eat it here, as the rest of the
		 * routine only expects to deal with key down events.
		 */

		if ((lParam & 0x80000000))
			return -1;

		break;

	}

#endif /* defined(CONFIG_HACK_PAGE_KEYS) */

	if (wParam == VK_MENU && (HIWORD(lParam) & KF_EXTENDED)) {
		keystate[VK_RMENU] = keystate[VK_MENU];
	}

	/* Nastyness with NUMLock - Shift-NUMLock is left alone though */
	if ((funky_type == FUNKY_VT400 ||
	     (funky_type <= FUNKY_LINUX && term->app_keypad_keys &&
	      !no_applic_k))
	    && wParam == VK_NUMLOCK && !(keystate[VK_SHIFT] & 0x80)) {

	    wParam = VK_EXECUTE;

	    /* UnToggle NUMLock */
	    if ((HIWORD(lParam) & (KF_UP | KF_REPEAT)) == 0)
		keystate[VK_NUMLOCK] ^= 1;
	}
    }

    /* Disable Auto repeat if required */
    if (term->repeat_off &&
	(HIWORD(lParam) & (KF_UP | KF_REPEAT)) == KF_REPEAT)
	return 0;

    if ((HIWORD(lParam) & KF_ALTDOWN) && (keystate[VK_RMENU] & 0x80) == 0)
	left_alt = 1;

    key_down = ((HIWORD(lParam) & KF_UP) == 0);

    /* Make sure Ctrl-ALT is not the same as AltGr for ToAscii unless told. */
    if (left_alt && (keystate[VK_CONTROL] & 0x80)) {
	if (ctrlaltkeys)
	    keystate[VK_MENU] = 0;
	else {
	    keystate[VK_RMENU] = 0x80;
	    left_alt = 0;
	}
    }

    scan = (HIWORD(lParam) & (KF_UP | KF_EXTENDED | 0xFF));
    shift_state = ((keystate[VK_SHIFT] & 0x80) != 0)
	+ ((keystate[VK_CONTROL] & 0x80) != 0) * 2;

    /* Note if AltGr was pressed and if it was used as a compose key */
    if (!compose_state) {
	compose_keycode = 0x100;
	if (conf_get_int(conf, CONF_compose_key)) {
	    if (wParam == VK_MENU && (HIWORD(lParam) & KF_EXTENDED))
		compose_keycode = wParam;
	}
	if (wParam == VK_APPS)
	    compose_keycode = wParam;
    }

    if (wParam == compose_keycode) {
	if (compose_state == 0
	    && (HIWORD(lParam) & (KF_UP | KF_REPEAT)) == 0) compose_state =
		1;
	else if (compose_state == 1 && (HIWORD(lParam) & KF_UP))
	    compose_state = 2;
	else
	    compose_state = 0;
    } else if (compose_state == 1 && wParam != VK_CONTROL)
	compose_state = 0;

    if (compose_state > 1 && left_alt)
	compose_state = 0;

    /* Sanitize the number pad if not using a PC NumPad */
    if (left_alt || (term->app_keypad_keys && !no_applic_k
		     && funky_type != FUNKY_XTERM)
	|| funky_type == FUNKY_VT400 || nethack_keypad || compose_state) {
	if ((HIWORD(lParam) & KF_EXTENDED) == 0) {
	    int nParam = 0;
	    switch (wParam) {
	      case VK_INSERT:
		nParam = VK_NUMPAD0;
		break;
	      case VK_END:
		nParam = VK_NUMPAD1;
		break;
	      case VK_DOWN:
		nParam = VK_NUMPAD2;
		break;
	      case VK_NEXT:
		nParam = VK_NUMPAD3;
		break;
	      case VK_LEFT:
		nParam = VK_NUMPAD4;
		break;
	      case VK_CLEAR:
		nParam = VK_NUMPAD5;
		break;
	      case VK_RIGHT:
		nParam = VK_NUMPAD6;
		break;
	      case VK_HOME:
		nParam = VK_NUMPAD7;
		break;
	      case VK_UP:
		nParam = VK_NUMPAD8;
		break;
	      case VK_PRIOR:
		nParam = VK_NUMPAD9;
		break;
	      case VK_DELETE:
		nParam = VK_DECIMAL;
		break;
	    }
	    if (nParam) {
		if (keystate[VK_NUMLOCK] & 1)
		    shift_state |= 1;
		wParam = nParam;
	    }
	}
    }

    /* If a key is pressed and AltGr is not active */
    if (key_down && (keystate[VK_RMENU] & 0x80) == 0 && !compose_state) {
	/* Okay, prepare for most alts then ... */
	if (left_alt)
	    *p++ = '\033';

	/* Lets see if it's a pattern we know all about ... */
	if (wParam == VK_PRIOR && shift_state == 1) {
	    SendMessage(hwnd, WM_VSCROLL, SB_PAGEUP, 0);
	    return 0;
	}
	if (wParam == VK_PRIOR && shift_state == 2) {
	    SendMessage(hwnd, WM_VSCROLL, SB_LINEUP, 0);
	    return 0;
	}
	if (wParam == VK_NEXT && shift_state == 1) {
	    SendMessage(hwnd, WM_VSCROLL, SB_PAGEDOWN, 0);
	    return 0;
	}
	if (wParam == VK_NEXT && shift_state == 2) {
	    SendMessage(hwnd, WM_VSCROLL, SB_LINEDOWN, 0);
	    return 0;
	}
	if ((wParam == VK_PRIOR || wParam == VK_NEXT) && shift_state == 3) {
	    term_scroll_to_selection(term, (wParam == VK_PRIOR ? 0 : 1));
	    return 0;
	}
	if (wParam == VK_INSERT && shift_state == 1) {
	    request_paste(NULL);
	    return 0;
	}
	if (left_alt && wParam == VK_F4 && conf_get_int(conf, CONF_alt_f4)) {
	    return -1;
	}
	if (left_alt && wParam == VK_SPACE && conf_get_int(conf,
							   CONF_alt_space)) {
	    SendMessage(hwnd, WM_SYSCOMMAND, SC_KEYMENU, 0);
	    return -1;
	}
	if (left_alt && wParam == VK_RETURN &&
	    conf_get_int(conf, CONF_fullscreenonaltenter) &&
	    (conf_get_int(conf, CONF_resize_action) != RESIZE_DISABLED)) {
 	    if ((HIWORD(lParam) & (KF_UP | KF_REPEAT)) != KF_REPEAT)
 		flip_full_screen();
	    return -1;
	}
	/* Control-Numlock for app-keypad mode switch */
	if (wParam == VK_PAUSE && shift_state == 2) {
	    term->app_keypad_keys ^= 1;
	    return 0;
	}

	/* Nethack keypad */
	if (nethack_keypad && !left_alt) {
	    switch (wParam) {
	      case VK_NUMPAD1:
		*p++ = "bB\002\002"[shift_state & 3];
		return p - output;
	      case VK_NUMPAD2:
		*p++ = "jJ\012\012"[shift_state & 3];
		return p - output;
	      case VK_NUMPAD3:
		*p++ = "nN\016\016"[shift_state & 3];
		return p - output;
	      case VK_NUMPAD4:
		*p++ = "hH\010\010"[shift_state & 3];
		return p - output;
	      case VK_NUMPAD5:
		*p++ = shift_state ? '.' : '.';
		return p - output;
	      case VK_NUMPAD6:
		*p++ = "lL\014\014"[shift_state & 3];
		return p - output;
	      case VK_NUMPAD7:
		*p++ = "yY\031\031"[shift_state & 3];
		return p - output;
	      case VK_NUMPAD8:
		*p++ = "kK\013\013"[shift_state & 3];
		return p - output;
	      case VK_NUMPAD9:
		*p++ = "uU\025\025"[shift_state & 3];
		return p - output;
	    }
	}

	/* Application Keypad */
	if (!left_alt) {
	    int xkey = 0;

	    if (funky_type == FUNKY_VT400 ||
		(funky_type <= FUNKY_LINUX &&
		 term->app_keypad_keys && !no_applic_k)) switch (wParam) {
		  case VK_EXECUTE:
		    xkey = 'P';
		    break;
		  case VK_DIVIDE:
		    xkey = 'Q';
		    break;
		  case VK_MULTIPLY:
		    xkey = 'R';
		    break;
		  case VK_SUBTRACT:
		    xkey = 'S';
		    break;
		}
	    if (term->app_keypad_keys && !no_applic_k)
		switch (wParam) {
		  case VK_NUMPAD0:
		    xkey = 'p';
		    break;
		  case VK_NUMPAD1:
		    xkey = 'q';
		    break;
		  case VK_NUMPAD2:
		    xkey = 'r';
		    break;
		  case VK_NUMPAD3:
		    xkey = 's';
		    break;
		  case VK_NUMPAD4:
		    xkey = 't';
		    break;
		  case VK_NUMPAD5:
		    xkey = 'u';
		    break;
		  case VK_NUMPAD6:
		    xkey = 'v';
		    break;
		  case VK_NUMPAD7:
		    xkey = 'w';
		    break;
		  case VK_NUMPAD8:
		    xkey = 'x';
		    break;
		  case VK_NUMPAD9:
		    xkey = 'y';
		    break;

		  case VK_DECIMAL:
		    xkey = 'n';
		    break;
		  case VK_ADD:
		    if (funky_type == FUNKY_XTERM) {
			if (shift_state)
			    xkey = 'l';
			else
			    xkey = 'k';
		    } else if (shift_state)
			xkey = 'm';
		    else
			xkey = 'l';
		    break;

		  case VK_DIVIDE:
		    if (funky_type == FUNKY_XTERM)
			xkey = 'o';
		    break;
		  case VK_MULTIPLY:
		    if (funky_type == FUNKY_XTERM)
			xkey = 'j';
		    break;
		  case VK_SUBTRACT:
		    if (funky_type == FUNKY_XTERM)
			xkey = 'm';
		    break;

		  case VK_RETURN:
		    if (HIWORD(lParam) & KF_EXTENDED)
			xkey = 'M';
		    break;
		}
	    if (xkey) {
		if (term->vt52_mode) {
		    if (xkey >= 'P' && xkey <= 'S')
			p += sprintf((char *) p, "\x1B%c", xkey);
		    else
			p += sprintf((char *) p, "\x1B?%c", xkey);
		} else
		    p += sprintf((char *) p, "\x1BO%c", xkey);
		return p - output;
	    }
	}

	if (wParam == VK_BACK && shift_state == 0) {	/* Backspace */
	    *p++ = (conf_get_int(conf, CONF_bksp_is_delete) ? 0x7F : 0x08);
	    *p++ = 0;
	    return -2;
	}
	if (wParam == VK_BACK && shift_state == 1) {	/* Shift Backspace */
	    /* We do the opposite of what is configured */
	    *p++ = (conf_get_int(conf, CONF_bksp_is_delete) ? 0x08 : 0x7F);
	    *p++ = 0;
	    return -2;
	}
	if (wParam == VK_TAB && shift_state == 1) {	/* Shift tab */
	    *p++ = 0x1B;
	    *p++ = '[';
	    *p++ = 'Z';
	    return p - output;
	}
	if (wParam == VK_SPACE && shift_state == 2) {	/* Ctrl-Space */
	    *p++ = 0;
	    return p - output;
	}
	if (wParam == VK_SPACE && shift_state == 3) {	/* Ctrl-Shift-Space */
	    *p++ = 160;
	    return p - output;
	}
	if (wParam == VK_CANCEL && shift_state == 2) {	/* Ctrl-Break */
		if (back)
			back->special(backhandle, TS_BRK);
		return 0;
	}
	if (wParam == VK_PAUSE) {      /* Break/Pause */
	    *p++ = 26;
	    *p++ = 0;
	    return -2;
	}
	/* Control-2 to Control-8 are special */
	if (shift_state == 2 && wParam >= '2' && wParam <= '8') {
	    *p++ = "\000\033\034\035\036\037\177"[wParam - '2'];
	    return p - output;
	}
	if (shift_state == 2 && (wParam == 0xBD || wParam == 0xBF)) {
	    *p++ = 0x1F;
	    return p - output;
	}
	if (shift_state == 2 && (wParam == 0xDF || wParam == 0xDC)) {
	    *p++ = 0x1C;
	    return p - output;
	}
	if (shift_state == 3 && wParam == 0xDE) {
	    *p++ = 0x1E;	       /* Ctrl-~ == Ctrl-^ in xterm at least */
	    return p - output;
	}
	if (shift_state == 0 && wParam == VK_RETURN && term->cr_lf_return) {
	    *p++ = '\r';
	    *p++ = '\n';
	    return p - output;
	}

	/*
	 * Next, all the keys that do tilde codes. (ESC '[' nn '~',
	 * for integer decimal nn.)
	 *
	 * We also deal with the weird ones here. Linux VCs replace F1
	 * to F5 by ESC [ [ A to ESC [ [ E. rxvt doesn't do _that_, but
	 * does replace Home and End (1~ and 4~) by ESC [ H and ESC O w
	 * respectively.
	 */
	code = 0;
	switch (wParam) {
	  case VK_F1:
	    code = (keystate[VK_SHIFT] & 0x80 ? 23 : 11);
	    break;
	  case VK_F2:
	    code = (keystate[VK_SHIFT] & 0x80 ? 24 : 12);
	    break;
	  case VK_F3:
	    code = (keystate[VK_SHIFT] & 0x80 ? 25 : 13);
	    break;
	  case VK_F4:
	    code = (keystate[VK_SHIFT] & 0x80 ? 26 : 14);
	    break;
	  case VK_F5:
	    code = (keystate[VK_SHIFT] & 0x80 ? 28 : 15);
	    break;
	  case VK_F6:
	    code = (keystate[VK_SHIFT] & 0x80 ? 29 : 17);
	    break;
	  case VK_F7:
	    code = (keystate[VK_SHIFT] & 0x80 ? 31 : 18);
	    break;
	  case VK_F8:
	    code = (keystate[VK_SHIFT] & 0x80 ? 32 : 19);
	    break;
	  case VK_F9:
	    code = (keystate[VK_SHIFT] & 0x80 ? 33 : 20);
	    break;
	  case VK_F10:
	    code = (keystate[VK_SHIFT] & 0x80 ? 34 : 21);
	    break;
	  case VK_F11:
	    code = 23;
	    break;
	  case VK_F12:
	    code = 24;
	    break;
	  case VK_F13:
	    code = 25;
	    break;
	  case VK_F14:
	    code = 26;
	    break;
	  case VK_F15:
	    code = 28;
	    break;
	  case VK_F16:
	    code = 29;
	    break;
	  case VK_F17:
	    code = 31;
	    break;
	  case VK_F18:
	    code = 32;
	    break;
	  case VK_F19:
	    code = 33;
	    break;
	  case VK_F20:
	    code = 34;
	    break;
	}
	if ((shift_state&2) == 0) switch (wParam) {
	  case VK_HOME:
	    code = 1;
	    break;
	  case VK_INSERT:
	    code = 2;
	    break;
	  case VK_DELETE:
	    code = 3;
	    break;
	  case VK_END:
	    code = 4;
	    break;
	  case VK_PRIOR:
	    code = 5;
	    break;
	  case VK_NEXT:

	    code = 6;
	    break;
	}
	/* Reorder edit keys to physical order */
	if (funky_type == FUNKY_VT400 && code <= 6)
	    code = "\0\2\1\4\5\3\6"[code];

	if (term->vt52_mode && code > 0 && code <= 6) {
	    p += sprintf((char *) p, "\x1B%c", " HLMEIG"[code]);
	    return p - output;
	}

	if (funky_type == FUNKY_SCO && code >= 11 && code <= 34) {
	    /* SCO function keys */
	    char codes[] = "MNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz@[\\]^_`{";
	    int index = 0;
	    switch (wParam) {
	      case VK_F1: index = 0; break;
	      case VK_F2: index = 1; break;
	      case VK_F3: index = 2; break;
	      case VK_F4: index = 3; break;
	      case VK_F5: index = 4; break;
	      case VK_F6: index = 5; break;
	      case VK_F7: index = 6; break;
	      case VK_F8: index = 7; break;
	      case VK_F9: index = 8; break;
	      case VK_F10: index = 9; break;
	      case VK_F11: index = 10; break;
	      case VK_F12: index = 11; break;
	    }
	    if (keystate[VK_SHIFT] & 0x80) index += 12;
	    if (keystate[VK_CONTROL] & 0x80) index += 24;
	    p += sprintf((char *) p, "\x1B[%c", codes[index]);
	    return p - output;
	}
	if (funky_type == FUNKY_SCO &&     /* SCO small keypad */
	    code >= 1 && code <= 6) {
	    char codes[] = "HL.FIG";
	    if (code == 3) {
		*p++ = '\x7F';
	    } else {
		p += sprintf((char *) p, "\x1B[%c", codes[code-1]);
	    }
	    return p - output;
	}
	if ((term->vt52_mode || funky_type == FUNKY_VT100P) && code >= 11 && code <= 24) {
	    int offt = 0;
	    if (code > 15)
		offt++;
	    if (code > 21)
		offt++;
	    if (term->vt52_mode)
		p += sprintf((char *) p, "\x1B%c", code + 'P' - 11 - offt);
	    else
		p +=
		    sprintf((char *) p, "\x1BO%c", code + 'P' - 11 - offt);
	    return p - output;
	}
	if (funky_type == FUNKY_LINUX && code >= 11 && code <= 15) {
	    p += sprintf((char *) p, "\x1B[[%c", code + 'A' - 11);
	    return p - output;
	}
	if (funky_type == FUNKY_XTERM && code >= 11 && code <= 14) {
	    if (term->vt52_mode)
		p += sprintf((char *) p, "\x1B%c", code + 'P' - 11);
	    else
		p += sprintf((char *) p, "\x1BO%c", code + 'P' - 11);
	    return p - output;
	}
	if ((code == 1 || code == 4) &&
	    conf_get_int(conf, CONF_rxvt_homeend)) {
	    p += sprintf((char *) p, code == 1 ? "\x1B[H" : "\x1BOw");
	    return p - output;
	}
	if (code) {
	    p += sprintf((char *) p, "\x1B[%d~", code);
	    return p - output;
	}

check_for_arrow_key:
	/*
	 * Now the remaining keys (arrows and Keypad 5. Keypad 5 for
	 * some reason seems to send VK_CLEAR to Windows...).
	 */
	{
	    char xkey = 0;
	    switch (wParam) {
	      case VK_UP:
		xkey = 'A';
		break;
	      case VK_DOWN:
		xkey = 'B';
		break;
	      case VK_RIGHT:
		xkey = 'C';
		break;
	      case VK_LEFT:
		xkey = 'D';
		break;
	      case VK_CLEAR:
		xkey = 'G';
		break;
	    }
	    if (xkey) {
		if (term->vt52_mode)
		    p += sprintf((char *) p, "\x1B%c", xkey);
		else {
		    int app_flg = (term->app_cursor_keys && !conf_get_int(conf, CONF_no_applic_c));
#if 0
		    /*
		     * RDB: VT100 & VT102 manuals both state the
		     * app cursor keys only work if the app keypad
		     * is on.
		     *
		     * SGT: That may well be true, but xterm
		     * disagrees and so does at least one
		     * application, so I've #if'ed this out and the
		     * behaviour is back to PuTTY's original: app
		     * cursor and app keypad are independently
		     * switchable modes. If anyone complains about
		     * _this_ I'll have to put in a configurable
		     * option.
		     */
		    if (!term->app_keypad_keys)
			app_flg = 0;
#endif
		    /* Useful mapping of Ctrl-arrows */
		    if (shift_state == 2)
			app_flg = !app_flg;

		    if (app_flg)
			p += sprintf((char *) p, "\x1BO%c", xkey);
		    else
			p += sprintf((char *) p, "\x1B[%c", xkey);
		}
		return p - output;
	    }
	}

	/*
	 * Finally, deal with Return ourselves. (Win95 seems to
	 * foul it up when Alt is pressed, for some reason.)
	 */
	if (wParam == VK_RETURN) {     /* Return */
	    *p++ = 0x0D;
	    *p++ = 0;
	    return -2;
	}

	if (left_alt && wParam >= VK_NUMPAD0 && wParam <= VK_NUMPAD9)
	    alt_sum = alt_sum * 10 + wParam - VK_NUMPAD0;
	else
	    alt_sum = 0;
    }

    /* Okay we've done everything interesting; let windows deal with
     * the boring stuff */
    {
	BOOL capsOn=0;

	/* helg: clear CAPS LOCK state if caps lock switches to cyrillic */
	if(keystate[VK_CAPITAL] != 0 &&
	   conf_get_int(conf, CONF_xlat_capslockcyr)) {
	    capsOn= !left_alt;
	    keystate[VK_CAPITAL] = 0;
	}

#if 0 /* WinCE? */	/* XXX how do we know what the max size of the keys array should
	 * be is? There's indication on MS' website of an Inquire/InquireEx
	 * functioning returning a KBINFO structure which tells us. */
	if (osVersion.dwPlatformId == VER_PLATFORM_WIN32_NT && p_ToUnicodeEx) {
	    r = p_ToUnicodeEx(wParam, scan, keystate, keys,
			      lenof(keys), 0, kbd_layout);
	} else
#endif
	{
	    /* XXX 'keys' parameter is declared in MSDN documentation as
	     * 'LPWORD lpChar'.
	     * The experience of a French user indicates that on
	     * Win98, WORD[] should be passed in, but on Win2K, it should
	     * be BYTE[]. German WinXP and my Win2K with "US International"
	     * driver corroborate this.
	     * Experimentally I've conditionalised the behaviour on the
	     * Win9x/NT split, but I suspect it's worse than that.
	     * See wishlist item `win-dead-keys' for more horrible detail
	     * and speculations. */
	    int i;
	    static WORD keys[3];
	    static BYTE keysb[3];
	    r = ToAsciiEx(wParam, scan, keystate, keys, 0, kbd_layout);
	    if (r > 0) {
	        for (i = 0; i < r; i++) {
	            keysb[i] = (BYTE)keys[i];
	        }
	        MultiByteToWideChar(CP_ACP, 0, (LPCSTR)keysb, r,
				    keys, lenof(keys));
	    }
	}
#ifdef SHOW_TOASCII_RESULT
	if(debug_level & DBG_VERBOSE) {
		if (r == 1 && !key_down) {
			if (alt_sum) {
				if (in_utf(term) || ucsdata.dbcs_screenfont)
					debug_level((DBG_VERBOSE, ", (U+%04x)", alt_sum));
				else
					debug_level((DBG_VERBOSE, ", LCH(%d)", alt_sum));
			} else {
				debug_level((DBG_VERBOSE, ", ACH(%d)", keys[0]));
			}
		} else if (r > 0) {
			int r1;
			debug_level((DBG_VERBOSE, ", ASC("));

			for (r1 = 0; r1 < r; r1++) {
				debug_level((DBG_VERBOSE, "%s%d", r1 ? "," : "", keys[r1]));
			}
			debug_level((DBG_VERBOSE, ")"));
		}
	}
	debug_level((DBG_VERBOSE, ");\n"));
#endif
	if (r > 0) {
	    WCHAR keybuf;

	    /*
	     * Interrupt an ongoing paste. I'm not sure this is
	     * sensible, but for the moment it's preferable to
	     * having to faff about buffering things.
	     */
	    term_nopaste(term);

	    p = output;
	    for (i = 0; i < r; i++) {
		unsigned char ch = (unsigned char) keys[i];

		if (compose_state == 2 && (ch & 0x80) == 0 && ch > ' ') {
		    compose_char = ch;
		    compose_state++;
		    continue;
		}
		if (compose_state == 3 && (ch & 0x80) == 0 && ch > ' ') {
		    int nc;
		    compose_state = 0;

		    if ((nc = check_compose(compose_char, ch)) == -1) {
			MessageBeep(MB_ICONHAND);
			return 0;
		    }
		    keybuf = nc;
		    term_seen_key_event(term);
		    if (ldisc)
			luni_send(ldisc, &keybuf, 1, 1);
		    continue;
		}

		compose_state = 0;

		if (!key_down) {
		    if (alt_sum) {
			if (in_utf(term) || ucsdata.dbcs_screenfont) {
			    keybuf = alt_sum;
			    term_seen_key_event(term);
			    if (ldisc)
				luni_send(ldisc, &keybuf, 1, 1);
			} else {
			    ch = (char) alt_sum;
			    /*
			     * We need not bother about stdin
			     * backlogs here, because in GUI PuTTY
			     * we can't do anything about it
			     * anyway; there's no means of asking
			     * Windows to hold off on KEYDOWN
			     * messages. We _have_ to buffer
			     * everything we're sent.
			     */
			    term_seen_key_event(term);
			    if (ldisc)
				ldisc_send(ldisc, (char *)&ch, 1, 1);
			}
			alt_sum = 0;
		    } else {
			term_seen_key_event(term);
			if (ldisc)
			    lpage_send(ldisc, kbd_codepage, (char *)&ch, 1, 1);
		    }
		} else {
		    if(capsOn && ch < 0x80) {
			WCHAR cbuf[2];
			cbuf[0] = 27;
			cbuf[1] = xlat_uskbd2cyrllic(ch);
			term_seen_key_event(term);
			if (ldisc)
			    luni_send(ldisc, cbuf+!left_alt, 1+!!left_alt, 1);
		    } else {
			char cbuf[2];
			cbuf[0] = '\033';
			cbuf[1] = ch;
			term_seen_key_event(term);
			if (ldisc)
			    lpage_send(ldisc, kbd_codepage,
				       cbuf+!left_alt, 1+!!left_alt, 1);
		    }
		}
		show_mouseptr(0);
	    }

	    /* This is so the ALT-Numpad and dead keys work correctly. */
	    keys[0] = 0;

	    return p - output;
	}
	/* If we're definitly not building up an ALT-54321 then clear it */
	if (!left_alt)
	    keys[0] = 0;
	/* If we will be using alt_sum fix the 256s */
	else if (keys[0] && (in_utf(term) || ucsdata.dbcs_screenfont))
	    keys[0] = 10;
    }
    /*
     * ALT alone may or may not want to bring up the System menu.
     * If it's not meant to, we return 0 on presses or releases of
     * ALT, to show that we've swallowed the keystroke. Otherwise
     * we return -1, which means Windows will give the keystroke
     * its default handling (i.e. bring up the System menu).
     */
    if (wParam == VK_MENU && !conf_get_int(conf, CONF_alt_only))
	return 0;

    return -1;
}

void set_title(void *frontend, char *title)
{
	wchar_t *wtitle;

	debug_level((DBG_FUNCTION, "set_title(%s);\n", title));

	sfree(window_name);
	window_name = snewn(1 + strlen(title), char);
	strcpy(window_name, title);

	if (conf_get_int(conf, CONF_win_name_always)) {
		wtitle = ansi2unicode(title);
		SetWindowText(hwnd, wtitle);
		free(wtitle);
	}
}

void set_icon(void *frontend, char *title)
{
	wchar_t *wtitle;

	debug_level((DBG_FUNCTION, "set_icon(%s);\n", title));

	sfree(icon_name);
	icon_name = snewn(1 + strlen(title), char);
	strcpy(icon_name, title);

	if (!conf_get_int(conf, CONF_win_name_always)) {
		wtitle = ansi2unicode(title);
		SetWindowText(hwnd, wtitle);
		free(wtitle);
	}
}

void set_sbar(void *frontend, int total, int start, int page)
{
	SCROLLINFO si;

	debug_level((DBG_FUNCTION, "set_sbar(t: %i s: %i p: %i);\n", total, start, page));

	if (is_full_screen() ? !conf_get_int(conf, CONF_scrollbar_in_fullscreen) : !conf_get_int(conf, CONF_scrollbar))
		return;

	ZeroMemory(&si, sizeof(SCROLLINFO));
	si.cbSize = sizeof(si);
	si.fMask = SIF_ALL | SIF_DISABLENOSCROLL;
	si.nMin = 0;
	si.nMax = total - 1;
	si.nPage = page;
	si.nPos = start;
	if (hwnd)
		SetScrollInfo(hwnd, SB_VERT, &si, TRUE);
}

Context get_ctx(void *frontend)
{
	HDC hdc;

	debug_level((DBG_FUNCTION, "get_ctx();\n"));

	if (hwnd) {
		hdc = GetDC(hwnd);
		if (hdc && pal)
			SelectPalette(hdc, pal, FALSE);
		return hdc;
	} else {
		return NULL;
	}
}

void free_ctx(Context ctx)
{
	debug_level((DBG_FUNCTION, "free_ctx();\n"));
	SelectPalette(ctx, (HPALETTE)GetStockObject(DEFAULT_PALETTE), FALSE);
	ReleaseDC(hwnd, ctx);
}

static void real_palette_set(int n, int r, int g, int b)
{
	debug_level((DBG_FUNCTION, "real_palette_set(n: %i r: %i g: %i b: %i);\n", n, r, g, b));

	if (pal) {
		logpal->palPalEntry[n].peRed = r;
		logpal->palPalEntry[n].peGreen = g;
		logpal->palPalEntry[n].peBlue = b;
		logpal->palPalEntry[n].peFlags = PC_NOCOLLAPSE;
		colours[n] = PALETTERGB(r, g, b);
		SetPaletteEntries(pal, 0, NALLCOLOURS, logpal->palPalEntry);
	} else {
		colours[n] = RGB(r, g, b);
	}
}

void palette_set(void *frontend, int n, int r, int g, int b)
{
    debug_level((DBG_FUNCTION, "palette_set(n: %i r: %i g: %i b: %i);\n", n, r, g, b));

    if (n >= 16)
	n += 256 - 16;
    if (n > NALLCOLOURS)
	return;
    real_palette_set(n, r, g, b);
    if (pal) {
	HDC hdc = get_ctx(frontend);
	RealizePalette(hdc);
	free_ctx(hdc);
    } else {
	if (n == (ATTR_DEFBG>>ATTR_BGSHIFT))
	    /* If Default Background changes, we need to ensure any
	     * space between the text area and the window border is
	     * redrawn. */
	    InvalidateRect(hwnd, NULL, TRUE);
    }
}

void palette_reset(void *frontend)
{
    int i;

    debug_level((DBG_FUNCTION, "palette_reset();\n"));

    /* And this */
    for (i = 0; i < NALLCOLOURS; i++) {
	if (pal) {
	    logpal->palPalEntry[i].peRed = defpal[i].rgbtRed;
	    logpal->palPalEntry[i].peGreen = defpal[i].rgbtGreen;
	    logpal->palPalEntry[i].peBlue = defpal[i].rgbtBlue;
	    logpal->palPalEntry[i].peFlags = 0;
	    colours[i] = PALETTERGB(defpal[i].rgbtRed,
				    defpal[i].rgbtGreen,
				    defpal[i].rgbtBlue);
	} else
	    colours[i] = RGB(defpal[i].rgbtRed,
			     defpal[i].rgbtGreen, defpal[i].rgbtBlue);
    }

    if (pal) {
	HDC hdc;
	SetPaletteEntries(pal, 0, NALLCOLOURS, logpal->palPalEntry);
	hdc = get_ctx(frontend);
	RealizePalette(hdc);
	free_ctx(hdc);
    } else {
	/* Default Background may have changed. Ensure any space between
	 * text area and window border is redrawn. */
	InvalidateRect(hwnd, NULL, TRUE);
    }
}

static int is_shift_pressed(void)
{
    BYTE keystate[256];
    int r = GetKeyboardState(keystate);
    if (!r)
	return FALSE;
    if (keystate[VK_SHIFT] & 0x80)
	return TRUE;
    return FALSE;
}

/*
 * Note: unlike write_aclip() this will not append a nul.
 */
void write_clip(void *frontend, wchar_t * data, int *attr, int len, int must_deselect)
{
    HLOCAL clipdata;

    debug_level((DBG_FUNCTION, "write_clip(%i);\n", len));

    clipdata = LocalAlloc(LPTR, (len + 1) * sizeof(wchar_t));

    if (!clipdata) {
	LocalFree(clipdata);
	return;
    }

    CopyMemory(clipdata, data, len * sizeof(wchar_t));

    if (!must_deselect)
	SendMessage(hwnd, WM_IGNORE_CLIP, TRUE, 0);

    if (OpenClipboard(hwnd)) {
	EmptyClipboard();
	SetClipboardData(CF_UNICODETEXT, clipdata);
	CloseClipboard();
    } else {
	LocalFree(clipdata);
    }

    if (!must_deselect)
	SendMessage(hwnd, WM_IGNORE_CLIP, FALSE, 0);
}

static DWORD WINAPI clipboard_read_threadfunc(void *param)
{
    HWND hwnd = (HWND)param;
    HANDLE clipdata;

    if (OpenClipboard(NULL)) {
	if ((clipdata = (HANDLE)GetClipboardData(CF_UNICODETEXT))) {
	    SendMessage(hwnd, WM_GOT_CLIPDATA, (WPARAM)1, (LPARAM)clipdata);
	}
	CloseClipboard();
    }

    return 0;
}

static int process_clipdata(HLOCAL clipdata, int unicode)
{
    sfree(clipboard_contents);
    clipboard_contents = NULL;
    clipboard_length = 0;

    if (unicode) {
	wchar_t *p = (wchar_t *) clipdata;
	wchar_t *p2;

	if (p) {
	    /* Unwilling to rely on Windows having wcslen() */
	    for (p2 = p; *p2; p2++);
	    clipboard_length = p2 - p;
	    clipboard_contents = snewn(clipboard_length + 1, wchar_t);
	    memcpy(clipboard_contents, p, clipboard_length * sizeof(wchar_t));
	    clipboard_contents[clipboard_length] = L'\0';
	    return TRUE;
	}
    } else {
	char *s = (char *) clipdata;
	int i;

	if (s) {
	    i = MultiByteToWideChar(CP_ACP, 0, s, strlen(s) + 1, 0, 0);
	    clipboard_contents = snewn(i, wchar_t);
	    MultiByteToWideChar(CP_ACP, 0, s, strlen(s) + 1,
				clipboard_contents, i);
	    clipboard_length = i - 1;
	    clipboard_contents[clipboard_length] = L'\0';
	    return TRUE;
	}
    }

    return FALSE;
}

void request_paste(void *frontend)
{
    /*
     * I always thought pasting was synchronous in Windows; the

     * clipboard access functions certainly _look_ synchronous,
     * unlike the X ones. But in fact it seems that in some
     * situations the contents of the clipboard might not be
     * immediately available, and the clipboard-reading functions
     * may block. This leads to trouble if the application
     * delivering the clipboard data has to get hold of it by -
     * for example - talking over a network connection which is
     * forwarded through this very PuTTY.
     *
     * Hence, we spawn a subthread to read the clipboard, and do
     * our paste when it's finished. The thread will send a
     * message back to our main window when it terminates, and
     * that tells us it's OK to paste.
     */
    DWORD in_threadid; /* required for Win9x */
    CreateThread(NULL, 0, clipboard_read_threadfunc,
		 hwnd, 0, &in_threadid);
}

void get_clip(void *frontend, wchar_t **p, int *len)
{
    debug_level((DBG_FUNCTION, "get_clip(%i);\n", len));

    if (p) {
	*p = clipboard_contents;
	*len = clipboard_length;
    }
}

/*
 * Print a message box and perform a fatal exit.
 */
void fatalbox(char *fmt, ...)
{
	va_list ap;
	char *stuff;
	char *cap = dupprintf(load_ansi_string(IDS_WARN_CAPTION_FALAL_ERROR), appname);

	debug_level((DBG_FUNCTION, "fatalbox();\n"));

	va_start(ap, fmt);
	stuff = dupvprintf(fmt, ap);
	va_end(ap);

	fatalbox_visible = TRUE;
	MessageBoxA(hwnd, stuff, cap, MB_ICONERROR | MB_OK);
	fatalbox_visible = FALSE;

	sfree(stuff);
	sfree(cap);
	cleanup_exit(1);
}

/*
 * Print a modal (Really Bad) message box and perform a fatal exit.
 */
void modalfatalbox(char *fmt, ...)
{
	va_list ap;
	char *stuff;
	char *cap = dupprintf(load_ansi_string(IDS_WARN_CAPTION_FALAL_ERROR), appname);

	va_start(ap, fmt);
	stuff = dupvprintf(fmt, ap);
	va_end(ap);

	fatalbox_visible = TRUE;
	MessageBoxA(hwnd, stuff, cap, MB_APPLMODAL | MB_ICONERROR | MB_OK);
	fatalbox_visible = FALSE;

	sfree(stuff);
	sfree(cap);
	cleanup_exit(1);
}

/*
 * Print a message box and don't close the connection.
 */
void nonfatal(char *fmt, ...)
{
	va_list ap;
	char *stuff;
	char *cap = dupprintf(load_ansi_string(IDS_WARN_CAPTION_ERROR), appname);

	va_start(ap, fmt);
	stuff = dupvprintf(fmt, ap);
	va_end(ap);

	MessageBoxA(hwnd, stuff, cap, MB_ICONERROR | MB_OK);

	sfree(stuff);
	sfree(cap);
}

/*
 * Beep.
 */
void do_beep(void *frontend, int mode)
{
	debug_level((DBG_FUNCTION, "do_beep(%i);\n", mode));

	if (mode == BELL_DEFAULT) {
		/*
		* For MessageBeep style bells, we want to be careful of
		* timing, because they don't have the nice property of
		* PlaySound bells that each one cancels the previous
		* active one. So we limit the rate to one per 50ms or so.
		*/
		static long lastbeep = 0;
		long beepdiff;

		beepdiff = GetTickCount() - lastbeep;
		if (beepdiff >= 0 && beepdiff < 50)
			return;
		MessageBeep(MB_OK);
		/*
		* The above MessageBeep call takes time, so we record the
		* time _after_ it finishes rather than before it starts.
		*/
		lastbeep = GetTickCount();
	} else if (mode == BELL_WAVEFILE) {
		wchar_t *wavfile;
		wavfile = ansi2unicode(filename_to_str(conf_get_filename(conf, CONF_bell_wavefile)));

		if (!PlaySound(wavfile, NULL, SND_ASYNC | SND_FILENAME)) {
			char buf[1024 + 80];
			char otherbuf[100];
			wchar_t *text, *caption;

			sprintf(buf, "Unable to play sound file\n%.1024s\n"
				"Using default sound instead", filename_to_str(conf_get_filename(conf, CONF_bell_wavefile)));
			sprintf(otherbuf, "%.70s Sound Error", appname);
			MessageBoxA(hwnd, buf, otherbuf, MB_OK | MB_ICONEXCLAMATION);
			conf_set_int(conf, CONF_beep, BELL_DEFAULT);
		}
		sfree(wavfile);
	}
}

/*
* Move the window in response to a server-side request.
*/
void move_window(void *frontend, int x, int y)
{
	int resize_action = conf_get_int(conf, CONF_resize_action);

	debug_level((DBG_FUNCTION, "move_window(x: %i y: %i);\n", x, y));

	if (resize_action == RESIZE_DISABLED ||
		resize_action == RESIZE_FONT ||
		IsZoomed(hwnd))
		return;

	SetWindowPos(hwnd, NULL, x, y, 0, 0, SWP_NOSIZE | SWP_NOZORDER);
}

/*
 * Move the window to the top or bottom of the z-order in response
 * to a server-side request.
 */
void set_zorder(void *frontend, int top)
{
	debug_level((DBG_FUNCTION, "set_zorder(%i);\n", top));

	if (conf_get_int(conf, CONF_alwaysontop))
		return; /* ignore */

	SetWindowPos(hwnd, top ? HWND_TOP : HWND_BOTTOM, 0, 0, 0, 0,
		     SWP_NOMOVE | SWP_NOSIZE);
}

/*
 * Refresh the window in response to a server-side request.
 */
void refresh_window(void *frontend)
{
	debug_level((DBG_FUNCTION, "refresh_window();\n"));
	InvalidateRect(hwnd, NULL, TRUE);
}

/*
 * Maximise or restore the window in response to a server-side
 * request.
 */
void set_zoomed(void *frontend, int zoomed)
{
	debug_level((DBG_FUNCTION, "set_zoomed(%i);\n", zoomed));

	if (IsZoomed(hwnd)) {
		if (!zoomed)
			ShowWindow(hwnd, SW_RESTORE);
	} else {
		if (zoomed)
			ShowWindow(hwnd, SW_MAXIMIZE);
	}
}

/*
 * Report the window's position, for terminal reports.
 */
void get_window_pos(void *frontend, int *x, int *y)
{
	RECT r;

	GetWindowRect(hwnd, &r);
	*x = r.left;
	*y = r.top;

	debug_level((DBG_FUNCTION, "get_window_pos(x: %i y: %i);\n", *x, *y));
}

/*
 * Report the window's pixel size, for terminal reports.
 */
void get_window_pixels(void *frontend, int *x, int *y)
{
	RECT r;

	GetWindowRect(hwnd, &r);
	*x = r.right - r.left;
	*y = r.bottom - r.top;

	debug_level((DBG_FUNCTION, "get_window_pixels(x: %i y: %i);\n", *x, *y));
}

/*
 * Return the window or icon title.
 */
char *get_window_title(void *frontend, int icon)
{
	debug_level((DBG_FUNCTION, "get_window_title(%i);\n", icon));
	return icon ? icon_name : window_name;
}

HWND get_taskbar(HWND hwnd)
{
	return FindWindowW(L"HHTaskbar", NULL);
}

/*
 * See if we're in full-screen mode.
 */
static int is_full_screen()
{
//    if (IsWindowVisible(get_taskbar(hwnd)))
//	return FALSE;
    return current_full_screen;
}

static int get_desktop_rect(RECT * ss)
{
	RECT rt;

	debug_level((DBG_FUNCTION, "get_desktop_rect();\n"));

	if (cetype == CEHPC) {
		GetClientRect(GetDesktopWindow(), ss);
	}
	else {
		HWND hTkWnd = get_taskbar(hwnd);

		ss->left = ss->top = 0;
		ss->right = GetSystemMetrics(SM_CXSCREEN);
		ss->bottom = GetSystemMetrics(SM_CYSCREEN);

		if (hTkWnd && IsWindowVisible(hTkWnd)) {
			GetWindowRect(get_taskbar(hwnd), &rt);
			ss->top += rt.bottom - rt.top;
		}

#ifdef SHELL_AYGSHELL
		if (as_SHFindMenuBar) {
			HWND hTbWnd = as_SHFindMenuBar(hwnd);
			if (hTbWnd && IsWindowVisible(hTbWnd)) {
				GetWindowRect(hTbWnd, &rt);
				ss->bottom -= rt.bottom - rt.top;
			}
		}
#endif
	}

	return 1;
}

/* Get the rect/size of a full screen window using the nearest available
 * monitor in multimon systems; default to something sensible if only
 * one monitor is present. */
static int get_fullscreen_rect(RECT * ss)
{
	RECT rt;

	debug_level((DBG_FUNCTION, "get_fullscreen_rect();\n"));

	ss->left = ss->top = 0;
	ss->right = GetSystemMetrics(SM_CXSCREEN);
	ss->bottom = GetSystemMetrics(SM_CYSCREEN);

	if (cetype != CEHPC) {
#ifdef SHELL_AYGSHELL
		if (as_SHFindMenuBar) {
			HWND hTbWnd = as_SHFindMenuBar(hwnd);
			if (hTbWnd && IsWindowVisible(hTbWnd)) {
				GetWindowRect(hTbWnd, &rt);
				ss->bottom -= rt.bottom - rt.top;
			}
		}
#endif
	}

	debug_level((DBG_FUNCTION, "(%d,%d,%d,%d)\n", ss->left, ss->top, ss->right, ss->bottom));

	return 1;
}


static void recalc_sip_rect(RECT * ss)
{
	SIPINFO si;

	if (cetype == CEHPC)
		return;

	ZeroMemory(&si, sizeof(SIPINFO));

	si.cbSize = sizeof(SIPINFO);

	if (SipGetInfo(&si) && (si.fdwFlags & SIPF_ON)) {
		//ss->bottom -= si.rcSipRect.bottom - si.rcSipRect.top;
		ss->bottom = si.rcVisibleDesktop.bottom;
	}
}

/*
 * Go full-screen. This should only be called when we are already
 * maximised.
 */
static void make_full_screen()
{
	RECT ss;

	current_full_screen = TRUE;

	/* Remove the window furniture. */
	if (conf_get_int(conf, CONF_scrollbar_in_fullscreen))
		set_window_style(hwnd, WS_VSCROLL, 0);
	else
		set_window_style(hwnd, 0, WS_VSCROLL);

	/* Resize ourselves to exactly cover the nearest monitor. */
	get_fullscreen_rect(&ss);
	recalc_sip_rect(&ss);
	SetWindowPos(hwnd, 0, ss.left, ss.top,
		     ss.right - ss.left,
		     ss.bottom - ss.top,
		     SWP_NOZORDER | SWP_FRAMECHANGED);

	ShowWindow(get_taskbar(hwnd), SW_HIDE);
#ifdef SHELL_AYGSHELL
	if (as_SHFullScreen)
		as_SHFullScreen(hwnd, SHFS_HIDETASKBAR | SHFS_HIDESTARTICON);
	if (as_SHDoneButton)
		as_SHDoneButton(hwnd, SHDB_HIDE);
#endif
	if (cetype == CEHPC)
		EnableWindow(get_taskbar(hwnd), FALSE);

	/* We may have changed size as a result */

	reset_window(0);

	/* Tick the menu item in the System and context menus. */
	{
		int i;
		for (i = 0; i < lenof(popup_menus); i++)
			CheckMenuItem(popup_menus[i].menu, IDM_FULLSCREEN, MF_CHECKED);
	}
}

/*
 * Clear the full-screen attributes.
 */
static void clear_full_screen()
{
	RECT ss;

	current_full_screen = FALSE;

	/* Reinstate the window furniture. */
	if (conf_get_int(conf, CONF_scrollbar))
		set_window_style(hwnd, WS_VSCROLL, 0);
	else
		set_window_style(hwnd, 0, WS_VSCROLL);

	if (cetype == CEHPC)
		EnableWindow(get_taskbar(hwnd), TRUE);
#ifdef SHELL_AYGSHELL
	if (as_SHDoneButton)
		as_SHDoneButton(hwnd, SHDB_SHOWCANCEL);
	if (as_SHFullScreen)
		as_SHFullScreen(hwnd, SHFS_SHOWTASKBAR | SHFS_SHOWSTARTICON);
#endif
	ShowWindow(get_taskbar(hwnd), SW_SHOW);

	/* Resize ourselves to exactly cover the nearest monitor. */
	get_desktop_rect(&ss);
	recalc_sip_rect(&ss);
	SetWindowPos(hwnd, 0, ss.left, ss.top,
		     ss.right - ss.left,
		     ss.bottom - ss.top,
		     SWP_NOZORDER | SWP_FRAMECHANGED);

	/* We may have changed size as a result */

	reset_window(0);

	/* Untick the menu item in the System and context menus. */
	{
		int i;
		for (i = 0; i < lenof(popup_menus); i++)
			CheckMenuItem(popup_menus[i].menu, IDM_FULLSCREEN, MF_UNCHECKED);
	}
}

/*
 * Toggle full-screen mode.
 */
static void flip_full_screen()
{
    if (is_full_screen()) {
	clear_full_screen();
    } else {
	make_full_screen();
    }
}

int from_backend(void *frontend, int is_stderr, const char *data, int len)
{
	debug_level((DBG_FUNCTION, "from_backend(%i);\n", len));
	return term_data(term, is_stderr, data, len);
}

int from_backend_untrusted(void *frontend, const char *data, int len)
{
	debug_level((DBG_FUNCTION, "from_backend_untrusted(%i);\n", len));
	return term_data_untrusted(term, data, len);
}

int from_backend_eof(void *frontend)
{
	return TRUE;   /* do respond to incoming EOF with outgoing */
}

int get_userpass_input(prompts_t *p, unsigned char *in, int inlen)
{
	int ret;

	debug_level((DBG_FUNCTION, "get_userpass_input();\n"));

	ret = cmdline_get_passwd_input(p, in, inlen);
	if (ret == -1)
		ret = term_get_userpass_input(term, p, in, inlen);
	return ret;
}

void agent_schedule_callback(void (*callback)(void *, void *, int),
			     void *callback_ctx, void *data, int len)
{
    struct agent_callback *c = snew(struct agent_callback);
    c->callback = callback;
    c->callback_ctx = callback_ctx;
    c->data = data;
    c->len = len;
    PostMessage(hwnd, WM_AGENT_CALLBACK, 0, (LPARAM)c);
}

static HWND CreateRpCommandBar(HWND hwnd)
{
	HWND hMenu;
#ifdef SHELL_AYGSHELL
	SHMENUBARINFO mbi;

	ZeroMemory(&mbi, sizeof(SHMENUBARINFO));
	mbi.cbSize     = sizeof(SHMENUBARINFO);
	mbi.hwndParent = hwnd;
	mbi.nToolBarId = IDM_MENU;
	mbi.hInstRes   = hinst;
	mbi.nBmpId     = 0;
	mbi.cBmpImages = 0;

	if (as_SHCreateMenuBar && as_SHCreateMenuBar(&mbi) && mbi.hwndMB) {
		popup_menus[0].menu = (HMENU)SendMessage(mbi.hwndMB, SHCMBM_GETSUBMENU, 0, IDM_MAIN_COMMAND1);
		popup_menus[1].menu = (HMENU)SendMessage(mbi.hwndMB, SHCMBM_GETSUBMENU, 0, IDM_MAIN_COMMAND2);
		override_menu_height = 0;

		return mbi.hwndMB;
	}
#endif
	hMenu = CommandBar_Create(hinst, hwnd, 1);
	CommandBar_InsertMenubar(hMenu,  hinst, IDM_MENU, 0);
	GetWindow(hMenu, GW_CHILD);
	CommandBar_DrawMenuBar(hMenu, 0);
	popup_menus[0].menu = GetSubMenu(CommandBar_GetMenu(hMenu, 0), 0);
	popup_menus[1].menu = GetSubMenu(CommandBar_GetMenu(hMenu, 0), 1);
	override_menu_height = CommandBar_Height(hMenu);

	return hMenu;
}

/*
 * Actually do the job requested by a WM_NETEVENT
 */
static void enact_pending_netevent(void)
{
	static int reentering = 0;
	extern int select_result(WPARAM, LPARAM);

	if (reentering)
		return;	/* don't unpend the pending */

	pending_netevent = FALSE;

	reentering = 1;
	select_result(pend_netevent_wParam, pend_netevent_lParam);
	reentering = 0;
}

static BOOL set_window_style(HWND hwnd, DWORD_PTR add, DWORD_PTR remove)
{
	DWORD_PTR s, ns;

	if (hwnd == NULL)
		return FALSE;

	s = GetWindowLongPtr(hwnd, GWL_STYLE);
	ns = (s & ~remove) | add;
	if (s == ns)
		return FALSE;

	SetWindowLongPtr(hwnd, GWL_STYLE, ns);
	SetWindowPos(hwnd, NULL, 0, 0, 0, 0,
		     SWP_NOSIZE | SWP_NOMOVE | SWP_NOZORDER | SWP_NOACTIVATE);

	return TRUE;
}

